/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#define _sph_ifs_simple_adi_run_C
#include <cpl.h>
#include "sph_ifs_simple_adi.h"
#include "sph_simple_adi.h"
#include "sph_error.h"
#include "sph_ifs_keywords.h"
#include "sph_fits.h"
#include "sph_keyword_manager.h"
#include <math.h>
#undef _sph_ifs_simple_adi_run_C
static
cpl_vector*
__sph_ifs_simple_adi_extract_angles( sph_ifs_simple_adi* self );

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides functionality for apertures, extending the functionality
 * as it exists for cpl_apertures.
 */
/*----------------------------------------------------------------------------*/
/**@{*/
cpl_error_code
sph_ifs_simple_adi_run( sph_ifs_simple_adi* self ) {
    sph_simple_adi*                 adi = NULL;
    sph_transform*                  transform = NULL;
    int                             nf          = 0;
    sph_cube*                       outcube     = NULL;
    int                             np          = 0;
    int                             pp          = 0;
    int                             ff          = 0;
    sph_master_frame*               mframe      = NULL;
    double                          angle       = 0.0;
    cpl_vector*                     angles       = NULL;

    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    nf = cpl_frameset_get_size(self->current_raw_frameset);
    cpl_ensure_code(nf > 0, CPL_ERROR_ILLEGAL_INPUT);

    np = sph_fits_get_nplanes( cpl_frame_get_filename(
            cpl_frameset_get_first(self->current_raw_frameset)), 0 );
    cpl_ensure_code(np > 0, CPL_ERROR_ILLEGAL_INPUT);


    if ( self->transform_method == SPH_TRANSFORM_METHOD_FFT ) {
        if ( self->filter_method == SPH_FFT_FILTER_METHOD_NONE ) {
            transform = sph_transform_new_fft(SPH_FFT_FILTER_METHOD_NONE,0.0,0.0);
        }
        else if ( self->filter_method == SPH_FFT_FILTER_METHOD_TH ) {
            transform = sph_transform_new_fft(SPH_FFT_FILTER_METHOD_TH,self->filter_radius,0.0);
        }
        else if ( self->filter_method == SPH_FFT_FILTER_METHOD_FERMI ) {
            transform = sph_transform_new_fft(SPH_FFT_FILTER_METHOD_FERMI,self->filter_radius,self->fermi_temp);
        }
        else if ( self->filter_method == SPH_FFT_FILTER_METHOD_BUTTER ) {
            transform = sph_transform_new_fft(SPH_FFT_FILTER_METHOD_BUTTER,self->butter_pass,self->butter_stop);
        }
    }
    else if ( self->transform_method == 1) {
        transform = sph_transform_new_cpl_warp();
    }
    cpl_ensure_code(transform,cpl_error_get_code());

    angles = __sph_ifs_simple_adi_extract_angles( self );
    outcube = sph_cube_new( self->outfilename );
    for (pp = 0; pp < np; ++pp) {
        adi = sph_simple_adi_new(transform);
        for (ff = 0; ff < nf; ++ff) {
            mframe = sph_master_frame_load( cpl_frame_get_filename(
                    cpl_frameset_get_frame(self->current_raw_frameset,ff)
                    ),
                    pp);
            if ( mframe ) {
                angle = cpl_vector_get( angles, ff );
                SPH_ERROR_ENSURE_GOTO_EXIT(
                		sph_simple_adi_append_image(
                				adi,
                				mframe->image,
                				mframe->badpixelmap,
                				mframe->ncombmap,
                				angle) == CPL_ERROR_NONE,
                				cpl_error_get_code()
                				);
            }
            sph_master_frame_delete(mframe); mframe = NULL;
        }
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
        //sph_simple_adi_calculate_speckle_frame(adi);
        mframe = sph_simple_adi_process_imlists(adi,
                self->coll_alg);
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT
        if ( mframe ) {
            sph_cube_append_master(outcube,mframe,pp);
        }
        sph_master_frame_delete(mframe); mframe = NULL;
        sph_simple_adi_delete(adi); adi = NULL;
    }

EXIT:
    sph_master_frame_delete(mframe); mframe = NULL;
    sph_simple_adi_delete(adi); adi = NULL;
    sph_cube_finalise_file(outcube->filename);
    sph_cube_delete(outcube); outcube = NULL;
    sph_transform_delete(transform); transform = NULL;
    cpl_vector_delete(angles); angles = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
cpl_vector*
__sph_ifs_simple_adi_extract_angles( sph_ifs_simple_adi* self ) {
    cpl_vector*             result = NULL;
    int                     nframes      = 0;
    int                     ff          = 0;
    double                  angle       = 0.0;
    cpl_propertylist*       prop        = NULL;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT, NULL);

    nframes = cpl_frameset_get_size(self->current_raw_frameset);
    result = cpl_vector_new(nframes);
    cpl_vector_fill(result, 0.0 );
    for (ff = 0; ff < nframes; ++ff) {
        prop = sph_keyword_manager_load_properties(
                cpl_frame_get_filename( cpl_frameset_get_frame(
                        self->current_raw_frameset,ff)), 0 );
        if ( prop ) {
            if ( cpl_propertylist_has(prop,
                    SPH_IFS_KEYWORD_ROT_ANGLE) )
            {
                angle = cpl_propertylist_get_double(prop,
                        SPH_IFS_KEYWORD_ROT_ANGLE);
            }
            cpl_propertylist_delete(prop); prop = NULL;
        }
        cpl_vector_set(result,ff,angle);
    }
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;


}
/**@}*/
