/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ifs_keywords.h"
#include "sph_ifs_tags.h"
#include "sph_ifs_spectra_positions.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_fft.h"
#include "sph_utils.h"
#include "sph_ifs_lenslet_model.h"
#include "sph_point_pattern.h"
#include "sph_ifs_proc_apply_master_dark.h"
#include "sph_polynomial_distortion_model.h"
#include "sph_framecombination.h"
#include <gsl_rng.h>

#define IFS_MIN_SPECWIDTH 5.0

/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IFS_SPECTRA_POSITIONS_GENERAL;
extern sph_error_code SPH_IFS_SPECTRA_POSITIONS_FRAMES_MISSING;

static
sph_point_pattern*
sph_ifs_spectra_positions_extract_pp(
        cpl_image* specid_image,
        int nobs,
	int minpix )
{
    sph_point_pattern* point_pattern = NULL;
    double xpos,ypos;
    int aa;
    int        xx = 0;
    int        yy = 0;
    int        bpix = 0;
    int        id = 0;
    double        nnpix = 0;
    cpl_vector*        bottom = NULL;
    cpl_vector*        left = NULL;
    cpl_vector*        top = NULL;
    cpl_vector*        right = NULL;
    cpl_vector*	       npix = NULL;

    point_pattern = sph_point_pattern_new();

    bottom = cpl_vector_new(nobs);
    left = cpl_vector_new(nobs);
    top = cpl_vector_new(nobs);
    right = cpl_vector_new(nobs);
    npix = cpl_vector_new(nobs);
    for (xx = 0; xx < nobs; ++xx) {
        cpl_vector_set(left,xx,(double)INT_MAX);
        cpl_vector_set(bottom,xx,(double)INT_MAX);
        cpl_vector_set(right,xx,-1.0);
        cpl_vector_set(top,xx,-1.0);
        cpl_vector_set(npix,xx,0.0);
    }
    for (xx = 0; xx < cpl_image_get_size_x(specid_image); ++xx) {
        for (yy = 0; yy < cpl_image_get_size_y(specid_image); ++yy) {
            bpix = 0;
            id = cpl_image_get(specid_image,xx+1,yy+1,&bpix);
            if ( id > 0 && bpix == 0 && id <= nobs ) {
                xpos = cpl_vector_get(left,id-1);
                ypos = cpl_vector_get(bottom,id-1);
                nnpix = cpl_vector_get(npix,id-1);
                if ( xx < xpos ) cpl_vector_set(left,id-1,(double)xx);
                if ( yy < ypos ) cpl_vector_set(bottom,id-1,(double)yy);
                xpos = cpl_vector_get(right,id-1);
                ypos = cpl_vector_get(top,id-1);
                if ( xx > xpos ) cpl_vector_set(right,id-1,(double)xx);
                if ( yy > ypos ) cpl_vector_set(top,id-1,(double)yy);
                cpl_vector_set(npix,id-1,nnpix+1.0);
            }
        }
    }

    for (aa = 0; aa < nobs; ++aa) {
        xpos = 0.5 * ( cpl_vector_get(right,aa) + cpl_vector_get(left,aa) + 1.0 );
        ypos = 0.5 * ( cpl_vector_get(top,aa) + cpl_vector_get(bottom,aa) + 1.0 );
        if ( xpos < 0.99 * (double)INT_MAX && ypos < 0.99 * (double)INT_MAX ) {
            	nnpix = cpl_vector_get(npix,aa);
		if (nnpix > minpix )
			sph_point_pattern_add_point( point_pattern,xpos,ypos);
        }
    }
    cpl_vector_delete(top);
    cpl_vector_delete(bottom);
    cpl_vector_delete(left);
    cpl_vector_delete(right);
    cpl_vector_delete(npix);
    return point_pattern;
}

static
sph_error_code
sph_ifs_spectra_positions_determine_offset( sph_point_pattern* pp1,
        sph_point_pattern* pp2,
        int nsamples, double* offx, double* offy,
        int detsize)
{
    cpl_vector*        distsx = NULL;
    cpl_vector*        distsy = NULL;
    gsl_rng*        pRNG = NULL;
    int                ii    = 0;
    double            distx = 0.0;
    double            disty = 0.0;
    int                np1 = 0;
    int                np2    = 0;
    int                cc = 0;
    sph_point*        p = NULL;
    sph_point*        p2 = NULL;
    double			  centre_x = 0.0;
    double			  centre_y = 0.0;

    if ( nsamples < 1 ) {
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__,
                __LINE__,
                SPH_ERROR_ERROR,
                "Need to take at "
                "least 1 sample !");
        return CPL_ERROR_ILLEGAL_INPUT;
    }
    pRNG = gsl_rng_alloc( gsl_rng_taus );


    np1 = sph_point_pattern_get_size( pp1 );
    np2 = sph_point_pattern_get_size( pp2 );
    centre_x = detsize / 2.0;
    centre_y = detsize / 2.0;
    cc = 0;

    for (ii = 0; ii < np1; ++ii)
    {
        p = sph_point_pattern_get_point( pp1, ii );
        if ( fabs(p->x - centre_x) < nsamples && fabs(p->y - centre_y) < nsamples ) {
            cc++;
        }
        sph_point_delete(p); p = NULL;
    }
    distsx = cpl_vector_new( cc );
    distsy = cpl_vector_new( cc );
    cc = 0;
    for (ii = 0; ii < np1; ++ii)
    {
        p = sph_point_pattern_get_point( pp1, ii );
        if ( fabs(p->x - centre_x) < nsamples && fabs(p->y - centre_y) < nsamples ) {
        	if ( ( p2 = sph_point_pattern_find_closest(pp2, p->x, p->y ) ) != NULL )
        	{
        		distx = ( p->x - p2->x );
        		disty = ( p->y - p2->y );
        		cpl_vector_set( distsx, cc, distx );
        		cpl_vector_set( distsy, cc, disty );
        		cc++;
        		sph_point_delete(p2); p2 = NULL;
        	}
        }
        sph_point_delete(p); p = NULL;
    }
    *offx = -1.0 * cpl_vector_get_median( distsx );
    *offy = -1.0 * cpl_vector_get_median( distsy );
    cpl_vector_delete( distsx ); distsx = NULL;
    cpl_vector_delete( distsy ); distsy = NULL;

    gsl_rng_free( pRNG ); pRNG = NULL;

    return CPL_ERROR_NONE;
}
static
double
sph_ifs_spectra_positions_get_angle_model(sph_ifs_lenslet_model* model) {
    sph_polygon*        poly = NULL;
    double              x1      = 0.0;
    double              x2      = 0.0;
    double              y1      = 0.0;
    double              y2      = 0.0;
    double      ll = 0.0;
    double angle = 0.0;

    poly = sph_ifs_lenslet_model_get_spec_region_pixel_coords(model,0,0,0.0,0.0);
    x1 = sph_polygon_get_midx(poly);
    y1 = sph_polygon_get_midy(poly);
    poly = sph_ifs_lenslet_model_get_spec_region_pixel_coords(model,-2,3,0.0,0.0);
    x2 = sph_polygon_get_midx(poly);
    y2 = sph_polygon_get_midy(poly);
    ll = ( x2 - x1 ) * ( x2 - x1 ) + ( y2 - y1 ) * ( y2 - y1 );

    angle = acos((x2-x1)/sqrt(ll)) / CPL_MATH_PI * 180.0;
    angle = 90.0 - angle;
    return angle;
}

static
double
sph_ifs_spectra_positions_get_scale_model(sph_ifs_lenslet_model* model) {
    sph_polygon*        poly = NULL;
    double              x1      = 0.0;
    double              x2      = 0.0;
    poly = sph_ifs_lenslet_model_get_spec_region_pixel_coords(model,0,0,0.0,0.0);
    x1 = poly->points[0]->x;
    poly = sph_ifs_lenslet_model_get_spec_region_pixel_coords(model,1,-1,0.0,0.0);
    x2 = poly->points[0]->x;
    return fabs(x1 - x2);
}

static
double
sph_ifs_spectra_positions_get_scale(cpl_image* specid_image_obs) {
	sph_fft* sphfft   = NULL;
	cpl_image*		im2 = NULL;
	cpl_image*		imcut = NULL;
	cpl_image* 		im = NULL;
	double          offx_sigma = 0.0;
	int				yy = 0;
	double offx = 0.0;

    im2 = cpl_image_cast( specid_image_obs, CPL_TYPE_DOUBLE);
    if ( cpl_image_get_size_x(im2) > 500 + 1024 ) {
        imcut = cpl_image_extract(im2, 500, 500, 500 + 1024, 500 + 1024);
    }
    else {
        imcut = cpl_image_duplicate(im2);
    }
    cpl_image_delete( im2 );
        im2 = cpl_image_collapse_create(imcut,0);
    cpl_image_save(im2,"input_image_row.fits", CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    if ( im2 == NULL ) return CPL_ERROR_NULL_INPUT;
    im = cpl_image_new( cpl_image_get_size_x(imcut), cpl_image_get_size_y(imcut), CPL_TYPE_DOUBLE );
    if ( im == NULL ) return CPL_ERROR_NULL_INPUT;
	for (yy = 0; yy < cpl_image_get_size_y(imcut); ++yy) {
		cpl_image_copy(im,im2,1,yy + 1);
		if ( cpl_error_get_code()) {
			SPH_RAISE_CPL
			exit(1);
		}
	}
    cpl_image_save(im,"input_image2.fits", CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);

    sphfft = sph_fft_new( SPH_FFT_GSL_MIXEDRADIX );
    offx = sph_fft_get_scale_vector( sphfft, im, &offx_sigma );
    cpl_image_delete(im2);
    cpl_image_delete(imcut);
    cpl_image_delete(im);
    sph_fft_delete(sphfft);
    return offx;
}


static
sph_polynomial_distortion_model*
sph_ifs_spectra_positions_create_distmap(
        sph_point_pattern* pp2,
        sph_point_pattern* pp,
        double length_threshold,
        int polyfit_order )
{
    sph_polynomial_distortion_model*    result = NULL;
    sph_distortion_map*                    dmap = NULL;
    sph_distortion_map*                    ddmap = NULL;
    int                                    nremoved = 0;
    if ( !pp2 ) return NULL;

    //Do NOT centralise since this can cause problem if some points are missing in pp2.
    //sph_point_pattern_centralise(pp2);
    //sph_point_pattern_centralise(pp);
    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "The number of points are: %d,%d",
            sph_point_pattern_get_size(pp),
            sph_point_pattern_get_size(pp2) );
    dmap = sph_distortion_map_new(pp2,pp);
    if ( !dmap ) {
        sph_point_pattern_delete(pp2);pp2=NULL;
        return NULL;
    }

    //sph_distortion_map_remove_offset(dmap);

    nremoved = sph_distortion_map_remove_above_length(dmap,length_threshold);
    sph_error_raise(SPH_ERROR_GENERAL,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO,
            "Removed %d vectors above distortion %f "
            "from distortion map from a total of %d "
            "initally.",
            nremoved, length_threshold, sph_distortion_map_get_size(dmap)+nremoved);
    result = sph_polynomial_distortion_model_new(NULL,NULL);
    if ( !result ) {
        sph_distortion_map_delete(dmap);dmap = NULL;
        return NULL;
    }

    ddmap = sph_polynomial_distortion_model_fit_distortion_map(result,dmap,polyfit_order);
    sph_distortion_map_delete(ddmap); ddmap = NULL;
    sph_distortion_map_delete(dmap); dmap = NULL;
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ifs_spectra_positions_run Run the Instrument Flat Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * instrument flat recipe for IFS.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ifs_spectra_positions.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list
 @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static
sph_pixel_description_table*
sph_ifs_spectra_positions_process_prism( sph_ifs_spectra_positions* self,
        cpl_frameset* inframes,
        sph_master_frame* flat,
        sph_master_frame* dark,
        sph_ifs_lenslet_model* lenslet_model,
        cpl_propertylist* pl)
{
    int                         rerr                = CPL_ERROR_NONE;
    cpl_error_code              prestate            = CPL_ERROR_NONE;
    sph_master_frame*            combined            = NULL;
    sph_pixel_description_table*    predicted_pdt    = NULL;
    cpl_image*                    centres    = NULL;
    cpl_image*                    centres_obs    = NULL;
    cpl_propertylist*             ipl = NULL;
    sph_point_pattern*            point_pattern = NULL;
    sph_point_pattern*            point_pattern_model = NULL;
    cpl_image*                    specid_image_obs = NULL;
    cpl_mask*                    mask = NULL;
    cpl_size                            nobs = 0;
    double                        median = 0.0;
    double                        dev = 0.0;
    double                        scale = 0.0;
    double                        offx = 0.0;
    double                        offy = 0.0;
    double                        doffx = 0.0;
    double                        doffy = 0.0;
    int                         pixsin = 0;
    int                         pixsout = 0;
    int                         darkin = 0;
    int                         xx = 0;
    int                         yy = 0;
    int                         bp = 0;
    sph_polynomial_distortion_model* poldist = NULL;
    if ( !lenslet_model ) {
        sph_error_raise( CPL_ERROR_NULL_INPUT,
                __FILE__,
                __func__,
                __LINE__,
                SPH_ERROR_ERROR,
                "I need a valid IFS lenslet model to "
                "produce the spectra positions table. "
                "Please provide one with the %s tag",
                SPH_IFS_TAG_LENSLET_MODEL_CALIB);
        return NULL;
    }
    prestate = cpl_error_get_code();
    cpl_error_reset();

    combined = sph_framecombination_master_frame_from_cpl_frameset( inframes,
            self->coll_alg, self->framecomb_parameterlist );

    // recover the DIT from the inframes to allow for later scaling
    ipl = sph_keyword_manager_load_properties( cpl_frame_get_filename(
    		cpl_frameset_get_first(inframes)),0);

    if (ipl != NULL && cpl_propertylist_has(ipl, SPH_COMMON_KEYWORD_DIT))
    {
    	cpl_propertylist_append_property(combined->properties,cpl_propertylist_get_property_const(ipl,SPH_COMMON_KEYWORD_DIT));
        cpl_propertylist_delete(ipl); ipl=NULL;
    }
    else
    	SPH_WARNING("Cannot find DIT in raw data, will subtract unscaled dark subsequently!");

    if ( dark ) {
        sph_ifs_proc_apply_master_dark(combined,dark);
    }
    if ( flat )
    {
        sph_master_frame_divide_master_frame( combined, flat );
    }

    median = cpl_image_get_median_dev(combined->image,&dev);

    if ( self->threshold < 0.0 ) {
        self->threshold = median + 0.1 * dev;
    }
    sph_error_raise( SPH_ERROR_INFO,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO,
            "Am attemtping to detect regions on image "
            "with median %f and average difference to "
            "median of %f using a threshold of %f",
            median,dev,
            self->threshold);
    cpl_propertylist_append_double(pl, SPH_IFS_KEYWORD_SPECPOS_QC_THRESHOLD, self->threshold );
#ifdef SPHERE_DEBUG
    cpl_image_save( combined->image, "combined.fits",
            CPL_TYPE_DOUBLE,
            NULL,
            CPL_IO_DEFAULT);
#endif
    mask = cpl_mask_threshold_image_create(combined->image, self->threshold, 1000000.0 );
    specid_image_obs = cpl_image_labelise_mask_create(mask,&nobs);
    cpl_mask_delete( mask ); mask = NULL;

    SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO,"Detected %d spectra",(int)nobs );

    cpl_propertylist_append_int(pl, SPH_IFS_KEYWORD_SPECPOS_QC_NREGS, nobs );
    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Creating point pattern...");
#ifdef SPHERE_DEBUG
    cpl_image_save( specid_image_obs, "detected_spectra_regions.fits",
            CPL_TYPE_DOUBLE,
            NULL,
            CPL_IO_DEFAULT);
#endif
    point_pattern = sph_ifs_spectra_positions_extract_pp( specid_image_obs, nobs, 20 );

    if ( point_pattern ) {
        centres_obs = sph_point_pattern_create_image( point_pattern,
                cpl_image_get_size_x(combined->image),
                cpl_image_get_size_y(combined->image),
                1.0);
#ifdef SPHERE_DEBUG
        cpl_image_save( centres_obs, "observed_point_pattern.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_DEFAULT);
#endif
        scale = sph_ifs_spectra_positions_get_scale(specid_image_obs);
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
        sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_INFO,
                "Determined a scale of %f.", scale );
        scale = scale / sph_ifs_spectra_positions_get_scale_model(lenslet_model);
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;

        cpl_propertylist_append_double(pl, SPH_IFS_KEYWORD_SPECPOS_QC_SCALE, scale );

        if ( scale > 0.0 ) {
            sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_INFO,
                    "Determined a scale of %f.", scale );
            lenslet_model->stretch_x *= scale;
            lenslet_model->stretch_y *= scale;
            lenslet_model->specwidth_pixels *= scale;
            lenslet_model->specwidth_microns *= scale;
            // Below commented out, since IFS request is that spectra alwyays have
            // same length
            //lenslet_model->speclength_pixels *= scale;
            //lenslet_model->speclength_microns *= scale;
        }
        else {
            sph_error_raise( SPH_ERROR_GENERAL, __FILE__, __func__,__LINE__,SPH_ERROR_ERROR,"Problem determining scale");
            goto EXIT;
        }
        if ( lenslet_model->specwidth_pixels < IFS_MIN_SPECWIDTH ) {
            SPH_ERROR_RAISE_ERR( CPL_ERROR_ILLEGAL_INPUT,
                "Determined a avg. spectra width of %f, which is "
                "below the tolerance of %f.",
                lenslet_model->specwidth_pixels ,
                IFS_MIN_SPECWIDTH);
                goto EXIT;
        }
        offx = offy = 0.0;
        doffx = 2.0; doffy = 40.0;
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;


        point_pattern_model = sph_ifs_lenslet_model_predict_spec_loc_pp( lenslet_model, 0.0,0.0 );
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
        sph_ifs_spectra_positions_determine_offset(
                point_pattern_model,
                point_pattern,
                500,
                &offx,
                &offy,
                lenslet_model->detsize_pixels / 2);
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
        sph_point_pattern_delete(point_pattern_model); point_pattern_model = NULL;
        SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;

        if ( rerr == CPL_ERROR_NONE ) {
            sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_INFO,
                    "Determined an offset of %f in x and %f in y.", offx, offy );
            lenslet_model->zero_offsetx += offx;
            lenslet_model->zero_offsety += offy;
        }
        else {
            sph_error_raise( SPH_ERROR_GENERAL, __FILE__, __func__,__LINE__,SPH_ERROR_ERROR,"Problem determining offset");
            goto EXIT;
        }

        cpl_propertylist_append_double(pl, SPH_IFS_KEYWORD_SPECPOS_QC_OFFSETX, offx );
        cpl_propertylist_append_double(pl, SPH_IFS_KEYWORD_SPECPOS_QC_OFFSETY, offy );
        if ( self->dist ) {
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Determining distortion...");
            point_pattern_model = sph_ifs_lenslet_model_predict_spec_loc_pp( lenslet_model, 0.0,0.0 );
            poldist = sph_ifs_spectra_positions_create_distmap(point_pattern,point_pattern_model,10.0,4);
            if ( poldist ) {
                sph_polynomial_distortion_model_save(poldist,"spectra_positions_distortion.fits",2048,2048,CPL_IO_CREATE,NULL);
                sph_ifs_lenslet_model_set_detector_distortion(lenslet_model,poldist);
            }
            else SPH_ERROR_RAISE_WARNING(SPH_ERROR_GENERAL,
                    "Could not determine distortion. Assuming zero distortion.");
            sph_point_pattern_delete(point_pattern_model); point_pattern_model = NULL;
        }
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Creating new final PDT...");
        predicted_pdt = sph_pixel_description_table_new_from_model( lenslet_model , 0.0, 0.0);
        mask = sph_pixel_description_table_get_mask(predicted_pdt);
        cpl_image_threshold(specid_image_obs,0.0,0.9,0.0,1.0);

        for (xx = 0; xx < cpl_image_get_size_x(specid_image_obs); ++xx) {
            for (yy = 0; yy < cpl_image_get_size_y(specid_image_obs); ++yy) {
                if ( cpl_image_get(specid_image_obs, xx + 1, yy + 1,&bp) > 0 ) {
                    if ( cpl_mask_get(mask, xx + 1, yy + 1) > 0 ) {
                        pixsin++;
                    }
                    else {
                        pixsout++;
                    }
                }
                else {
                    if ( cpl_mask_get(mask, xx + 1, yy + 1) > 0 ) {
                        darkin++;
                    }
                }
            }
        }

        cpl_propertylist_append_int(pl, SPH_IFS_KEYWORD_SPECPOS_QC_PIXS_IN, pixsin);
        cpl_propertylist_append_int(pl, SPH_IFS_KEYWORD_SPECPOS_QC_PIXS_OUT, pixsout);
        cpl_propertylist_append_int(pl, SPH_IFS_KEYWORD_SPECPOS_QC_DARK_PIXS_IN, darkin );

        cpl_mask_delete(mask); mask = NULL;
        sph_point_pattern_delete(point_pattern); point_pattern = NULL;
        SPH_RAISE_CPL_RESET;
    }
    else {
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__,__func__,__LINE__,
                SPH_ERROR_ERROR,
                "Could not create the point pattern"
                " of model.");
    }
    sph_master_frame_delete(combined);combined = NULL;

    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
    	SPH_RAISE_CPL
    }
    cpl_image_delete( centres_obs ); centres_obs = NULL;
    cpl_image_delete( centres ); centres = NULL;
    cpl_image_delete(specid_image_obs); specid_image_obs = NULL;
    cpl_propertylist_delete(ipl); ipl = NULL;
    return predicted_pdt;
EXIT:    
    sph_master_frame_delete(combined);combined = NULL;
    sph_point_pattern_delete(point_pattern); point_pattern = NULL;
    sph_point_pattern_delete(point_pattern_model); point_pattern_model = NULL;
    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
    	SPH_RAISE_CPL
    }
    cpl_image_delete( centres_obs ); centres_obs = NULL;
    cpl_image_delete( centres ); centres = NULL;
    cpl_image_delete(specid_image_obs); specid_image_obs = NULL;
    cpl_propertylist_delete(ipl); ipl = NULL;
    
    sph_pixel_description_table_delete(predicted_pdt);
    return NULL;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list
 @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ifs_spectra_positions_run( sph_ifs_spectra_positions* self )
{
    sph_error_code rerr = CPL_ERROR_NONE;
    sph_master_frame* dark = NULL;
    sph_master_frame* flat = NULL;
    sph_ifs_lenslet_model* model = NULL;
    sph_pixel_description_table* predicted_pdt = NULL;
    cpl_propertylist*           pl = NULL;
    if ( self->master_dark_frame ) {
        dark = sph_master_frame_load( cpl_frame_get_filename(self->master_dark_frame),0 );
    }

    if ( self->master_flat_frame ) {
        flat = sph_master_frame_load( cpl_frame_get_filename(self->master_dark_frame),0 );
    }
    if ( self->model_frame ) {
        model = sph_ifs_lenslet_model_load( cpl_frame_get_filename(self->model_frame) );
    }
    else {
        model = sph_ifs_lenslet_model_new();
        if ( self->hmode ) {
            // These two lines were active BEFORE the Sept. 12 IFS tests
            // They may have been correct, but baseline now is that
            // speclength and specwidth are mode independent
            //model->speclength_pixels = 39.4951734171714;
            //model->specwidth_pixels =  4.9293975935862;
            model->maxlambda = SPH_IFS_LENSLET_MODEL_MAX_LAMBDA_JH;
            model->minlambda = SPH_IFS_LENSLET_MODEL_MIN_LAMBDA_JH;
            model->dispersion = ( model->maxlambda - model->minlambda ) / ( model->speclength_pixels - 1 );
        }
    }

    if ( self->angle < -360.0 ) {
        model->rotangle = 0.0;
        self->angle = sph_ifs_spectra_positions_get_angle_model(model);
    }
    model->rotangle = self->angle;
    pl = cpl_propertylist_new();

    predicted_pdt = sph_ifs_spectra_positions_process_prism( self,
            self->rawframes,flat, dark, model,
            pl);

    if ( predicted_pdt ) {
        if ( sph_pixel_description_table_save_dfs(
                predicted_pdt,
                self->spectra_positions_filename,
                model, self->inframes, NULL, self->inparams,
                SPH_IFS_TAG_SPEC_POS_CALIB,
                SPH_RECIPE_NAME_IFS_SPECTRA_POSITIONS,
                SPH_PIPELINE_NAME_IFS, pl) != CPL_ERROR_NONE ) {
            SPH_ERROR_RAISE_ERR(CPL_ERROR_FILE_IO, "Could not write main product.");
        }
    }
    sph_master_frame_delete(flat); flat = NULL;
    sph_master_frame_delete(dark); dark = NULL;
    sph_ifs_lenslet_model_delete(model); model = NULL;
    sph_pixel_description_table_delete(predicted_pdt); predicted_pdt = NULL;
    cpl_propertylist_delete(pl); pl = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/**@}*/
