/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"

#include "sph_ird_tags.h"
#include "sph_ird_gain.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_filemanager.h"
#include "sph_ird_instrument_model.h"
#include "sph_ird_star_center.h"
#include "sph_keyword_manager.h"
#include "sph_fft.h"
#include "sph_fits.h"
#include "sph_smart_imagelist.h"
#include "sph_common_science.h"
#include "sph_ird_common_science.h"
#include "sph_differential_imaging.h"
#include "sph_time.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include "sph_fctable.h"
#include "sph_andromeda_support.h"
#include "sph_cube.h"

const char* const SPH_IRD_STAR_CENTER_TIME_UT_NAME = "TIME-UT";
const char* const SPH_IRD_STAR_CENTER_CENTLEFTX_NAME = "CENTRE LEFT X";
const char* const SPH_IRD_STAR_CENTER_CENTLEFTY_NAME = "CENTRE LEFT Y";
const char* const SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME = "CENTRE RIGHT X";
const char* const SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME = "CENTRE RIGHT Y";
const char* const SPH_IRD_STAR_CENTER_DMS_POS_X_NAME = "DMS POS X";
const char* const SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME = "DMS POS Y";


static
sph_ird_common_science*
__sph_ird_common_science_new(void);
static
sph_error_code
sph_ird_common_science_interpolate_centers(
        cpl_frame* fcframe,
        cpl_vector* timev,
        int plane,
        double* clx,
        double* cly,
        double* crx,
        double* cry);
static
sph_fctable*
sph_ird_common_science_create_table_single_cube(
        cpl_frame* fcframe,
        cpl_frame* arawframe);
static
sph_error_code
__sph_ird_common_science_processing_iteration(
        sph_ird_common_science* sci,
        cpl_frameset* loopframes,
        cpl_frameset* scframes,
        cpl_frameset** left_out_frames,
        cpl_frameset** right_out_frames,
        cpl_frameset** total_out_frames,
        int ff,
        int xcl,
        int ycl,
        int xcr,
        int ycr);

static
sph_error_code
sph_ird_common_science_process_single_cube(
        sph_ird_common_science* sci,
        cpl_frame* frame_of_raw_cube,
        cpl_vector* centx_left,
        cpl_vector* centy_left,
        cpl_vector* centx_right,
        cpl_vector* centy_right,
        sph_cube* left_cube,
        sph_cube* right_cube,
        sph_cube* total_cube);
/*----------------------------------------------------------------------------*/
/**
 * @brief Create new sph_common_science structure
 *
 * @return pointer to new sph_common_science structure
 */
/*----------------------------------------------------------------------------*/
sph_ird_common_science*
sph_ird_common_science_new(
        const cpl_frameset* inframes,
        const cpl_parameterlist* inparams,
        cpl_frame*    dark,
        cpl_frame*    flat,
        cpl_frame*    bpix,
        cpl_frame*    distortion,
        cpl_frame* filter,
        cpl_frame*    irdmodel,
        const char* recipe,
        const char* tag
        )
{
    sph_ird_common_science* result = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    result = __sph_ird_common_science_new();
    result->inframes = (cpl_frameset*)inframes;
    result->inparams = (cpl_parameterlist*)inparams;
    if ( dark ) {
        result->dark = sph_master_frame_load(cpl_frame_get_filename(dark),0);
    }
    if ( flat ) {
        result->flat = sph_master_frame_load(cpl_frame_get_filename(flat),0);
    }
    if ( bpix ) {
        result->bpix = cpl_image_load(cpl_frame_get_filename(bpix),CPL_TYPE_INT,0,0);
    }

    if ( distortion ) {
        result->distort_left = sph_polynomial_distortion_model_load(
                cpl_frame_get_filename(distortion),0);
        result->distort_right = sph_polynomial_distortion_model_load(
                cpl_frame_get_filename(distortion),4);
    }
    if ( filter ) {
        result->filter_table = sph_filter_table_load_ascii(
                cpl_frame_get_filename(filter));
    }
    if ( irdmodel ) result->irdmodel = sph_ird_instrument_model_load(
            cpl_frame_get_filename(irdmodel));
    else {
        if ( flat ) result->irdmodel = sph_ird_instrument_model_load(
                cpl_frame_get_filename(flat));
        else result->irdmodel = sph_ird_instrument_model_new();
    }
    result->transform = sph_transform_new_default();
    result->recipe = recipe;
    result->tag = tag;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Set the telescope flat field
 * @param self          pointer to self
 * @param tff_frame     frame of the telescope flat or NULL
 *
 * @return error code
 *
 * Set the telescope flat field to use.
 * If this is non NULL, the telescope flat field is read
 * from the frame and used in the processing.
 *
 * If NULL is passed the telescope flat is unset.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_ird_common_science_set_tff(
        sph_ird_common_science* self,
        cpl_frame* tff)
{
    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT );

    sph_master_frame_delete(self->tff); self->tff = NULL;

    if ( tff ) {
        self->tff = sph_master_frame_load( cpl_frame_get_filename( tff ), 0 );
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Set the transform method
 * @param self          the IRDIS common science
 * @param transform     the transform to set for future operations
 *
 * @return error code
 *
 * This sets the new transform method. This new transform will be used
 * for all future calls to ird common science functionality.
 * The transform structure will be managed by the ird common science
 * module. Do not call delete on the transform structure passed after
 * set_transform has been called.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_ird_common_science_set_transform(
        sph_ird_common_science* self,
        sph_transform* transform)
{
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(transform,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    if ( self->transform ) {
        sph_transform_delete(self->transform);
        self->transform = NULL;
    }
    self->transform = transform;

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Delete sph_common_science structure
 * @param self   pointer to structure to delete
 */
/*----------------------------------------------------------------------------*/
void
sph_ird_common_science_delete(sph_ird_common_science* self) {

    if ( self ) {
        sph_master_frame_delete(self->dark);
        sph_master_frame_delete(self->flat);
        cpl_image_delete(self->bpix);
        sph_polynomial_distortion_model_delete(self->distort_left);
        sph_polynomial_distortion_model_delete(self->distort_right);
        sph_filter_table_delete(self->filter_table);
        sph_ird_instrument_model_delete(self->irdmodel);
        self->irdmodel = NULL;
        sph_transform_delete(self->transform); self->transform = NULL;
        if ( self->tff) sph_master_frame_delete(self->tff);
        self->tff = NULL;
        cpl_free(self);
    }
}


/*----------------------------------------------------------------------------*/
/**
 * @brief Process a set of raw input cubes
 * @param inframes      the frameset containing the IRDIS raw cubes
 * @param scframes      the frameset containing the star center tables (optional)
 * @param dark          the master dark (optional)
 * @param flat          the master flat (optional)
 * @param filer_frame   the frame of the filter table (optional)
 * @param distortion    the distortion frame (optional)
 * @param inparams      the input parameterlist
 * @param minr          the minimum annulus radius
 * @param maxr          the maximim annulus radius
 * @param filter_radius the size of the filter kernel
 * @param rotate        flag to set rotation on or off
 * @param sdi           flag to set SDI (scaling+difference imaging) on or off
 * @param tag           the tag to give to resulting frames
 * @param recipe        the name of the calling recipe
 * @param left_out      output frameset of processed left FOV (optional)
 * @param right_out      output frameset of processed right FOV (optional)
 * @param total_out      output frames of processed sum or difference (optional)
 *
 * @return error code
 *
 * This function processes a set of input cubes. Most inputs are optional.
 * Depending on whether SDI is selected or not, a resulting total frameset is
 * either the sum of left and right FOV divided by 2 or the difference of
 * right FOV and left FOV (when SDI is on).
 * Note that if SDI is off, frames will NOT be rescaled even if the filter in the
 * filter list has different wavelengths for left and right FOV.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_ird_common_science_process_cubes(
        sph_ird_common_science* sci,
        cpl_frameset* rawframes,
        cpl_frameset* scframes,
        cpl_frameset** left_out,
        cpl_frameset** right_out,
        cpl_frameset** total_out
        )
{
    sph_master_frame*   mdark        = NULL;
    sph_master_frame*   mflat        = NULL;
    sph_polynomial_distortion_model*    distort_left = NULL;
    sph_polynomial_distortion_model*    distort_right = NULL;
    cpl_frame*              aframe = NULL;
    cpl_propertylist*       plist = NULL;
    double                  xcl = 0.0;
    double                  xcr = 0.0;
    double                  ycl = 0.0;
    double                  ycr = 0.0;
    int                     ff      = 0;
    char*                       fname_left  = NULL;
    char*                       fname_right  = NULL;
    char*                       fname_total  = NULL;
    cpl_frame*                  cube_frame_left = NULL;
    cpl_frame*               cube_frame_right = NULL;
    cpl_frame*               cube_frame_total = NULL;
    cpl_frameset*           loopframes = NULL;
    char*                   basename = NULL;
    sph_error_code          rerr = CPL_ERROR_NONE;
    sph_error_code          rerrtmp = CPL_ERROR_NONE;

    cpl_ensure_code(sci, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(sci->rotate_flag == 0, CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(sci->inframes, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(sci->inparams, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(rawframes, CPL_ERROR_NULL_INPUT);

    if ( scframes ) {
        if ( cpl_frameset_get_size(rawframes) !=
                cpl_frameset_get_size(scframes) ) {
            SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                    "The number of frames in the star center frameset "
                    "is not the same "
                    "as the number of raw files. "
                    "Please ensure that the .sof file "
                    "specifies exactly as many star center frames as raw frames.");
            return cpl_error_get_code();
        }
    }
    else {
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "No star center table provided. "
                "Am assuming dithering is not used.");
    }

    sph_ird_instrument_model_get_centre_left_subframe_coords(
            sci->irdmodel,
            &xcl,&ycl);
    sph_ird_instrument_model_get_centre_right_subframe_coords(
            sci->irdmodel,
            &xcr,&ycr);
    if ( left_out ) *left_out = cpl_frameset_new();
    if ( right_out ) *right_out = cpl_frameset_new();
    if ( total_out ) *total_out = cpl_frameset_new();
    SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;

    for (ff = 0; ff < cpl_frameset_get_size(rawframes); ++ff) {
        aframe = cpl_frameset_get_frame(rawframes,ff);
        basename = sph_filemanager_remove_dir(
                cpl_frame_get_filename(aframe));
        fname_left = sph_filemanager_new_filename_from_base(
                basename,"left");
        fname_right = sph_filemanager_new_filename_from_base(
                basename,"right");
        fname_total = sph_filemanager_new_filename_from_base(
                basename,"total");
        cube_frame_left = cpl_frame_new();
        cube_frame_right = cpl_frame_new();
        cube_frame_total = cpl_frame_new();
        cpl_frame_set_filename(cube_frame_left,fname_left);
        cpl_frame_set_filename(cube_frame_right,fname_right);
        cpl_frame_set_filename(cube_frame_total,fname_total);
        cpl_frame_set_tag(cube_frame_left,sci->tag);
        cpl_frame_set_tag(cube_frame_right,sci->tag);
        cpl_frame_set_tag(cube_frame_total,sci->tag);
        cpl_frame_set_group(cube_frame_left,CPL_FRAME_GROUP_CALIB);
        cpl_frame_set_group(cube_frame_right,CPL_FRAME_GROUP_CALIB);
        cpl_frame_set_group(cube_frame_total,CPL_FRAME_GROUP_CALIB);
        if ( left_out ) cpl_frameset_insert(*left_out,cube_frame_left);
        else { cpl_frame_delete(cube_frame_left); cube_frame_left = NULL;}
        if ( right_out) cpl_frameset_insert(*right_out,cube_frame_right);
        else { cpl_frame_delete(cube_frame_right); cube_frame_right = NULL;}
        if ( total_out) cpl_frameset_insert(*total_out,cube_frame_total);
        else { cpl_frame_delete(cube_frame_total); cube_frame_total = NULL;}
        cpl_free(fname_left); fname_left = NULL;
        cpl_free(fname_right); fname_right = NULL;
        cpl_free(fname_total); fname_total = NULL;
        cpl_free(basename); basename = NULL;
    }

    // Need to make duplicate since DFS product frames are added to inframes
    // but those should not to be looped over.
    loopframes = cpl_frameset_duplicate(rawframes);
#ifdef _OPENMP
#pragma omp parallel for firstprivate(rerr,rerrtmp), lastprivate(rerr)
#endif
    for (ff = 0; ff < cpl_frameset_get_size(loopframes); ++ff) {
        if ( (rerrtmp = __sph_ird_common_science_processing_iteration(
                sci, loopframes, scframes,
                left_out,
                right_out,
                total_out,
                ff,
                xcl,ycl,xcr,ycr)) != CPL_ERROR_NONE) {
            rerr = rerrtmp;
        }
    }
    if ( rerr == CPL_ERROR_DATA_NOT_FOUND ) {
        SPH_ERROR_RAISE_ERR(SPH_ERROR_GENERAL,
                "Could not find matching fctable "
                "for some input frames."
                "Please double check the input.");
    }
    SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
EXIT:
    cpl_frameset_delete(loopframes); loopframes = NULL;
    sph_master_frame_delete(mdark); mdark = NULL;
    sph_master_frame_delete(mflat); mflat = NULL;
    sph_polynomial_distortion_model_delete(distort_left); distort_left = NULL;
    sph_polynomial_distortion_model_delete(distort_right); distort_right = NULL;
    cpl_propertylist_delete(plist); plist = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/
static
sph_ird_common_science*
__sph_ird_common_science_new(void) {
    sph_ird_common_science* result = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    result = cpl_calloc(1,sizeof(sph_ird_common_science));
    result->least_square = 1;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}


static
sph_error_code
sph_ird_common_science_process_single_cube(
        sph_ird_common_science* sci,
        cpl_frame* frame_of_raw_cube,
        cpl_vector* cx_left,
        cpl_vector* cy_left,
        cpl_vector* cx_right,
        cpl_vector* cy_right,
        sph_cube* left_cube,
        sph_cube* right_cube,
        sph_cube* total_cube)
{
    cpl_vector*         posind = NULL;
    cpl_vector*         negind = NULL;
    cpl_vector*         normalisations = NULL;
    cpl_image*          rawimage = NULL;
    sph_master_frame*          calibmframe = NULL;
    sph_master_frame*          leftmframe = NULL;
    sph_master_frame*          rightmframe = NULL;
    sph_master_frame*          tmppointer = NULL;
    double              scale_left = 1.0;
    double              scale_right = 1.0;
    double              lambda_left = 1.0;
    double              lambda_right = 1.0;
    cpl_imagelist*      diffimlist = NULL;
    sph_error_code      rerr = CPL_ERROR_NONE;
    const char*         filtername = NULL;
    int                 pp = 0 ;
    cpl_vector*         centx_left = NULL;
    cpl_vector*         centy_left = NULL;
    cpl_vector*         centx_right = NULL;
    cpl_vector*         centy_right = NULL;
	int nplanes = 0;
    cpl_ensure_code(frame_of_raw_cube,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(sci,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cx_left,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cy_left,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cx_right,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cy_right,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cpl_vector_get_size(cx_left) ==
            cpl_vector_get_size(cy_left), CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(cpl_vector_get_size(cx_left) ==
            cpl_vector_get_size(cx_right), CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(cpl_vector_get_size(cx_left) ==
            cpl_vector_get_size(cy_right), CPL_ERROR_ILLEGAL_INPUT);
    nplanes = sph_fits_get_nplanes(
            cpl_frame_get_filename( frame_of_raw_cube ), 0 );
    if ( nplanes != cpl_vector_get_size(cx_left)) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT, "The number of entries in the "
                "field centre table is not equal to the number of planes in the "
                "raw frames. Please check the field centre table, if needed, "
                "recreate one using the sph_ird_star_center recipe.");
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    if ( sci->flag_sdi ) {
        posind = cpl_vector_new(1);
        negind = cpl_vector_new(1);
        cpl_vector_set(posind,0,0);
        cpl_vector_set(negind,0,1);
    }

    sph_ird_instrument_model_adjust_centre_coords_v(
            sci->irdmodel,
            cx_left,cy_left,
            &centx_left,&centy_left);
    sph_ird_instrument_model_adjust_centre_coords_v(
            sci->irdmodel,
            cx_right,cy_right,
            &centx_right,&centy_right);

    for (pp = 0; pp < nplanes; ++pp)
    {
        rawimage = cpl_image_load(cpl_frame_get_filename(frame_of_raw_cube), CPL_TYPE_DOUBLE, pp, 0);
        if ( rawimage )
        {
            calibmframe = sph_common_science_calibrate_raw(rawimage,sci->dark,sci->flat);
            if ( calibmframe )
            {
                cpl_image_delete(rawimage); rawimage = NULL;
                if ( sci->bpix )
                    sph_master_frame_set_bads_from_image(calibmframe,sci->bpix);
                leftmframe =
                        sph_ird_instrument_model_extract_left_master_frame(sci->irdmodel,
                        calibmframe);
                rightmframe =
                        sph_ird_instrument_model_extract_right_master_frame(sci->irdmodel,
                        calibmframe);
                if ( sci->filter_table && sci->flag_sdi ) {
                    if ( cpl_propertylist_has(leftmframe->properties,
                            SPH_IRD_KEYWORD_COMMON_FILTER_NAME) ) {
                        filtername = cpl_propertylist_get_string(leftmframe->properties,
                                SPH_IRD_KEYWORD_COMMON_FILTER_NAME);
                    }
                    else {
                        filtername = sci->filter_table->filterNames[0];
                        SPH_ERROR_RAISE_WARNING(CPL_ERROR_ILLEGAL_INPUT,
                                "Could not find correct entry in the filter table,"
                                "so I am using the default filter (first line in "
                                "filter table).");
                    }
                    lambda_left = sph_filter_table_get_left_lambda(sci->filter_table,filtername);
                    lambda_right = sph_filter_table_get_right_lambda(sci->filter_table,filtername);
                    if ( lambda_left < lambda_right ) {
                        scale_left = 1.0;
                        scale_right = lambda_left / lambda_right;
                    }
                    else {
                        scale_right = 1.0;
                        scale_left = lambda_right / lambda_left;
                    }
                    SPH_ERROR_RAISE_INFO(CPL_ERROR_ILLEGAL_INPUT,
                                "Sclaing by (left,right): %f,%f", scale_left,scale_right);


                }
                else {
                    scale_left = 1.0;
                    scale_right = 1.0;
                    if ( sci->flag_sdi ) {
                        SPH_ERROR_RAISE_WARNING(CPL_ERROR_ILLEGAL_INPUT,
                                "Am not rescaling since no filter table could be read.");
                    }
                }
                rerr = sph_transform_apply(sci->transform,
                        leftmframe,
                        sci->distort_left,
                        cpl_vector_get(centx_left,pp),
                        cpl_vector_get(centy_left,pp),
                        0.0,scale_left,sci->irdmodel);
                SPH_ERROR_ENSURE_GOTO_EXIT(rerr == CPL_ERROR_NONE,rerr);
                rerr = sph_transform_apply(sci->transform,
                        rightmframe,
                        sci->distort_right,
                        cpl_vector_get(centx_right,pp),
                        cpl_vector_get(centy_right,pp),
                        0.0,scale_right,sci->irdmodel);
                SPH_ERROR_ENSURE_GOTO_EXIT(rerr == CPL_ERROR_NONE,rerr);
                if ( sci->tff ) {
                    sph_master_frame_divide_master_frame(
                            leftmframe,
                            sci->tff);
                }
                if ( sci->tff ) {
                    sph_master_frame_divide_master_frame(
                            rightmframe,
                            sci->tff);
                }

                if ( left_cube ) {
                    SPH_ERROR_ENSURE_GOTO_EXIT(
                            sph_cube_append_master(
                                    left_cube,
                                    leftmframe,
                                    pp) == CPL_ERROR_NONE,
                            CPL_ERROR_FILE_IO);
                }
                if ( right_cube ) {
                    SPH_ERROR_ENSURE_GOTO_EXIT(
                            sph_cube_append_master(
                                    right_cube,
                                    rightmframe,
                                    pp) == CPL_ERROR_NONE,
                            CPL_ERROR_FILE_IO);
                }
                if ( total_cube ) {
                    if ( sci->flag_sdi ) {
                        diffimlist = cpl_imagelist_new();
                        cpl_imagelist_set(diffimlist,
                                sph_master_frame_get_image(leftmframe,1),0);
                        cpl_imagelist_set(diffimlist,
                                sph_master_frame_get_image(rightmframe,1),1);
                        lambda_left = sph_filter_table_get_left_lambda(sci->filter_table,filtername);
                        lambda_right = sph_filter_table_get_right_lambda(sci->filter_table,filtername);
                        if ( lambda_left < lambda_right ) {
                            cpl_vector_set(posind,0,0);
                            cpl_vector_set(negind,0,1);
                        }
                        else {
                            cpl_vector_set(posind,0,1);
                            cpl_vector_set(negind,0,0);
                        }
                        normalisations =
                                sph_andromeda_support_calc_normalisations(
                                        diffimlist,posind,negind,
                                        sci->minr,sci->maxr,sci->least_square);
                        cpl_imagelist_delete(diffimlist); diffimlist = NULL;
                        if ( lambda_left < lambda_right ) { // left frame is short wavelength
                            sph_master_frame_divide_double(rightmframe,
                                    1.0/cpl_vector_get(normalisations,0));
                            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                                    "Applied (multiplied) a renormalisation of %f to right frame",cpl_vector_get(normalisations,0));
                            sph_master_frame_subtract_master_frame(
                                    leftmframe,
                                    rightmframe);
                        }
                        else { // right frame is short wavelength
                            sph_master_frame_divide_double(leftmframe,
                                    1.0/cpl_vector_get(normalisations,0));
                            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                                    "Applied (multiplied) a renormalisation of %f to left frame",cpl_vector_get(normalisations,0));
                            sph_master_frame_subtract_master_frame(
                                    rightmframe,
                                    leftmframe);
                            tmppointer = leftmframe;
		            leftmframe = rightmframe;
                            rightmframe = tmppointer;
                            tmppointer = NULL;
                        }
                        cpl_vector_delete(normalisations); normalisations = NULL;
                    }
                    else {
                        sph_master_frame_add_master_frame(leftmframe,rightmframe);
                        sph_master_frame_divide_double(leftmframe,2.0);
                    }
                    SPH_ERROR_ENSURE_GOTO_EXIT(
                            sph_cube_append_master(
                                    total_cube,
                                    leftmframe,
                                    pp) == CPL_ERROR_NONE,
                                    CPL_ERROR_FILE_IO);
                }
                sph_master_frame_delete(calibmframe); calibmframe = NULL;
                sph_master_frame_delete(leftmframe); leftmframe = NULL;
                sph_master_frame_delete(rightmframe); rightmframe = NULL;
                SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
            }
        }
    }

EXIT:
    cpl_vector_delete(posind); posind = NULL;
    cpl_vector_delete(negind); negind = NULL;
    cpl_vector_delete(centx_left); centx_left = NULL;
    cpl_vector_delete(centy_left); centy_left = NULL;
    cpl_vector_delete(centx_right); centx_right = NULL;
    cpl_vector_delete(centy_right); centy_right = NULL;
	sph_master_frame_delete(calibmframe); calibmframe = NULL;
	sph_master_frame_delete(leftmframe); leftmframe = NULL;
	sph_master_frame_delete(rightmframe); rightmframe = NULL;

	SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_ird_common_science_processing_iteration(
        sph_ird_common_science* sci,
        cpl_frameset* loopframes,
        cpl_frameset* scframes,
        cpl_frameset** left_out_frames,
        cpl_frameset** right_out_frames,
        cpl_frameset** total_out_frames,
        int ff,
        int xcl,
        int ycl,
        int xcr,
        int ycr)
{
    cpl_frame*          aframe = NULL;
    sph_fctable*        fctab = NULL;
    cpl_vector*         cent_lx = NULL;
    cpl_vector*         cent_ly = NULL;
    cpl_vector*         cent_rx = NULL;
    cpl_vector*         cent_ry = NULL;
    sph_cube*                   left_cube = NULL;
    sph_cube*                   right_cube = NULL;
    sph_cube*                   total_cube = NULL;


    aframe = cpl_frameset_get_frame(loopframes,ff);
     if ( scframes ) {
         fctab = sph_fctable_find_fctable(aframe,scframes);
         if ( fctab ) {
             cent_lx = sph_fctable_get_centre_x_left(fctab);
             cent_ly = sph_fctable_get_centre_y_left(fctab);
             cent_rx = sph_fctable_get_centre_x_right(fctab);
             cent_ry = sph_fctable_get_centre_y_right(fctab);
         }
         else {
             return CPL_ERROR_DATA_NOT_FOUND;
         }
     }
     else {
         cent_lx = cpl_vector_new(
                 sph_fits_get_nplanes(
                         cpl_frame_get_filename(aframe),
                         0)
                         );
         cent_ly = cpl_vector_duplicate(cent_lx);
         cent_rx = cpl_vector_duplicate(cent_lx);
         cent_ry = cpl_vector_duplicate(cent_lx);

         cpl_vector_fill(cent_lx,xcl);
         cpl_vector_fill(cent_ly,ycl);
         cpl_vector_fill(cent_rx,xcr);
         cpl_vector_fill(cent_ry,ycr);
     }

     if ( left_out_frames ) {
         left_cube = sph_cube_new_dfs(
                 cpl_frame_get_filename(
                         cpl_frameset_get_frame(*left_out_frames,ff)),
                 sci->inframes,aframe,
                 sci->inparams, sci->tag, sci->recipe,
                 SPH_PIPELINE_NAME_IRDIS, NULL);
     }
     if ( right_out_frames ) {
         right_cube = sph_cube_new_dfs(
                 cpl_frame_get_filename(
                         cpl_frameset_get_frame(*right_out_frames,ff)),
                 sci->inframes,
                 aframe, sci->inparams,
                 sci->tag, sci->recipe, SPH_PIPELINE_NAME_IRDIS, NULL);
     }
     if ( total_out_frames ) {
         total_cube = sph_cube_new_dfs(
                 cpl_frame_get_filename(
                         cpl_frameset_get_frame(*total_out_frames,ff)),
                         sci->inframes,
                         aframe, sci->inparams,
                         sci->tag, sci->recipe, SPH_PIPELINE_NAME_IRDIS, NULL);
     }
     sph_ird_common_science_process_single_cube(sci,aframe,
             cent_lx,cent_ly,cent_rx,cent_ry,
             left_cube,
             right_cube,
             total_cube);
     if ( left_cube )
         sph_cube_finalise_file(left_cube->filename);
     if ( right_cube )
         sph_cube_finalise_file(right_cube->filename);
     if ( total_cube )
         sph_cube_finalise_file(total_cube->filename);
     sph_cube_delete(left_cube); left_cube = NULL;
     sph_cube_delete(right_cube); right_cube = NULL;
     sph_cube_delete(total_cube); total_cube = NULL;
     cpl_vector_delete(cent_lx);cent_lx = NULL;
     cpl_vector_delete(cent_ly);cent_ly = NULL;
     cpl_vector_delete(cent_rx);cent_rx = NULL;
     cpl_vector_delete(cent_ry);cent_ry = NULL;
     sph_fctable_delete(fctab); fctab = NULL;
     return CPL_ERROR_NONE;
}


cpl_frameset*
sph_ird_common_science_create_fctables(cpl_frame* fcframe,
		cpl_frameset* rawframes,
		int keep_fctable) {
    int             ff  = 0;
    char*                       tmpfilename = NULL;
    char*                       tmpfilename2 = NULL;
    sph_fctable*    fctable = NULL;
    cpl_propertylist*       plist = NULL;
    cpl_table*          result = NULL;
    cpl_frameset*           resultset = NULL;
    cpl_frame*              aframe = NULL;

    cpl_ensure( rawframes, CPL_ERROR_NULL_INPUT, NULL );

    resultset = cpl_frameset_new();
    plist = cpl_propertylist_new();

    for (ff = 0; ff < cpl_frameset_get_size(rawframes); ++ff)
    {
        fctable =
                sph_ird_common_science_create_table_single_cube(fcframe,
                        cpl_frameset_get_frame(rawframes,ff) );

        if ( fctable )
        {
            tmpfilename2 = sph_fctable_construct_filename(
                    cpl_frameset_get_frame(rawframes,ff),1);

            aframe = cpl_frame_new();
            cpl_frame_set_tag(aframe,SPH_IRD_TAG_FCTABLE_CALIB);
            cpl_frame_set_filename(aframe,tmpfilename2);
            if ( keep_fctable ) sph_filemanager_add_tmp_file(tmpfilename2);
            cpl_frameset_insert(resultset,aframe);
            sph_fctable_save_ascii(fctable,tmpfilename2);
            sph_fctable_delete(fctable); fctable = NULL;
            cpl_table_delete(result); result = NULL;
            cpl_free(tmpfilename); tmpfilename = NULL;
            cpl_free(tmpfilename2); tmpfilename2 = NULL;
        }
        else {
            break;
        }
    }
    cpl_propertylist_delete(plist); plist = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return resultset;
}

static
sph_fctable*
sph_ird_common_science_create_table_single_cube(
        cpl_frame* fcframe,
        cpl_frame* arawframe)
{
    sph_error_code        rerr = CPL_ERROR_NONE;
    sph_fctable*          fctab   = NULL;
    double                crx = 0.0;
    double                cry = 0.0;
    double                clx = 0.0;
    double                cly = 0.0;
    double                dmsx = 0.0;
    double                dmsy = 0.0;
    int                   ff        = 0;
    int                   nplanes = 0;
    cpl_vector*           timev = NULL;
    cpl_propertylist*	  pl = NULL;

    fctab = sph_fctable_create_fctable(arawframe,SPH_FCTAB_DOUBLE);
    if ( !fctab ) {
        SPH_ERR("Could not create centre positions table.");
        return NULL;
    }
    nplanes = sph_fits_get_nplanes( cpl_frame_get_filename(
            arawframe), 0 );

    SPH_ERROR_ENSURE_GOTO_EXIT(
          nplanes == sph_fctable_get_size(fctab),
          CPL_ERROR_INCOMPATIBLE_INPUT );

    timev = sph_fctable_get_times(fctab);

    pl = sph_keyword_manager_load_properties( cpl_frame_get_filename( arawframe), 0 );
    SPH_ERROR_ENSURE_GOTO_EXIT(pl != NULL, CPL_ERROR_NULL_INPUT);
    SPH_ERROR_ENSURE_GOTO_EXIT(cpl_propertylist_has( pl, SPH_IRD_KEYWORD_DMS_POS_X ), CPL_ERROR_DATA_NOT_FOUND);
    SPH_ERROR_ENSURE_GOTO_EXIT(cpl_propertylist_has( pl, SPH_IRD_KEYWORD_DMS_POS_Y ), CPL_ERROR_DATA_NOT_FOUND);
    dmsx = cpl_propertylist_get_double( pl, SPH_IRD_KEYWORD_DMS_POS_X ) / SPH_IRD_INSTRUMENT_MODEL_PIX_SIZE_MICRONS;
    dmsy = cpl_propertylist_get_double( pl, SPH_IRD_KEYWORD_DMS_POS_Y ) / SPH_IRD_INSTRUMENT_MODEL_PIX_SIZE_MICRONS;
    for (ff = 0; ff < nplanes; ++ff) {
        rerr = sph_ird_common_science_interpolate_centers(
                fcframe,
                timev,
                ff,
                &clx,&cly,&crx,&cry);
        SPH_RAISE_CPL;

        SPH_ERROR_ENSURE_GOTO_EXIT( rerr == CPL_ERROR_NONE, rerr );

        clx += dmsx;
        cly += dmsy;
        crx += dmsx;
        cry += dmsy;

        SPH_ERROR_RAISE_INFO(
                SPH_ERROR_GENERAL,
                "Setting centre for %d as %f,%f and %f,%f.",
                ff,clx,cly,crx,cry);
        
        rerr = sph_fctable_set_cent_double(
                fctab,
                ff,
                clx,cly,crx,cry);
    }
    cpl_propertylist_delete( pl ); pl = NULL;
    cpl_vector_delete(timev); timev = NULL;
    return fctab;
EXIT:
    cpl_vector_delete(timev); timev = NULL;
    sph_fctable_delete(fctab); fctab = NULL;
    return NULL;
}

static
sph_error_code
sph_ird_common_science_interpolate_centers(
        cpl_frame* fcframe,
        cpl_vector* timev,
        int plane,
        double* clx,
        double* cly,
        double* crx,
        double* cry)
{
    double                  time_mjd = 0.0;
    double                  dtime1 = 10000000.0;
    double                  dtime2 = 10000000.0;
    double                  time1 = 0.0;
    double                  time2 = 0.0;
    double                  atime = 0.0;
    double					dmsx  = 0.0;
    double					dmsy  = 0.0;
    int                     rr      = 0;
    int                     r1      = -1;
    int                     r2      = -1;
    int                     bpix    = 0;
    double                  f       = 0.0;
    cpl_table*              interntab = NULL;

    cpl_ensure_code( fcframe, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( timev, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( clx, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( cly, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( crx, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( cry, CPL_ERROR_NULL_INPUT);

    time_mjd = cpl_vector_get( timev, plane );

    // Only one star center frame provided so take coords from that
    interntab = cpl_table_load( cpl_frame_get_filename( fcframe ), 1, 0 );
    if ( interntab == NULL ) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_OUTPUT,
                "Error loading star center calibration file from %s",
                cpl_frame_get_filename( fcframe ));
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    if ( cpl_table_get_nrow(interntab) == 1 ) {
        SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
                "Using single row field center table");
        dmsx = cpl_table_get_double(interntab,
                SPH_IRD_STAR_CENTER_DMS_POS_X_NAME,
                0,&bpix);
        dmsy = cpl_table_get_double(interntab,
                SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME,
                0,&bpix);
        *clx = cpl_table_get_double(interntab,
                SPH_IRD_STAR_CENTER_CENTLEFTX_NAME,
                0,&bpix) - dmsx;
        *cly = cpl_table_get_double(interntab,
                SPH_IRD_STAR_CENTER_CENTLEFTY_NAME,
                0,&bpix) - dmsy;
        *crx = cpl_table_get_double(interntab,
                SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME,
                0,&bpix) - dmsx;
        *cry = cpl_table_get_double(interntab,
                SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME,
                0,&bpix) - dmsy;
        cpl_table_delete(interntab);interntab = NULL;
        return cpl_error_get_code();
    }

    if ( sph_time_mjd_from_string(&atime, cpl_table_get_string(
            interntab,
            SPH_IRD_STAR_CENTER_TIME_UT_NAME,
            cpl_table_get_nrow(interntab) - 1) ) != CPL_ERROR_NONE ) {
        cpl_table_delete(interntab);interntab = NULL;
        return cpl_error_get_code();
    }

    if ( atime < time_mjd ) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                "The latest timestamp for the star center "
                "calibration data is %7.3f which is before %7.3f "
                "of the science frame. This is not allowed."
                "Rerun the star center calibration with input frames"
                " taken after the science frames or run it with a single"
                " input raw frame or remove the star center calibration "
                "inputs altogether.", atime, time_mjd);
        cpl_table_delete(interntab);interntab = NULL;
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    for (rr = 0; rr < cpl_table_get_nrow(interntab); ++rr) {
        cpl_ensure_code(sph_time_mjd_from_string(&atime, cpl_table_get_string(
                interntab,
                SPH_IRD_STAR_CENTER_TIME_UT_NAME,
                rr) ) == CPL_ERROR_NONE, cpl_error_get_code() ) ;
        if ( atime <= time_mjd && fabs(atime - time_mjd ) < dtime1 ) {
            dtime1 = fabs(atime - time_mjd );
            r1 = rr;
            time1 = atime;
        }
    }
    for (rr = 0; rr < cpl_table_get_nrow(interntab); ++rr) {
        if (sph_time_mjd_from_string(&atime, cpl_table_get_string(
                interntab,
                SPH_IRD_STAR_CENTER_TIME_UT_NAME,
                rr) ) != CPL_ERROR_NONE ) {
            cpl_table_delete(interntab);interntab = NULL;
            return cpl_error_get_code() ;
        }
        if ( atime >= time_mjd &&
                fabs(atime - time_mjd ) < dtime2 && rr != r1 ) {
            dtime2 = fabs(atime - time_mjd );
            r2 = rr;
            time2 = atime;
        }
    }

    if ( r1 < 0 ) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                "There is a problem with the timestamps for the star center "
                "calibration data. Please chekc the data.");
        cpl_table_delete(interntab);interntab = NULL;
        return CPL_ERROR_ILLEGAL_INPUT;
    }
    if ( r2 < 0 ) {
        f = 0.0; r2 = 0;
    }
    else {
        if ( time2 < time1 ) {
            f = time2; time2 = time1; time1 = f;
        }
        SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
                "time1: %f, time2: %f,, timemjd:%f",
                time1, time2, time_mjd);
        if ( time1 > time_mjd ) {
           SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                "There is a problem with the timestamps for the star center "
                "calibration data. Please chekc the data.");
           cpl_table_delete(interntab);interntab = NULL;
           return CPL_ERROR_UNSUPPORTED_MODE;
        }
        if ( time2 < time_mjd ) {
           SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                "There is a problem with the timestamps for the star center "
                "calibration data. Please chekc the data.");
           cpl_table_delete(interntab);interntab = NULL;
           return CPL_ERROR_UNSUPPORTED_MODE;
        }

        f = (time_mjd - time1 ) / ( time2 - time1 ) ;
    }
    dmsx = cpl_table_get_double(
            interntab,
            SPH_IRD_STAR_CENTER_DMS_POS_X_NAME,
            r1, &bpix) +
            f * ( cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_DMS_POS_X_NAME,
                    r2, &bpix) -
                  cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_DMS_POS_X_NAME,
                    r1, &bpix) );
    dmsy = cpl_table_get_double(
            interntab,
            SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME,
            r1, &bpix) +
            f * ( cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME,
                    r2, &bpix) -
                  cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME,
                    r1, &bpix) );
    *clx = cpl_table_get_double(
            interntab,
            SPH_IRD_STAR_CENTER_CENTLEFTX_NAME,
            r1, &bpix) +
            f * ( cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTLEFTX_NAME,
                    r2, &bpix) -
                  cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTLEFTX_NAME,
                    r1, &bpix) ) - dmsx;
    *cly = cpl_table_get_double(
            interntab,
            SPH_IRD_STAR_CENTER_CENTLEFTY_NAME,
            r1, &bpix) +
            f * ( cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTLEFTY_NAME,
                    r2, &bpix) -
                  cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTLEFTY_NAME,
                    r1, &bpix) ) - dmsy;
    *crx = cpl_table_get_double(
            interntab,
            SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME,
            r1, &bpix) +
            f * ( cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME,
                    r2, &bpix) -
                  cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME,
                    r1, &bpix) ) - dmsx;
    *cry = cpl_table_get_double(
            interntab,
            SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME,
            r1, &bpix) +
            f * ( cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME,
                    r2, &bpix) -
                  cpl_table_get_double(
                    interntab,
                    SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME,
                    r1, &bpix) ) - dmsy;

    cpl_table_delete(interntab);interntab = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}



