/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"

#include "sph_ird_tags.h"
#include "sph_ird_science_imaging.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_simple_adi.h"
#include "sph_utils.h"
#include "sph_ird_instrument_model.h"
#include "sph_framecombination.h"
#include "sph_common_science.h"
#include "sph_ird_common_science.h"
#include "sph_filemanager.h"
#include "sph_keyword_manager.h"
#include "sph_differential_imaging.h"

const char* const SPH_IRD_STAR_CENTER_FILE_NAME = "WAFFFRAME";
const char* const SPH_IRD_STAR_CENTER_ANGLE_NAME = "ANGLE";


/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IRD_SCIENCE_IMAGING_GENERAL;
extern sph_error_code SPH_IRD_SCIENCE_IMAGING_FRAMES_MISSING;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_science_imaging_run
 *
 * This module provides the core recipe for the IRDIS science recipe
 * in the classical imaging mode.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_science_imaging.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_science_imaging recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/


cpl_error_code sph_ird_science_imaging_run( sph_ird_science_imaging* self )
{
    sph_master_frame*           result = NULL;
    sph_master_frame*           leftFOV = NULL;
    sph_master_frame*           rightFOV = NULL;
    cpl_propertylist*            plist = NULL;
    sph_ird_common_science*     sci = NULL;
    sph_transform*              transform = NULL;
    cpl_frameset*               total_frameset = NULL;
    cpl_frameset*               left_frameset_out = NULL;
    cpl_frameset*               right_frameset_out = NULL;
    cpl_frameset*               adi_frameset = NULL;
    cpl_frameset*               fctable_frames = NULL;
    cpl_vector*                 angles  = NULL;

    if ( self->transform_method == 0 ) {
        if ( self->filter_method == 0 ) {
            transform = sph_transform_new_fft(
                    SPH_FFT_FILTER_METHOD_NONE,0.0,0.0);
        }
        else if ( self->filter_method == 1 ) {
            transform = sph_transform_new_fft(
                    SPH_FFT_FILTER_METHOD_TH,
                    self->filter_radius,0.0);
        }
        else if ( self->filter_method == 2 ) {
            transform = sph_transform_new_fft(
                    SPH_FFT_FILTER_METHOD_FERMI,
                    self->filter_radius,
                    self->fermi_temp);
        }
        else if ( self->filter_method == 3 ) {
            transform = sph_transform_new_fft(
                    SPH_FFT_FILTER_METHOD_BUTTER,
                    self->butter_pass,
                    self->butter_stop);
        }
    }
    else if ( self->transform_method == 1) {
        transform = sph_transform_new_cpl_warp();
    }

    sci = sph_ird_common_science_new(
            self->inframes,
            self->inparams,
            self->dark_frame,
            self->flat_frame,
            self->static_badpixel_frame,
            self->distmap_frame,
            NULL,
            self->flat_frame,
            SPH_RECIPE_NAME_IRD_SCIENCE_IMAGING,
            SPH_IRD_TAG_SCIENCE_IMAGING_CALIB);

    if ( self->tff_frame ) {
        SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
                "Setting telescope flat %s. "
                "Note that this flat is applied after "
                "shifting/de-rotation",
                cpl_frame_get_filename( self->tff_frame ) );
        sph_ird_common_science_set_tff( sci, self->tff_frame );
    }
    sci->flag_sdi = 0;
    sci->rotate_flag = 0;

    if ( transform )
        sph_ird_common_science_set_transform(sci,
                transform);

    if ( self->fcasciis != NULL ) {
          fctable_frames = cpl_frameset_duplicate(self->fcasciis);
          if ( fctable_frames == NULL ) {
              goto EXIT;
          }
      }
      else {
          if ( self->fcframe ) {
              fctable_frames = sph_ird_common_science_create_fctables(self->fcframe,self->rawframes,self->keep_fctable);
              if ( fctable_frames == NULL ) {
                  goto EXIT;
              }
          }
      }


    if ( sph_ird_common_science_process_cubes(sci,
            self->rawframes,
            fctable_frames,
            &left_frameset_out,
            &right_frameset_out,
            &total_frameset) == CPL_ERROR_NONE)
    {

        if ( self->flag_adi ) {
            angles = sph_common_science_extract_angles(fctable_frames);

            adi_frameset = sph_simple_adi_process_cubes(
                    total_frameset,
                    self->rawframes,
                    angles,
                    sci->transform,
                    sci->irdmodel,0,1,2,self->fullset);
            result = sph_common_science_combine(
                    adi_frameset,
                    self->coll_alg,
                    0,1,-1);
            cpl_frameset_delete(adi_frameset);

            adi_frameset = sph_simple_adi_process_cubes(
                     left_frameset_out,
                     self->rawframes,
                     angles,
                     sci->transform,
                     sci->irdmodel,0,1,2,self->fullset);
             leftFOV = sph_common_science_combine(
                     adi_frameset,
                     self->coll_alg,
                     0,1,-1);
             cpl_frameset_delete(adi_frameset);

             adi_frameset = sph_simple_adi_process_cubes(
                      right_frameset_out,
                      self->rawframes,
                      angles,
                      sci->transform,
                      sci->irdmodel,0,1,2,self->fullset);
              rightFOV = sph_common_science_combine(
                      adi_frameset,
                      self->coll_alg,
                      0,1,-1);
              cpl_frameset_delete(adi_frameset);
              cpl_vector_delete(angles); angles = NULL;
        }
        else {
            result = sph_common_science_combine(
                    total_frameset,
                    self->coll_alg,
                    0,1,-1);
            leftFOV = sph_common_science_combine(
                    left_frameset_out,
                    self->coll_alg,
                    0,1,-1);
            rightFOV = sph_common_science_combine(
                    right_frameset_out,
                    self->coll_alg,
                    0,1,-1);
        }
        if ( result ) {
            sph_master_frame_save_dfs(result,
                    self->science_imaging_outfilename,
                    self->inframes,
                    cpl_frameset_get_first(self->rawframes),
                    self->inparams,
                    SPH_IRD_TAG_SCIENCE_IMAGING_CALIB,
                    SPH_RECIPE_NAME_IRD_SCIENCE_IMAGING,
                    SPH_PIPELINE_NAME_IRDIS, plist );
        }
        else {
            SPH_ERR("Could not create product frame.");
        }
        if ( leftFOV ) {
            sph_master_frame_save_dfs(leftFOV,
                    self->science_imaging_outfilename_left,
                    self->inframes,
                    cpl_frameset_get_first(self->rawframes),
                    self->inparams,
                    SPH_IRD_TAG_SCIENCE_IMAGING_CALIB_LEFT,
                    SPH_RECIPE_NAME_IRD_SCIENCE_IMAGING,
                    SPH_PIPELINE_NAME_IRDIS, plist );
        }
        if ( rightFOV ) {
            sph_master_frame_save_dfs(rightFOV,
                    self->science_imaging_outfilename_right,
                    self->inframes,
                    cpl_frameset_get_first(self->rawframes),
                    self->inparams,
                    SPH_IRD_TAG_SCIENCE_IMAGING_CALIB_RIGHT,
                    SPH_RECIPE_NAME_IRD_SCIENCE_IMAGING,
                    SPH_PIPELINE_NAME_IRDIS, plist );
        }
    }
EXIT:
    sph_master_frame_delete(result);result=NULL;
    sph_master_frame_delete(leftFOV);result=NULL;
    sph_master_frame_delete(rightFOV);result=NULL;
    cpl_propertylist_delete(plist); plist = NULL;
    cpl_frameset_delete(total_frameset);
    cpl_frameset_delete(left_frameset_out);
    cpl_frameset_delete(right_frameset_out);
    cpl_frameset_delete(adi_frameset);
    sph_ird_common_science_delete(sci);
    sph_filemanager_clean();
    return (int)cpl_error_get_code();
}

/**@}*/


