/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_keyword_manager.h"
#include "sph_ird_keywords.h"

#include "sph_ird_tags.h"
#include "sph_ird_spectra_resolution.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_apertures.h"
#include "sph_ird_instrument_model.h"
#include "sph_framecombination.h"

/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IRD_SPECTRA_RESOLUTION_GENERAL;
extern sph_error_code SPH_IRD_SPECTRA_RESOLUTION_FRAMES_MISSING;

typedef struct _sph_ird_spectra_resolution_peak_ {
    double lambda0;
    double lambda;
    double pos;
    double pos_width;
    double lambda_width;
} sph_ird_spectra_resolution_peak;

static double
sph_ird_spectra_resolution_get_lambda( sph_ird_spectra_resolution* self,
        cpl_apertures* aps, int n );

static
double
sph_ird_spectra_resolution_get_dy( sph_ird_spectra_resolution* self,
        cpl_apertures* aps, int n ) ;

static
sph_master_frame*
create_science( sph_ird_spectra_resolution* self,
        sph_master_frame* dark,
        sph_master_frame* flat,
        sph_ird_instrument_model* model,
        cpl_frameset* rawframes )
{
    sph_master_frame*                mframe = NULL;
    int                              xx = 0;
    int                              yy = 0;
    cpl_mask*                         mask = NULL;
    mframe = sph_framecombination_master_frame_from_cpl_frameset(
            rawframes,
            self->coll_alg,
            self->framecomb_parameterlist);
    if ( !mframe ) return NULL;

    if ( dark ) sph_master_frame_subtract_master_frame(mframe,dark);
    if ( flat ) sph_master_frame_divide_master_frame(mframe,flat);
    mask = cpl_mask_new(cpl_image_get_size_x(mframe->image),cpl_image_get_size_y(mframe->image));
    for (xx = 0; xx < cpl_image_get_size_x(mframe->image); ++xx) {
        for (yy = 0; yy < cpl_image_get_size_y(mframe->image); ++yy) {
            if ( sph_ird_instrument_model_pixel_inside_left(model,xx,yy) == 0
                    && sph_ird_instrument_model_pixel_inside_right(model,xx,yy) == 0 ) {
                cpl_mask_set(mask,xx+1,yy+1,1);
            }
            else {
                cpl_mask_set(mask,xx+1,yy+1,0);
            }
        }
    }
    sph_master_frame_set_bads_from_mask(mframe,mask);
    sph_master_frame_set_bads(mframe,0.0);
    cpl_mask_delete(mask); mask = NULL;
    SPH_RAISE_CPL_RESET;
    return mframe;
}

static
cpl_propertylist*
find_peaks( sph_ird_spectra_resolution* self, double sigma,
        sph_master_frame* mframe ) {
    cpl_apertures*        ap = NULL;
    double                dx = 0.0;
    double                dy = 0.0;
    int                    ii = 0;
    cpl_propertylist*    pl = cpl_propertylist_new();
    int                     npeaks = 0;
    cpl_vector*             respower = NULL;
    double lowest_line = 100000.0;
    int lowest_index = -1;

    ap = cpl_apertures_extract_sigma( mframe->image, sigma );
    if ( !ap ) {
        SPH_RAISE_CPL;
        return NULL;
    }
    npeaks = cpl_apertures_get_size(ap);
    sph_error_raise( SPH_ERROR_GENERAL,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO, "Found %d peaks",
            npeaks);
    if ( npeaks < 1 ) {
        SPH_ERR("Can't measure the spectra resolution without "
                "at least 1 detected peak. "
                "Ideall should detect 2 lines for the point source "
                "in both of the two IRDIS optical paths, "
                "giving 4 detections in total. "
                "Change the sigma for detection and/or check hardware alignment.");
        cpl_apertures_delete(ap); ap = NULL;
        return NULL;
    }
    for (ii = 0; ii < npeaks; ++ii) {
        dy = cpl_apertures_get_centroid_y(ap,ii + 1);
        if ( dy < lowest_line ) {
            lowest_line = dy;
            lowest_index =ii;
            self->y0 = dy;
        }
    }
    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
             "Lowest peak found at : x = %5.2f, y = %5.2f, "
             "so setting Y0 to %5.2f",
             cpl_apertures_get_centroid_x(ap,lowest_index + 1),
             cpl_apertures_get_centroid_y(ap,lowest_index + 1),
             self->y0 );

    respower = cpl_vector_new(npeaks);
    for (ii = 0; ii < npeaks; ++ii) {
        double lambd =  sph_ird_spectra_resolution_get_lambda(self,ap,ii+1);
        double dly = sph_ird_spectra_resolution_get_dy(self,ap,ii+1);
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Found peak %d: x = %5.2f, y = %5.2f, "
                "w = %5.2f, lambda= %5.2f, dlambda = %5.2f",
                ii,
                cpl_apertures_get_centroid_x(ap,ii + 1),
                cpl_apertures_get_centroid_y(ap,ii + 1),
                sph_apertures_get_width_x(ap,ii+1,mframe),
                lambd,
                dly
                );

        sph_keyword_manager_update_double1(pl,
                SPH_IRD_KEYWORD_SPEC_RES_LINE_X,
                ii+1,
                cpl_apertures_get_centroid_x(ap,ii + 1) );
        sph_keyword_manager_update_double1(pl,
                SPH_IRD_KEYWORD_SPEC_RES_LINE_Y,
                ii+1,
                cpl_apertures_get_centroid_y(ap,ii + 1) );
        sph_keyword_manager_update_double1(pl,
                SPH_IRD_KEYWORD_SPEC_RES_LINE_WIDTHX,
                ii+1,
                sph_apertures_get_width_x(ap,ii+1,mframe) );
        sph_keyword_manager_update_double1(pl,
                SPH_IRD_KEYWORD_SPEC_RES_LINE_LAMBDA,
                ii+1,
                lambd );
        sph_keyword_manager_update_double1(pl,
                SPH_IRD_KEYWORD_SPEC_RES_LINE_LAMBDA_WIDTH,
                ii+1,
                dly );
        dx += sph_apertures_get_width_x(ap,ii+1,mframe);
        dy += dly;
        if ( dly ) cpl_vector_set(respower,ii,
                lambd /
                fabs(dly));
    }
    dx = dx / npeaks;
    dy = dy / npeaks;

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Number of lines : %d",
            npeaks
            );
    cpl_propertylist_append_int(pl,SPH_IRD_KEYWORD_SPEC_RES_NLINES,npeaks);

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Wavelength resolution : %5.2f",
            dy
            );
    cpl_propertylist_append_double(pl,SPH_IRD_KEYWORD_SPEC_RES_LAMBDA_RES,dy);

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Spatial resolution : %5.2f",
            dx
            );
    cpl_propertylist_append_double(pl,SPH_IRD_KEYWORD_SPEC_RES_SPATIAL_RES,dx);

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Mean resolution power : %5.2f +/- %5.2f",
            cpl_vector_get_mean(respower),
            cpl_vector_get_stdev(respower)
            );
    cpl_propertylist_append_double(pl,SPH_IRD_KEYWORD_SPEC_RES_RESPOWER,cpl_vector_get_mean(respower));
    cpl_propertylist_append_double(pl,SPH_IRD_KEYWORD_SPEC_RES_RESPOWER_RMS,cpl_vector_get_stdev(respower));
    cpl_apertures_delete(ap); ap = NULL;
    cpl_vector_delete(respower); respower = NULL;
    return pl;
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_spectra_resolution_run Create Master Dark Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * master dakr
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_spectra_resolution.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_spectra_resolution recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ird_spectra_resolution_run( sph_ird_spectra_resolution* self )
{
        sph_master_frame*           dark         = NULL;
        sph_master_frame*            flat     = NULL;
        sph_master_frame*            result = NULL;
        cpl_propertylist*            plist = NULL;
        cpl_propertylist*            plist2 = NULL;
        sph_ird_instrument_model*    model = NULL;
        sph_pixel_description_table*    pdt = NULL;
        cpl_image*                      maskimage = NULL;

        /*------------------------------------------------------------------
        -  Placing master dark in cube structure for later use
        --------------------------------------------------------------------*/


        /*------------------------------------------------------------------
         -  Now process all sets of wavelength frames separately
         --------------------------------------------------------------------*/
        if ( self->master_dark_frame )
        {
            dark = sph_master_frame_load(
                    cpl_frame_get_filename(self->master_dark_frame),0);
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Will be applying dark %s",
                    cpl_frame_get_filename(self->master_dark_frame));
        }
        if ( self->master_flat_frame )
        {
            flat = sph_master_frame_load(
                    cpl_frame_get_filename(self->master_flat_frame),0);
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Will be applying flat %s",
                    cpl_frame_get_filename(self->master_flat_frame));
        }
        if ( self->wave_calib_frame ) {
            pdt = sph_pixel_description_table_load(
                    cpl_frame_get_filename(self->wave_calib_frame));
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Will be using wave calib %s",
                    cpl_frame_get_filename(self->wave_calib_frame));
            plist = sph_keyword_manager_load_properties(cpl_frame_get_filename(self->wave_calib_frame),0);
            if ( plist ) {
                if ( cpl_propertylist_has( plist, SPH_IRD_KEYWORD_WAVECALIB_C1 ) ) {
                    self->c1 = cpl_propertylist_get_double( plist, SPH_IRD_KEYWORD_WAVECALIB_C1 );
                }
                cpl_propertylist_delete(plist); plist = NULL;
            }
        }
        if ( self->mask_frame ) {
            maskimage = cpl_image_load(
                    cpl_frame_get_filename(self->mask_frame),
                    CPL_TYPE_INT,0,0);
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Will be use mask %s",
                    cpl_frame_get_filename(self->mask_frame));
        }

        plist = sph_keyword_manager_load_properties(cpl_frame_get_filename(self->master_flat_frame),0);
        model = sph_ird_instrument_model_new_from_propertylist(plist);
        cpl_propertylist_delete(plist); plist = NULL;
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Combining "
                "raw frames to create master...");
        result = create_science(self,dark,flat,model,self->rawframes);
        if ( result ) {
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"...Done");
            if ( maskimage ) {
                sph_master_frame_set_bads_from_image(result,maskimage);
                sph_master_frame_set_bads(result,0.0);
            }
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Finding peaks...");
            plist2 = find_peaks(self,self->threshold_sigma,result);
            plist = sph_ird_instrument_model_get_as_propertylist(model);
            cpl_propertylist_append(plist,plist2);
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"Saving master.");
            sph_master_frame_save_dfs(result,self->outfilename,
                    self->inframes,NULL,self->inparams,
                    SPH_IRD_TAG_SPECTRA_RESOLUTION_CALIB,
                    SPH_RECIPE_NAME_IRD_SPECTRA_RESOLUTION,
                    SPH_PIPELINE_NAME_IRDIS, plist);
            sph_master_frame_delete(result);result=NULL;
        }
        if ( dark ) {
            sph_master_frame_delete(dark); dark = NULL;
        }
        if ( flat ) {
            sph_master_frame_delete(flat); flat = NULL;
        }
        sph_pixel_description_table_delete(pdt); pdt = NULL;
        cpl_propertylist_delete(plist); plist = NULL;
        cpl_propertylist_delete(plist2); plist2 = NULL;
        sph_ird_instrument_model_delete(model); model = NULL;
        cpl_image_delete(maskimage); maskimage = NULL;
        //sph_filemanager_clean();
        return (int)cpl_error_get_code();
}

static
double
sph_ird_spectra_resolution_get_lambda( sph_ird_spectra_resolution* self,
        cpl_apertures* aps, int n) {
    double dx = 0.0;
    double dy = 0.0;
    double lambda = -1.0;

    dx = cpl_apertures_get_centroid_x(aps,n);
    dy = cpl_apertures_get_centroid_y(aps,n);
    dy = dy - self->y0;

    lambda = self->wavelength_line2 + self->c1 / 100.0 * dy + self->c2 * dy * dy / (100.0 * 100.0);
    lambda = lambda + self->c3 / ( 100.0 * 100.0 * 100.0 ) * dy * dy * dy;
    lambda = lambda + self->c4 / ( 100.0 * 100.0 * 100.0 * 100.0 ) * dy * dy * dy * dy;
    return lambda;
}

static
double
sph_ird_spectra_resolution_get_dy( sph_ird_spectra_resolution* self,
        cpl_apertures* aps, int n ) {
    double dx = 0.0;
    double dy = 0.0;
    double lambda = -1.0;

    dx = cpl_apertures_get_centroid_x(aps,n);
    dy = cpl_apertures_get_centroid_y(aps,n);
    dy = dy - self->y0;

    lambda = self->c1 / 100.0 + 2.0 * self->c2 * dy / (100.0 * 100.0);
    lambda = lambda + self->c3 / ( 100.0 * 100.0 * 100.0 ) * 3.0 * dy * dy;
    lambda = lambda + self->c4 / ( 100.0 * 100.0 * 100.0 * 100.0 ) * 4.0 * dy * dy * dy;
    return lambda;
}

/**@}*/
