/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"

#include "sph_ird_tags.h"
#include "sph_ird_star_center.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ird_instrument_model.h"
#include "sph_framecombination.h"
#include "sph_ird_common_science.h"
#include "sph_keyword_manager.h"
#include "sph_time.h"
#include "sph_common_science.h"
#include "sph_cube.h"
#include "sph_filemanager.h"
#include "sph_fitting.h"
#include "sph_fits.h"

/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IRD_STAR_CENTER_GENERAL;
extern sph_error_code SPH_IRD_STAR_CENTER_FRAMES_MISSING;
const sph_error_code SPH_IRD_STAR_CENTER_BAD_NUMBER_WAFFIMS = SPH_IRD_STAR_CENTER_ERR_START+5;

const int SPH_IRD_STAR_CENTER_MINPIX = 10;
const int SPH_IRD_STAR_CENTER_MAX_WAFFIMS = 4;
const double SPH_IRD_STAR_CENTER_FIT_FWHM_GUESS = 3.0;

const char* const SPH_IRD_STAR_CENTER_FILE_NAME = "WAFFFRAME";
const char* const SPH_IRD_STAR_CENTER_ANGLE_NAME = "ANGLE";
const double SPH_IRD_PIX_SIZE = 18.0;

typedef struct _sph_ird_star_center_star_image_ {
    double            cx;
    double            cy;
    int                npix;
} sph_ird_star_center_star_image;

static
sph_error_code
__sph_ird_star_center_init_inter_time_cent_tab(
        sph_ird_star_center* self,
        int nwaffle_images )
{
    cpl_ensure_code( self,
            CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( nwaffle_images > 0,
            CPL_ERROR_ILLEGAL_INPUT );

    if ( self->interntab ) {
        cpl_table_delete( self->interntab );
        self->interntab = NULL;
    }
    self->interntab = cpl_table_new(nwaffle_images);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_TIME_UT_NAME,
            CPL_TYPE_STRING);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_CENTLEFTX_NAME,
            CPL_TYPE_DOUBLE);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_CENTLEFTY_NAME,
            CPL_TYPE_DOUBLE);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME,
            CPL_TYPE_DOUBLE);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME,
            CPL_TYPE_DOUBLE);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_DMS_POS_X_NAME,
            CPL_TYPE_DOUBLE);
    cpl_table_new_column(self->interntab,
            SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME,
            CPL_TYPE_DOUBLE);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


static
cpl_image*
__sph_ird_star_center_get_extraction_image(
        sph_master_frame* mf,
        sph_ird_instrument_model* model,
        cpl_image** errim)
{
    cpl_image*              tmprms = NULL;
    cpl_mask*               bmask = NULL;
    cpl_image*              im = NULL;

    cpl_ensure( mf, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure( model, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure( errim, CPL_ERROR_NULL_INPUT, NULL);

    im = sph_master_frame_get_image(mf,1);
    tmprms = sph_master_frame_get_rms(mf);
    sph_master_frame_set_rms_poisson(mf,1e10,0);
    *errim = sph_master_frame_get_rms(mf);
    cpl_image_delete(mf->rmsmap); mf->rmsmap = tmprms;
    bmask = sph_ird_instrument_model_get_mask_dbi_fov(
            model,10.0);
    cpl_image_reject_from_mask(im,bmask);
    cpl_mask_delete(bmask); bmask = NULL;
    cpl_image_fill_rejected(im,0.0);
    cpl_image_accept_all(*errim);
    return im;
}

static
sph_error_code
__sph_ird_star_center_fit_center(
        cpl_apertures* aps,
        int aa,
        cpl_image* im,
        cpl_image* errim,
        sph_ird_star_center_star_image* sim
        )
{
    double          xfit        = 0;
    double          yfit        = 0;
    double          wx          = 0;
    double          wy          = 0;

    cpl_ensure_code( aps, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( im, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( errim, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( sim, CPL_ERROR_NULL_INPUT );
    sim->cx = cpl_apertures_get_centroid_x(aps,aa+1) - 0.5;
    sim->cy = cpl_apertures_get_centroid_y(aps,aa+1) - 0.5;
    sim->npix = cpl_apertures_get_npix(aps,aa+1);

    SPH_ERROR_RAISE_INFO(
            SPH_ERROR_GENERAL,
            "Peak found: centroid: %f, %f and npix %d",
            sim->cx,
            sim->cy,
            sim->npix);

    xfit = sim->cx + 0.5;
    yfit = sim->cy + 0.5;
    wx = wy = SPH_IRD_STAR_CENTER_FIT_FWHM_GUESS;
    if ( sph_fitting_fit_gauss2D(
            im,errim,
            &xfit,&yfit,&wx,&wy) == CPL_ERROR_NONE )
    {
        SPH_ERROR_RAISE_INFO(
                SPH_ERROR_GENERAL,
                "Centroid: %f, %f. "
                "Gauss fit :%f, %f.",
                sim->cx,
                sim->cy,
                xfit - 0.5,
                yfit - 0.5);

        sim->cx = xfit - 0.5;
        sim->cy = yfit - 0.5;
    }
    else {
        cpl_error_reset();
        SPH_ERROR_RAISE_WARNING(
                SPH_ERROR_GENERAL,
                "Could not do gauss fit. Am using centroid.");
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_ird_star_center_process_next(
        sph_ird_star_center* self,
        cpl_frameset* frameset,
        sph_master_frame** left,
        sph_master_frame** right)
{
    cpl_image*          im = NULL;
    sph_master_frame*   master_raw = NULL;
    cpl_mask*           bmask = NULL;

    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( left, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( right, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( frameset, CPL_ERROR_NULL_INPUT);

    master_raw = sph_common_science_get_next_master_frame(frameset);
    if ( master_raw ) {
        if ( master_raw ) {
            if ( self->dark )
                sph_master_frame_subtract_master_frame(
                        master_raw,
                        self->dark);
            if ( self->flat )
                sph_master_frame_divide_master_frame(
                        master_raw,
                        self->flat);


            bmask = sph_master_frame_get_badpixelmask(master_raw);
            sph_master_frame_interpolate_bpix(master_raw);

            sph_master_frame_set_bads_from_mask(master_raw,bmask);

            cpl_mask_delete(bmask); bmask = NULL;


            if ( self->stat_badpix_mask ) {
                   SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                        "Applying mask frame %s to image.",
                         cpl_frame_get_filename(self->static_badpixel_frame) );
                         sph_master_frame_set_bads_from_mask(master_raw,self->stat_badpix_mask);
            }
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Setting all masked pixels to 0.");
                sph_master_frame_set_bads(master_raw,0.0);

            *left  =
                    sph_ird_instrument_model_extract_left_master_frame(
                            self->model,
                            master_raw);
            *right  =
                    sph_ird_instrument_model_extract_right_master_frame(
                            self->model,
                            master_raw);
            self->current_frame = sph_common_science_get_current_raw_frame();
        }
        sph_master_frame_delete(master_raw); master_raw = NULL;
    }
    else {
        return CPL_ERROR_EOL;
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
sph_ird_star_center_get_center(
        sph_master_frame* mf, double* dx,
        double* dy,double sigma, int nsources,
        sph_ird_instrument_model* model)
{
    sph_error_code         rerr = CPL_ERROR_NONE;
    cpl_apertures*        aps = NULL;
    cpl_image*            im = NULL;
    int                    aa    = 0;
    int                    nwaffles = 0;
    int                    usedwaffs = 0;
    double                cx = 0.0;
    double                cy = 0.0;
    cpl_image*              errim = NULL;

    sph_ird_star_center_star_image        sim[4];

    cpl_ensure_code( mf, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( dx, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( dy, CPL_ERROR_NULL_INPUT );

    im = __sph_ird_star_center_get_extraction_image(
            mf,
            model,
            &errim);

    if ( im ) {
        aps = cpl_apertures_extract_sigma(im,sigma);

        if ( aps == NULL )
        {
            SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                    "Could not find any sources on the input above a sigma "
                    "of %f. Check that the quality of the input is sufficient, "
                    "and that the flat field and dark have reasonable quality."
                    "", sigma);
            cpl_image_delete(im); im = NULL;
            cpl_image_delete(errim); errim = NULL;
            return cpl_error_get_code();
        }

        cpl_apertures_sort_by_npix(aps);
        cpl_image_accept_all(im);

        for (aa = 0; aa < nsources; ++aa) {
            if ( cpl_apertures_get_npix(aps,aa+1) >
                            SPH_IRD_STAR_CENTER_MINPIX )
            {
                if ( nwaffles < SPH_IRD_STAR_CENTER_MAX_WAFFIMS )
                {

                    __sph_ird_star_center_fit_center(
                            aps, aa, im, errim, &sim[nwaffles]);

                }
                nwaffles++;
            }
        }
        cpl_image_delete(im);im = NULL;
        cpl_apertures_delete(aps);aps = NULL;
    }
    usedwaffs = SPH_IRD_STAR_CENTER_MAX_WAFFIMS;
    if ( nwaffles < usedwaffs ) usedwaffs = nwaffles;
    for (aa = 0; aa < usedwaffs; ++aa) {
        cx += sim[aa].cx;
        cy += sim[aa].cy;
    }
    if ( usedwaffs ) {
        cx/=usedwaffs;
        cy/=usedwaffs;
    }
    else {
        rerr = CPL_ERROR_ILLEGAL_INPUT;
    }
    cpl_image_delete(errim); errim = NULL;
    *dx=cx;
    *dy=cy;
    return rerr;
}


static
cpl_frameset*
sph_ird_star_center_collapse_waffle_frames(
        sph_ird_star_center* self,
        cpl_frame* waffle_frame) 
{
    cpl_frameset*       result = NULL;
    cpl_frameset*       tmpset = NULL;
    cpl_frame*          aframe = NULL;
    cpl_frame*          newframe = NULL;
    sph_master_frame*   mframe = NULL;
    cpl_propertylist*   pl = NULL;

    cpl_ensure( self,
            CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure( waffle_frame,
            CPL_ERROR_NULL_INPUT, NULL);

    result = cpl_frameset_new();

    SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
            "Collapsing all raw waffle cubes "
            "(if any) using median...");

    aframe = cpl_frame_duplicate(
            waffle_frame );
    tmpset = cpl_frameset_new();
    cpl_frameset_insert( tmpset, aframe );
    pl = cpl_propertylist_load(
            cpl_frame_get_filename(aframe), 0);

    newframe = sph_filemanager_get_unique_duplicate_frame(aframe);
    mframe = sph_common_science_combine( tmpset,
            SPH_COLL_ALG_MEDIAN, 0, -1, -1);
    cpl_frameset_delete(tmpset); tmpset = NULL;

    sph_master_frame_save( mframe,
            cpl_frame_get_filename(newframe), pl);
    cpl_frameset_insert(result,newframe);
    cpl_propertylist_delete(pl); pl = NULL;
    sph_master_frame_delete(mframe); mframe = NULL;

    return result;
}

static
sph_error_code
sph_ird_star_center_fill_waffle_table(
        sph_ird_star_center* self,
        cpl_frameset* collapsed_waffle_frames)
{
    sph_error_code        rerr = CPL_ERROR_NONE;
    sph_master_frame*    left = NULL;
    sph_master_frame*    right = NULL;
    double                crx = 0.0;
    double                 cry = 0.0;
    double                clx = 0.0;
    double                 cly = 0.0;
    int                    ff        = 0;
    double                  dms_x = 0.0;
    double                  dms_y = 0.0;
    double                  angle = 0.0;
    cpl_frameset*           tmpframeset = NULL;
    sph_transform*          transform = NULL;
    char                    ut_time[256];
    cpl_propertylist*       plist = NULL;

    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( collapsed_waffle_frames, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( self->interntab, CPL_ERROR_ILLEGAL_INPUT );

    transform = sph_transform_new_default();

    ff= 0;
    while ( __sph_ird_star_center_process_next(
                self, collapsed_waffle_frames,
                &left, &right ) == CPL_ERROR_NONE )
    {
        rerr = sph_ird_star_center_get_center(
                left,
                &clx,&cly,
                self->sigma,
                self->nsources,
                self->model);
        SPH_ERROR_ENSURE_GOTO_EXIT( rerr == CPL_ERROR_NONE, rerr );

        if ( self->qc_left )
        {
            sph_transform_apply(
                    transform,left,
                    NULL, clx, cly, 0.0, 1.0,
                    self->model);
            sph_master_frame_add_master_frame(
                    self->qc_left,left);
        }

        plist = sph_keyword_manager_load_properties(
                cpl_frame_get_filename(self->current_frame), 0);

        rerr = sph_ird_star_center_get_center(
                right,&crx,&cry,
                self->sigma,self->nsources,
                self->model);
        SPH_ERROR_ENSURE_GOTO_EXIT( rerr == CPL_ERROR_NONE, rerr );
        // TODO: Here comes treatment of angle...
        angle = 0.0;
        dms_x = dms_y = 0.0;
        if ( cpl_propertylist_has( plist,
                SPH_IRD_KEYWORD_DMS_POS_X) ) {
            dms_x = cpl_propertylist_get_double( plist,
                    SPH_IRD_KEYWORD_DMS_POS_X);
            dms_x /= SPH_IRD_PIX_SIZE;
        }
        else {
            SPH_ERROR_RAISE_WARNING(SPH_ERROR_GENERAL,
                    "Could not find the %s keyword. "
                    "Assuming value of 0.",
                    SPH_IRD_KEYWORD_DMS_POS_X);
        }
        if ( cpl_propertylist_has( plist,
                SPH_IRD_KEYWORD_DMS_POS_Y) ) {
            dms_y = cpl_propertylist_get_double( plist,
                    SPH_IRD_KEYWORD_DMS_POS_Y);
            dms_y /= SPH_IRD_PIX_SIZE;
        }
        else {
            SPH_ERROR_RAISE_WARNING(SPH_ERROR_GENERAL,
                    "Could not find the %s keyword. "
                    "Assuming value of 0.",
                    SPH_IRD_KEYWORD_DMS_POS_Y);
        }
        if ( sph_time_get_time_ut(
                self->current_frame,
                ut_time ) == CPL_ERROR_NONE)
        {
            cpl_table_set_string(self->interntab,
                    SPH_IRD_STAR_CENTER_TIME_UT_NAME,
                    ff, ut_time
                    );
            cpl_table_set_double(self->interntab,
                    SPH_IRD_STAR_CENTER_CENTLEFTX_NAME,
                    ff, clx
                    );
            cpl_table_set_double(self->interntab,
                    SPH_IRD_STAR_CENTER_CENTLEFTY_NAME,
                    ff, cly
                    );
            cpl_table_set_double(self->interntab,
                    SPH_IRD_STAR_CENTER_CENTRIGHTX_NAME,
                    ff, crx
                    );
            cpl_table_set_double(self->interntab,
                    SPH_IRD_STAR_CENTER_CENTRIGHTY_NAME,
                    ff, cry
                    );
            cpl_table_set_double(self->interntab,
                    SPH_IRD_STAR_CENTER_DMS_POS_X_NAME,
                    ff, dms_x
                    );
            cpl_table_set_double(self->interntab,
                    SPH_IRD_STAR_CENTER_DMS_POS_Y_NAME,
                    ff, dms_y
                    );
        }
        else {
            SPH_ERROR_RAISE_ERR(
                    CPL_ERROR_ILLEGAL_INPUT,
                    "Could not extract timing info from frame %s",
                    cpl_frame_get_filename(self->current_frame));
            goto EXIT;
        }
        SPH_ERROR_ENSURE_GOTO_EXIT( rerr == CPL_ERROR_NONE, rerr );

        SPH_ERROR_RAISE_INFO(
                SPH_ERROR_GENERAL,
                "Adding centre for %d as %f,%f and %f,%f.",
                ff,clx,cly,crx,cry);

        if ( self->qc_right ) {
            sph_transform_apply(
                    transform,right,
                    NULL, crx, cry, 0.0, 1.0,
                    self->model);
            sph_master_frame_add_master_frame(
                    self->qc_right,right);
        }

        sph_master_frame_delete(right);right = NULL;
        sph_master_frame_delete(left);left = NULL;
        cpl_propertylist_delete(plist); plist = NULL;
        ff++;
    }
    self->current_frame = NULL;
    sph_master_frame_delete(right);right = NULL;
    sph_master_frame_delete(left);left = NULL;
    cpl_frameset_delete(tmpframeset); tmpframeset  = NULL;
    sph_transform_delete(transform); transform = NULL;

EXIT:
    sph_master_frame_delete(right);right = NULL;
    sph_master_frame_delete(left);left = NULL;
    cpl_frameset_delete(tmpframeset); tmpframeset  = NULL;
    sph_transform_delete(transform); transform = NULL;
    cpl_propertylist_delete(plist); plist = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}



/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_star_center_run Create Master Dark Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * master dakr
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_star_center.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_star_center recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ird_star_center_run( sph_ird_star_center* self )
{
    cpl_table*              result = NULL;
    cpl_propertylist*            plist = NULL;
    int                         nwaffles = 0;
    char*                       tmpfilename = NULL;
    FILE*                              fp = NULL;
    cpl_frameset*               collapsed_waffles = NULL;
    cpl_mask*                   stat_badpix_mask = NULL;
    cpl_image*                  stat_badpix_image = NULL;

    /*------------------------------------------------------------------
    -  Placing master dark in cube structure for later use
    --------------------------------------------------------------------*/


    /*------------------------------------------------------------------
     -  Now process all sets of wavelength frames separately
     --------------------------------------------------------------------*/
//    if ( self->make_template )
//    	return sph_ird_star_center_make_template(self);

    nwaffles = sph_fits_get_nplanes(cpl_frame_get_filename(self->rawframe),0);
    cpl_ensure_code(nwaffles >=0, cpl_error_get_code());


    if ( self->static_badpixel_frame ) {
        stat_badpix_image = cpl_image_load(
                cpl_frame_get_filename(
                        self->static_badpixel_frame), CPL_TYPE_INT, 0, 0 );
        cpl_ensure_code(stat_badpix_image, cpl_error_get_code() );
        self->stat_badpix_mask = cpl_mask_threshold_image_create( stat_badpix_image, -0.5, 0.5);
        cpl_ensure_code(self->stat_badpix_mask, cpl_error_get_code() );
        cpl_mask_not(self->stat_badpix_mask);
        cpl_image_delete(stat_badpix_image); stat_badpix_image = NULL;
    }

    if ( self->dark_frame )
    {
        self->dark = sph_master_frame_load( cpl_frame_get_filename(self->dark_frame),0);
    }

    if ( self->flat_frame ) {
        self->flat = sph_master_frame_load( cpl_frame_get_filename(self->flat_frame),0);
        plist = sph_keyword_manager_load_properties(cpl_frame_get_filename(self->flat_frame),0);
        self->model = sph_ird_instrument_model_new_from_propertylist(plist);
    }
    if ( !self->model ) {
        self->model = sph_ird_instrument_model_new();
        plist = sph_ird_instrument_model_get_as_propertylist(self->model);
    }

    if ( self->qc ) {
        self->qc_left = sph_master_frame_new(
                self->model->window_size_y,
                self->model->window_size_y);
        self->qc_right = sph_master_frame_new(
                self->model->window_size_y,
                self->model->window_size_y);
    }
    self->tot_fctab = sph_fctable_new();


    SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
            "Using %d waffle images to create a field center"
            " information. ",
            nwaffles);


    collapsed_waffles = sph_ird_star_center_collapse_waffle_frames(
            self, self->rawframe);

    __sph_ird_star_center_init_inter_time_cent_tab( self, cpl_frameset_get_size(collapsed_waffles) );

    if ( cpl_error_get_code() == CPL_ERROR_NONE ) {
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL, "All waffle images collapsed.");
        if ( sph_ird_star_center_fill_waffle_table(
                self, collapsed_waffles) == CPL_ERROR_NONE )
        {
#ifdef SPHERE_DEBUG
            fp = fopen("waffle_table.txt","w");
            cpl_table_dump(self->interntab,0,cpl_table_get_nrow(self->interntab),fp);
            fclose(fp); fp = NULL;
#endif
            cpl_propertylist_update_string( plist,
                                SPH_COMMON_KEYWORD_PRO_CATG,
                                SPH_IRD_TAG_STAR_CENTER_CALIB );
            result = cpl_table_duplicate(self->interntab);
            SPH_RAISE_CPL;
            cpl_dfs_save_table(self->inframes,
                    NULL,
                    self->inparams,
                    self->inframes,
                    NULL,
                    result,
                    NULL,
                    SPH_RECIPE_NAME_IRD_STAR_CENTER,
                    plist,NULL,SPH_PIPELINE_NAME_IRDIS,
                    self->star_center_outfilename);
            SPH_RAISE_CPL;
            if ( self->qc_left ) {
                tmpfilename = sph_filemanager_new_filename_from_base(
                        self->star_center_outfilename, "qc_left");
                sph_master_frame_save_dfs(self->qc_left,
                        tmpfilename,
                        self->inframes,
                        self->rawframe,
                        self->inparams,
                        SPH_IRD_TAG_STAR_CENTER_QC,
                        SPH_RECIPE_NAME_IRD_STAR_CENTER,
                        SPH_PIPELINE_NAME_IRDIS,NULL);
                cpl_free(tmpfilename); tmpfilename = NULL;
            }
            if ( self->qc_right )
            {
                tmpfilename = sph_filemanager_new_filename_from_base(
                        self->star_center_outfilename,
                        "qc_right");
                sph_master_frame_save_dfs(
                        self->qc_right,
                        tmpfilename,
                        self->inframes,
                        self->rawframe,
                        self->inparams,
                        SPH_IRD_TAG_STAR_CENTER_QC,
                        SPH_RECIPE_NAME_IRD_STAR_CENTER,
                        SPH_PIPELINE_NAME_IRDIS,NULL);
                cpl_free(tmpfilename); tmpfilename = NULL;
            }
            cpl_table_delete(result); result = NULL;
        }
        else {
            SPH_ERROR_RAISE_ERR(SPH_ERROR_GENERAL,"Could not process science frame.");
        }
        if ( cpl_error_get_code() == CPL_ERROR_NONE ) {
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL, "All frames processed.");
        }
    }
    sph_master_frame_delete(self->dark); self->dark = NULL;
    sph_master_frame_delete(self->flat); self->flat = NULL;
    sph_master_frame_delete(self->qc_left); self->qc_left = NULL;
    sph_master_frame_delete(self->qc_right); self->qc_right = NULL;
    sph_fctable_delete(self->tot_fctab); self->tot_fctab = NULL;
    cpl_propertylist_delete(plist); plist = NULL;
    sph_ird_instrument_model_delete(self->model); self->model = NULL;
    cpl_table_delete(self->interntab); self->interntab = NULL;
    cpl_frameset_delete(collapsed_waffles); collapsed_waffles = NULL;
    cpl_mask_delete(stat_badpix_mask); stat_badpix_mask = NULL;
    cpl_mask_delete(self->stat_badpix_mask); self->stat_badpix_mask = NULL;
    return (int)cpl_error_get_code();

}

/**@}*/


