/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <cpl.h>
#include <math.h>
#include "sph_zpl_basic_polarization.h"
#include "sph_zpl_exposure.h"
#include "sph_quad_image.h"
#include "sph_double_image.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_utils.h"
#include "sph_zpl_utils.h"
#include "sph_common_keywords.h"
#include "sph_zpl_keywords.h"
#include "sph_zpl_tags.h"
#include "sph_filemanager.h"


/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_zpl_basic_polarization_run Basic Calibration Polarization Recipe
 *
 * This module provides the algorithm implementation for the basic (detector)
 * corrected images (polarimetric modes)
 *
 * @par Synopsis:
 * @code
 *   #include "sph_zpl_basic_polarization.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_zpl_basic_polarization recipe
 (polarimetric mode). The error  code returned is always a cpl error code
 (to allow maximal compatibility with  esorex, gasgano, etc.) even if during
 recipe execution an error in the SPHERE API is the cause. In this case
 (and if the underlying error is not a cpl error)  the cpl error code is set
 to the cpl_error_code that matches the failure reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/


static
cpl_frameset* sph_zpl_basic_polarization_create_basic( sph_zpl_basic_polarization* self, cpl_frame* inframe,
		sph_quad_image* master_bias_quadimage, sph_quad_image* master_dark_quadimage, sph_quad_image* intensity_flat_quadimage,
		sph_double_image* polarization_flat_doubleimage, sph_master_frame* modem_efficiency_masterframe )
{
	//sph_zpl_exposure*	curzplexp				= NULL;
	sph_quad_image*		curquadimage			= NULL;
	sph_double_image*	stokes					= NULL;
	cpl_frameset*		inframeset				= NULL;
    cpl_frameset*		ovsc_subtract_frames	= NULL;
	cpl_frame*			outframe				= NULL;
	cpl_frameset*		result					= NULL;
	cpl_frameset*		qiframeset				= NULL;
	cpl_frame*			qiframe					= NULL;
	cpl_frame*			dupinframe				= NULL;
	cpl_propertylist*	pl						= NULL;
	int					iplane					= 0;
	int					rerr					= CPL_ERROR_NONE;
	char				outfilename[256];
	char				outfilename_tmp[256];


    SPH_INFO_MSG("Starting sph_zpl_basic_polarization_create_basic...");

	if (!inframe ){
        SPH_ERR("no inframes to basic calibrate")
        return NULL;
	}

	// create inframeset and put there only one inframe
	// we do so, because interface for the transformation zplexp frames(cubes) into quad image frames(cubes)
	// uses cpl_frameset not cpl_frame
	dupinframe = cpl_frame_duplicate(inframe);
	inframeset = cpl_frameset_new();
	rerr = cpl_frameset_insert( inframeset,  dupinframe  );
	if ( rerr != CPL_ERROR_NONE ){
		SPH_ERR("Could not create inframeset");
		cpl_frameset_delete(inframeset); inframeset = NULL;
		return NULL;
	}

	if ( self->subtract_overscan ){
    	SPH_INFO_MSG("Subtract overscans from rawframes...")
       	ovsc_subtract_frames = sph_zpl_utils_subtract_overscans( inframeset );
    	if ( ovsc_subtract_frames ){
    		//create quad image from frameset(cubes) with subtracted overscans
    		qiframeset = sph_quad_image_create_quad_image_frameset_from_zplexp_cubes( ovsc_subtract_frames );
    		cpl_frameset_delete( ovsc_subtract_frames );
    	} else {
    		SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "Overscans could not be subtracted, trying to proceed further.")
			//create quad image frameset(cubes) from zplexp inframes (cubes)
			qiframeset = sph_quad_image_create_quad_image_frameset_from_zplexp_cubes( inframeset );
    	}
	} else {
		//create quad image frameset(cubes) from zplexp inframes (cubes)
		qiframeset = sph_quad_image_create_quad_image_frameset_from_zplexp_cubes( inframeset );
	}

	if ( !qiframeset ) {
		SPH_ERR("Failed creation of the quad image frameset(cubes) from zplexp inframes (cubes).");
		cpl_frameset_delete(inframeset); inframeset = NULL;
		return NULL;
	}
	qiframe = cpl_frameset_get_first( qiframeset );
	iplane = 0;
	result = cpl_frameset_new();
	while ( qiframe ){
		curquadimage = sph_quad_image_load( cpl_frame_get_filename( qiframe), 0);
		if ( !curquadimage ) {
			//if ( iplane == 0 ) {
			//	SPH_ERR("Could not load at least one plane from the qiframe");
			//	return NULL;
			//}
			SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO, "All planes from the current qiframe  calibrated or there was a problem by , "
					"the number of the iplane = %d", iplane)
			//break;
		}

		if ( master_bias_quadimage ) {
			rerr = sph_quad_image_subtract_quad_image( curquadimage, master_bias_quadimage);
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by subtracting master bias:\n"
						"cpl error code is: %d", cpl_error_get_code() );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}
		}


		if ( master_dark_quadimage ) {
			rerr = sph_quad_image_subtract_quad_image( curquadimage, master_dark_quadimage);
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by subtracting dark:\n"
						"cpl error code is: %d", cpl_error_get_code() );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}
		}


		if ( intensity_flat_quadimage ) {
			rerr = sph_quad_image_divide_quad_image( curquadimage, intensity_flat_quadimage);
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by applying (dividing) intensity flat:\n"
						"cpl error code is: %d", cpl_error_get_code() );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}
		}

	    SPH_INFO_MSG("Calculating stokes...");
		stokes = sph_zpl_utils_calculate_stokes_param_double_image_from_quad_image( curquadimage );

	    if (!stokes){
	    	SPH_ERR("stokes parameters is NULL");
			if (curquadimage) {
				sph_quad_image_delete( curquadimage );
				curquadimage = NULL;
			}
			if ( qiframeset ) {
				cpl_frameset_delete(qiframeset); qiframeset = NULL;
			}
			cpl_frameset_delete(inframeset); inframeset = NULL;
			return NULL;
	    }


	    if ( polarization_flat_doubleimage ){
	    	rerr = sph_zpl_utils_apply_polarization_flat( stokes, polarization_flat_doubleimage );
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by applying polarization flat:\n"
						"sph error code is: %d", sph_error_get_last_code() );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if (stokes) {
					sph_double_image_delete( stokes );
					stokes = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}

	    }

	    if ( modem_efficiency_masterframe ){
	    	rerr = sph_zpl_utils_apply_modem_efficiency( stokes, modem_efficiency_masterframe );
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by applying modem efficiency:\n"
						"sph error code is: %d", sph_error_get_last_code() );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if (stokes) {
					sph_double_image_delete( stokes );
					stokes = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}

	    }

	    //quality check (calculate statistical moments for the basic polarization)
	    rerr = sph_double_image_quality_check( stokes );
	    if ( rerr ){
	        sph_error_raise( SPH_ERROR_GENERAL,
	                __FILE__, __func__, __LINE__,
	                SPH_ERROR_ERROR,
	                "sph_double_image_quality_check returns cpl error.\n"
	                "cpl error code is: %d", rerr );
			if (curquadimage) {
				sph_quad_image_delete( curquadimage );
				curquadimage = NULL;
			}
			if (stokes) {
				sph_double_image_delete( stokes );
				stokes = NULL;
			}
			if ( qiframeset ) {
				cpl_frameset_delete(qiframeset); qiframeset = NULL;
			}
			cpl_frameset_delete(inframeset); inframeset = NULL;
	        return NULL;
	    }
	    //save
	    /* Attention:
	     * The following function includes a step to update SPH_COMMON_KEYWORD_SPH_TYPE by
	     * SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_DOUBLE_IMAGE. It is necessary because cpl_dsf_save_image
	     * updates the double image property list by the property list from the first input frames.
	     * In the case of some misbehavior exclude this step for checking.
	     */
		//create the name of the calibrated plane
	    SPH_INFO_MSG("Creating filename...");
	    sprintf( outfilename, "%s", self->basic_polarization_outfilename );
		sprintf( outfilename_tmp, "_%s_plane-", basename( (char*) cpl_frame_get_filename( inframe ) ));
		strcat( outfilename, outfilename_tmp );
		sprintf( outfilename_tmp, "%d.fits", iplane);
		strcat( outfilename, outfilename_tmp );
		SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "outfilename = %s", outfilename)
	    SPH_INFO_MSG("Saving...");
	    rerr = sph_double_image_save_dfs( stokes,
		                                  outfilename,
		                                  inframeset,
		                                  inframe,
		                                  self->inparams,
		                                  SPH_ZPL_TAG_BASIC_POL_CALIB,
		                                  SPH_RECIPE_NAME_ZPL_BASIC_POL,
		                                  SPH_PIPELINE_NAME_ZIMPOL,
		                                  pl);

	    if ( rerr ){
		       sph_error_raise( SPH_ERROR_GENERAL,
		              __FILE__, __func__, __LINE__,
		               SPH_ERROR_ERROR,
		               "Could not save basic polarization:\n"
		               "sph_double_image_save_dfs returns error code.\n"
		               "cpl error code is: %d", rerr );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if (stokes) {
					sph_double_image_delete( stokes );
					stokes = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
		       return NULL;
		}

	    outframe = cpl_frame_new();
	    cpl_frame_set_filename( outframe, outfilename );
	    cpl_frame_set_group( outframe, CPL_FRAME_GROUP_PRODUCT );
	    cpl_frame_set_tag( outframe, SPH_ZPL_TAG_BASIC_POL_CALIB );
	    rerr = cpl_frameset_insert( result, outframe);

	    if ( rerr ){
		       sph_error_raise( SPH_ERROR_GENERAL,
		              __FILE__, __func__, __LINE__,
		               SPH_ERROR_ERROR,
		               "Couldn't insert frame into the result frameset\n"
		               "cpl error code is: %d", rerr );
				if (curquadimage) {
					sph_quad_image_delete( curquadimage );
					curquadimage = NULL;
				}
				if (stokes) {
					sph_double_image_delete( stokes );
					stokes = NULL;
				}
				if ( qiframeset ) {
					cpl_frameset_delete(qiframeset); qiframeset = NULL;
				}
				if (result) {
			     	  cpl_frameset_delete(result); result = NULL;
			    }
				cpl_frameset_delete(inframeset); inframeset = NULL;
		       return NULL;
		}


	    sph_double_image_delete( stokes );
		sph_quad_image_delete( curquadimage ); curquadimage = NULL;
		qiframe = cpl_frameset_get_next( qiframeset) ;
		iplane = iplane + 1;
	}

	if ( qiframeset ) {
		cpl_frameset_delete(qiframeset); qiframeset = NULL;
	}
	if (inframeset) {
		cpl_frameset_delete(inframeset); inframeset = NULL;
	}
	if (result && cpl_frameset_get_size(result) == 0) {
    	cpl_frameset_delete(result); result = NULL;
    }

	return result;

}




cpl_error_code sph_zpl_basic_polarization_run( sph_zpl_basic_polarization* self )
{
    sph_quad_image*		master_bias_quadimage			= NULL;
    sph_quad_image*		master_dark_quadimage			= NULL;
    sph_quad_image*		intensity_flat_quadimage		= NULL;
    sph_double_image*	polarization_flat_doubleimage	= NULL;
    sph_master_frame*	modem_efficiency_masterframe	= NULL;


    cpl_frame*			curframe				= NULL;
    cpl_frameset*		basic_pol_frames		= NULL;

    int					recipe_error			= CPL_ERROR_NONE;


    SPH_INFO_MSG("Starting sph_zpl_basic_polarization_run...");
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE


    //if master bias exists load it
    if ( self->master_bias_frame ){
    	SPH_INFO_MSG("Loading master bias...\n");
        master_bias_quadimage = sph_quad_image_load( cpl_frame_get_filename( self->master_bias_frame ), 0);
    }

    //if master dark exists load it
    if ( self->master_dark_frame){
       	SPH_INFO_MSG("Loading master dark...\n");
        master_dark_quadimage = sph_quad_image_load( cpl_frame_get_filename( self->master_dark_frame ), 0);
    }

    //if intensity flat field exists load
    if ( self->intensity_flat_frame){
       	SPH_INFO_MSG("Loading intensity flat field...\n");
        intensity_flat_quadimage = sph_quad_image_load( cpl_frame_get_filename( self->intensity_flat_frame ), 0);
    }

    //if polarization flat field exists load it
    if ( self->polarization_flat_frame){
       	SPH_INFO_MSG("Loading polarization flat field...\n");
        polarization_flat_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->polarization_flat_frame ), 0);
    }

    //if modulation/demodulation efficiency exists load it
    if ( self->modem_efficiency_frame){
       	SPH_INFO_MSG("Loading modulation/demodulation efficiency...\n");
        modem_efficiency_masterframe = sph_master_frame_load( cpl_frame_get_filename( self->modem_efficiency_frame ), 0);
    }


    self->basic_pol_frameset = cpl_frameset_new();
    curframe = cpl_frameset_get_first( self->rawframes );
    while ( curframe ) {
        SPH_INFO_MSG("Invoke sph_zpl_basic_polarization_create_basic...");
        SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "curframe name = %s", cpl_frame_get_filename( curframe ));
        basic_pol_frames = sph_zpl_basic_polarization_create_basic( self, curframe,
        						master_bias_quadimage, master_dark_quadimage, intensity_flat_quadimage,
        						polarization_flat_doubleimage, modem_efficiency_masterframe );
       	recipe_error |= sph_utils_append_frames( self->basic_pol_frameset, basic_pol_frames );
       	cpl_frameset_delete(basic_pol_frames); basic_pol_frames = NULL;
        curframe = cpl_frameset_get_next( self->rawframes );
    }


    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "Could not create basic calibrated data:\n"
                "cpl error code is: %d", recipe_error );
    }


    if ( master_bias_quadimage ) {
    	sph_quad_image_delete( master_bias_quadimage ); master_bias_quadimage = NULL;
    }
    if ( master_dark_quadimage ) {
    	sph_quad_image_delete( master_dark_quadimage ); master_dark_quadimage = NULL;
    }
    if ( intensity_flat_quadimage ) {
    	sph_quad_image_delete( intensity_flat_quadimage ); intensity_flat_quadimage = NULL;
    }
    if ( polarization_flat_doubleimage ) {
    	sph_double_image_delete( polarization_flat_doubleimage ); polarization_flat_doubleimage = NULL;
    }
    if ( modem_efficiency_masterframe ) {
    	sph_master_frame_delete( modem_efficiency_masterframe ); modem_efficiency_masterframe = NULL;
    }


    SPH_INFO_MSG("Remove temporarily files: sph_filemanager_clean...");
    sph_filemanager_clean();

    SPH_INFO_MSG("sph_zpl_basic_polarization_run...End");
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}
/**@}*/
