/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include "sph_error.h"
#include "sph_master_frame.h"
#include "sph_framecombination.h"
#include "sph_fits.h"
#include "sph_common_science.h"
#include "sph_filemanager.h"
#include "sph_fits.h"
#include "sph_shared_irdifs.h"

static
cpl_frame*
sph_irdifs_master_dark_skip_copy( cpl_frame* frame, int nskip ) {
    cpl_frame*      result      = NULL;
    int             pp           = 0;
    int             nplanes     = 0;
    cpl_image*      tmpimage     = NULL;
    cpl_propertylist* pl         = NULL;
    const char*     tmpstr = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    cpl_ensure(frame,CPL_ERROR_NULL_INPUT,NULL);
    cpl_ensure(nskip >= 0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    result = cpl_frame_duplicate(frame);
    tmpstr = sph_filemanager_get_tmp_filename_const("DARK_SKIP");
    cpl_frame_set_filename(result,tmpstr);
    cpl_free((void*)tmpstr); tmpstr = NULL;
    tmpimage = cpl_image_load(cpl_frame_get_filename(frame),CPL_TYPE_INT,0,0);
    if ( !tmpimage ) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_FILE_IO,
                "Could not read plane %d in %s",
                0,cpl_frame_get_filename(frame));
        cpl_frame_delete(result); return NULL;
    }

    cpl_image_delete(tmpimage); tmpimage = NULL;

    sph_filemanager_add_tmp_file(cpl_frame_get_filename(result));

    nplanes = sph_fits_get_nplanes( cpl_frame_get_filename( frame ), 0 );
    pl = cpl_propertylist_load(cpl_frame_get_filename(frame),0);
    for (pp = nskip; pp < nplanes; ++pp) {
        tmpimage = cpl_image_load(cpl_frame_get_filename(frame),CPL_TYPE_INT,pp,0);
        if ( tmpimage ) {
            sph_cube_append_image(cpl_frame_get_filename(result),tmpimage,pl,0);
            cpl_image_delete(tmpimage); tmpimage = NULL;
        }
        else {
            SPH_ERROR_RAISE_ERR(CPL_ERROR_FILE_IO,
                    "Could not read plane %d in %s",
                    pp,cpl_frame_get_filename(frame));
            cpl_frame_delete(result); return NULL;
        }
    }
    cpl_propertylist_delete(pl); pl = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;

}
/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/

sph_master_frame*
sph_irdifs_master_dark_create( cpl_frameset* rawframes,
        sph_collapse_algorithm coll_alg,
        cpl_parameterlist* framecomb,
        double sigma_clip,
        double smoothing,
        double min_accept,
        double max_accept,
        int nskip,
        cpl_image** static_bpix)
{
    sph_master_frame*        result = NULL;
    sph_master_frame*        result_duplicate = NULL;
    sph_master_frame*        smooth_dark = NULL;
    int                      ff = 0;
    cpl_frame*               frame = NULL;
    cpl_frame*               nframe = NULL;
    cpl_frameset*            nframeset = NULL;
    int                      nplanes = 0;
    double                  rms = 0.0;
    /*------------------------------------------------------------------
      -  Sorting the frames into framesets of equal DIT setup
      --------------------------------------------------------------------*/

    if ( nskip == 0 ) {
        result = sph_framecombination_master_frame_from_cpl_frameset(
                rawframes,
                coll_alg,
                framecomb );
    }
    else {
        nframeset = cpl_frameset_new();
        for (ff = 0; ff < cpl_frameset_get_size(rawframes); ++ff) {
            frame = cpl_frameset_get_frame(rawframes,ff);
            if ( sph_fits_test_iscube(
            		cpl_frame_get_filename(frame), 0) == 1 ) {
                nplanes = sph_fits_get_nplanes(
                		cpl_frame_get_filename( frame ), 0 );
                if ( nplanes > nskip ) {
                    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                            "Copying cube in file %s, "
                            "skipping the first %d planes "
                            "so only %d planes are copied.",
                            cpl_frame_get_filename(frame),
                            nskip,nplanes-nskip);
                    nframe = sph_irdifs_master_dark_skip_copy(frame,nskip);
                    cpl_frameset_insert(nframeset,nframe);
                }
            }
        }
        result = sph_framecombination_master_frame_from_cpl_frameset(
                nframeset,
                coll_alg,
                framecomb );
        cpl_frameset_delete(nframeset); nframeset = NULL;
    }


    if ( result == NULL ) {
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "Could not create the dark -- "
                "Please check the previous error messages."
                "Error code was: %d", cpl_error_get_code() );

        return NULL;
    }

    sph_master_frame_mask_tolerance(result,min_accept,max_accept);

    result_duplicate = sph_master_frame_duplicate(result);
    sph_master_frame_quality_check(result);
    if ( smoothing > 0 ) {
        smooth_dark = sph_master_frame_duplicate(result);
        sph_master_frame_interpolate_bpix(smooth_dark);
        sph_master_frame_smooth(smooth_dark,smoothing);
        sph_master_frame_subtract_master_frame(result_duplicate,smooth_dark);
        sph_master_frame_delete(smooth_dark); smooth_dark = NULL;
    }
    sph_master_frame_quality_check(result_duplicate);
    sph_master_frame_get_mean(result_duplicate,&rms);
    sph_master_frame_mask_sigma(result_duplicate,sigma_clip);
    sph_master_frame_quality_check(result_duplicate);
    sph_master_frame_get_mean(result_duplicate,&rms);

    sph_master_frame_set_bads_from_image(result,result_duplicate->badpixelmap);
    sph_master_frame_delete(result_duplicate); result_duplicate = NULL;
    sph_master_frame_quality_check(result);

    if ( static_bpix ) {
        *static_bpix = sph_master_frame_get_badpixelmap(result);

        if ( !static_bpix ) {
            sph_error_raise( SPH_ERROR_ERROR,
                             __FILE__,
                             __func__,
                             __LINE__,
                             SPH_ERROR_ERROR,
                             "Could not find any badpixel map plane in the master dark." );
        }
    }

    return result;
}

/**@}*/
