/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include "sph_error.h"
#include "sph_fctable.h"
#include "sph_fits.h"
#include "sph_common_keywords.h"
#include "sph_keyword_manager.h"
#include "sph_filemanager.h"
#include "sph_time.h"
#include "sph_skycalc.h"
#include <math.h>
#include <strings.h>
#include <string.h>


/* Read one line from standard input, */
/* copying it to line array (but no more than max chars). */
/* Does not place terminating \n in line array. */
/* Returns line length, or 0 for empty line, or EOF for end-of-file. */
static
int __sph_fctable_getline(FILE* f,char line[], int max)
{
    int nch = 0;
    int c;
    max = max - 1;          /* leave room for '\0' */

    while((c = fgetc(f)) != EOF)
    {
        if(c == '\n')
            break;

        if(nch < max)
        {
            line[nch] = c;
            nch = nch + 1;
        }
    }

    if(c == EOF && nch == 0)
        return EOF;

    line[nch] = '\0';
    return nch;
}


/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides functionality for apertures, extending the functionality
 * as it exists for cpl_apertures.
 */
/*----------------------------------------------------------------------------*/
/**@{*/
static
void
__sph_fctable_increase_size(sph_fctable* self) {
    int n = 0;

    n = sph_fctable_get_size(self) + 1;

    cpl_vector_set_size(self->times,n+1);
    cpl_vector_set_size(self->centresL_x,n+1);
    cpl_vector_set_size(self->centresL_y,n+1);
    cpl_vector_set_size(self->angles,n+1);
    if ( self->mode == SPH_FCTAB_DOUBLE ) {
        cpl_vector_set_size(self->centresR_x,n+1);
        cpl_vector_set_size(self->centresR_y,n+1);
    }
    return;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Create a new table of field centres and angles
 * @param
 *
 * @return new pointer
 *
 * Creates a new table of field centres and angles.
 *
 */
/*----------------------------------------------------------------------------*/
sph_fctable*  sph_fctable_new(void) {
    sph_fctable*        fctab = NULL;

    fctab = calloc(1,sizeof(sph_fctable));
    fctab->angles = cpl_vector_new(1);
    fctab->times = cpl_vector_new(1);
    fctab->angles_err = cpl_vector_new(1);
    fctab->centresL_x = cpl_vector_new(1);
    fctab->centresL_x_err = cpl_vector_new(1);
    fctab->centresL_y = cpl_vector_new(1);
    fctab->centresL_y_err = cpl_vector_new(1);
    fctab->centresR_x = cpl_vector_new(1);
    fctab->centresR_x_err = cpl_vector_new(1);
    fctab->centresR_y = cpl_vector_new(1);
    fctab->centresR_y_err = cpl_vector_new(1);
    fctab->mode = SPH_FCTAB_UNDEFINED;
    return fctab;
}

int sph_fctable_get_size(sph_fctable* self) {
    cpl_ensure(self,CPL_ERROR_ILLEGAL_INPUT,-1);
    return cpl_vector_get_size(self->centresL_x)-1;
}

sph_error_code
sph_fctable_add_row_single(sph_fctable* self,
        double mjd,
        double cx, double cy, double angle) {
    int         n = 0;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(self->mode == SPH_FCTAB_SINGLE ||
            self->mode == SPH_FCTAB_UNDEFINED,
            CPL_ERROR_ILLEGAL_INPUT );
    n = sph_fctable_get_size(self);
    cpl_vector_set(self->times,n,mjd);
    cpl_vector_set(self->angles,n,angle);
    cpl_vector_set(self->centresL_x,n,cx);
    cpl_vector_set(self->centresL_y,n,cy);
    self->mode = SPH_FCTAB_SINGLE;
    __sph_fctable_increase_size(self);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

sph_error_code
sph_fctable_add_row_double(
        sph_fctable* self,
        double mjd,
        double clx, double cly,
        double crx, double cry, double angle) {
    int         n = 0;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(self->mode == SPH_FCTAB_DOUBLE ||
            self->mode == SPH_FCTAB_UNDEFINED,
            CPL_ERROR_ILLEGAL_INPUT );
    n = sph_fctable_get_size(self);
    cpl_vector_set(self->times,n,mjd);
    cpl_vector_set(self->angles,n,angle);
    cpl_vector_set(self->centresL_x,n,clx);
    cpl_vector_set(self->centresL_y,n,cly);
    cpl_vector_set(self->centresR_x,n,crx);
    cpl_vector_set(self->centresR_y,n,cry);
    self->mode = SPH_FCTAB_DOUBLE;
    __sph_fctable_increase_size(self);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

char*
sph_fctable_construct_filename( cpl_frame* inframe, int txtflag ) {
    char*                       tmpfilename = NULL;
    char*                       tmpfilename2 = NULL;

    cpl_ensure(inframe,CPL_ERROR_NULL_INPUT,NULL);
    tmpfilename2 = sph_filemanager_new_filename_from_base(
             cpl_frame_get_filename(inframe), "fctable");
     tmpfilename = sph_filemanager_remove_dir(tmpfilename2);
     cpl_free(tmpfilename2); tmpfilename2 = NULL;
     if (txtflag) {
         tmpfilename2 = sph_filemanager_filename_new_ext(
                 tmpfilename, "txt");
         cpl_free(tmpfilename); tmpfilename = NULL;
         return tmpfilename2;
     }
     return tmpfilename;
}

sph_error_code
sph_fctable_save_ascii( sph_fctable* self,const char* filename ) {
    FILE*       f = NULL;
    char        line[1024];
    int         n = 0;
    int         ii = 0;
    char        timechars[256];

    cpl_ensure_code( filename, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT);

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    f = fopen( filename, "w" );
    cpl_ensure_code( f, CPL_ERROR_FILE_IO);
    n = sph_fctable_get_size(self);
    cpl_ensure_code( n > 0, CPL_ERROR_ILLEGAL_INPUT);
    fprintf(f,"# Field centre and angles table\n");
    fprintf(f,"# \n");
    fprintf(f,"# This table gives the field centres and rotation angles\n");
    fprintf(f,"# for a set of frames.\n");
    fprintf(f,"# Format is: centres in pixel coordinates (0,0) at lower left\n");
    fprintf(f,"#            angles in degrees ccw wrt. x-axis.\n");
    fprintf(f,"# \n");
    if ( self->mode == SPH_FCTAB_SINGLE ) {
        fprintf(f,"#  Time-UT   Centre x  Centre y   Angle (deg)\n");
    }
    else {
        fprintf(f,"# Time-UT   Centre Left x    Left y   Right x   Right y   Angle (deg)\n");
    }
    for (ii = 0; ii < n; ++ii) {
        sph_time_mjd_to_iso8601string(
                cpl_vector_get(
                        self->times,ii), timechars);
        if ( self->mode == SPH_FCTAB_SINGLE ) {
            sprintf(line,"%s %9.2f %9.2f %12.6f",
                    timechars,
                    cpl_vector_get(self->centresL_x,ii),
                    cpl_vector_get(self->centresL_y,ii),
                    cpl_vector_get(self->angles,ii) );
        }
        else if ( self->mode == SPH_FCTAB_DOUBLE ) {
            sprintf(line,"%s  %9.2f %9.2f %9.2f %9.2f %12.6f",
                    timechars,
                    cpl_vector_get(self->centresL_x,ii),
                    cpl_vector_get(self->centresL_y,ii),
                    cpl_vector_get(self->centresR_x,ii),
                    cpl_vector_get(self->centresR_y,ii),
                    cpl_vector_get(self->angles,ii) );
        }
        fprintf(f,"%s\n",line);
    }
    fclose(f); f = NULL;

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

cpl_table*
sph_fctable_get_as_cpl_table(sph_fctable* self) {
    cpl_table*      result          = NULL;
    int             n                = 0;
    int             ii               = 0;
    char        timechars[256];

    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n > 0, CPL_ERROR_ILLEGAL_INPUT,NULL);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    result = cpl_table_new(n);
    cpl_ensure(result,cpl_error_get_code(),NULL);
    if ( self->mode == SPH_FCTAB_DOUBLE ) {
        cpl_table_new_column(result,"TIME-UT",CPL_TYPE_STRING);
        cpl_table_new_column(result,"CENTRE LEFT X",CPL_TYPE_DOUBLE);
        cpl_table_new_column(result,"CENTRE LEFT Y",CPL_TYPE_DOUBLE);
        cpl_table_new_column(result,"CENTRE RIGHT X",CPL_TYPE_DOUBLE);
        cpl_table_new_column(result,"CENTRE RIGHT Y",CPL_TYPE_DOUBLE);
    }
    else if ( self->mode == SPH_FCTAB_SINGLE ) {
        cpl_table_new_column(result,"TIME-UT",CPL_TYPE_STRING);
        cpl_table_new_column(result,"CENTRE X",CPL_TYPE_DOUBLE);
        cpl_table_new_column(result,"CENTRE Y",CPL_TYPE_DOUBLE);
    }
    cpl_table_new_column(result,"ANGLE",CPL_TYPE_DOUBLE);
    for (ii = 0; ii < n; ++ii) {
        sph_time_mjd_to_iso8601string(
                cpl_vector_get(
                        self->times,ii), timechars);
        cpl_table_set_string(result,"TIME-UT",ii,timechars);
        if ( self->mode == SPH_FCTAB_DOUBLE ) {
            cpl_table_set_double(result,"CENTRE LEFT X",ii,cpl_vector_get(self->centresL_x,ii));
            cpl_table_set_double(result,"CENTRE LEFT Y",ii,cpl_vector_get(self->centresL_y,ii));
            cpl_table_set_double(result,"CENTRE RIGHT X",ii,cpl_vector_get(self->centresR_x,ii));
            cpl_table_set_double(result,"CENTRE RIGHT Y",ii,cpl_vector_get(self->centresR_y,ii));
        }
        else if ( self->mode == SPH_FCTAB_SINGLE ) {
            cpl_table_set_double(result,"CENTRE X",ii,cpl_vector_get(self->centresL_x,ii));
            cpl_table_set_double(result,"CENTRE Y",ii,cpl_vector_get(self->centresL_y,ii));
        }
        cpl_table_set_double(result,"ANGLE",ii,cpl_vector_get(self->angles,ii));
    }
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}


sph_fctable*
sph_fctable_load_ascii(const char* filename) {
    FILE*       f = NULL;
    char       line[256];
    double      clx = 0.0;
    double      cly = 0.0;
    double      crx = 0.0;
    double      cry = 0.0;
    double      angle = 0.0;
    int         nmatches = 0;
    int         linesize = 0;
    int         year = 0;
    int         month = 0;
    int         day = 0;
    int         hour = 0;
    int         minute   =   0;
    double      second      = 0.0;
    double      mjd = 0.0;
    __sph_fctab_mode    mm = SPH_FCTAB_UNDEFINED;
    sph_fctable*    self = NULL;
    cpl_ensure(filename,CPL_ERROR_NULL_INPUT,NULL);
    f = fopen(filename,"r");
    cpl_ensure(f,CPL_ERROR_FILE_NOT_FOUND,NULL);
    self = sph_fctable_new();
    while ( (linesize = __sph_fctable_getline(f,line,256)) != EOF ) {
        if ( line[0]== '#' || linesize == 0 ) {
            ;
        }
        else {
            nmatches = sscanf(line,
                    "%4d-%2d-%2dT%2d:%2d:%lf %lf %lf %lf %lf %lf",
                    &year,
                    &month,
                    &day,
                    &hour,
                    &minute,
                    &second,
                    &clx,&cly,&crx,&cry,&angle);
            if ( nmatches == 9 ) {
                sph_time_mjd_from_iso8601(
                        &mjd,year,month,day,hour,minute,second);
                if ( mm == SPH_FCTAB_UNDEFINED ||
                        mm == SPH_FCTAB_SINGLE ) {
                    mm = SPH_FCTAB_SINGLE;
                    angle = crx;
                    sph_fctable_add_row_single(self,
                            mjd,clx,cly,angle);
                }
                else {
                    SPH_ERROR_ENSURE_GOTO_EXIT(0,CPL_ERROR_BAD_FILE_FORMAT);
                }
            }
            else if ( nmatches == 11 ) {
                sph_time_mjd_from_iso8601(
                        &mjd,year,month,day,hour,minute,second);
                if ( mm == SPH_FCTAB_UNDEFINED ||
                        mm == SPH_FCTAB_DOUBLE ) {
                    mm = SPH_FCTAB_DOUBLE;
                    sph_fctable_add_row_double(self,
                            mjd,clx,cly,crx,cry,angle);
                }
                else {
                    SPH_ERROR_ENSURE_GOTO_EXIT(0,CPL_ERROR_BAD_FILE_FORMAT);
                }
            }
            else {
                SPH_ERROR_ENSURE_GOTO_EXIT(0,CPL_ERROR_BAD_FILE_FORMAT);
            }
        }
    }
    fclose(f); f = NULL;

    return self;
EXIT:
    sph_fctable_delete(self); self = NULL;
    return NULL;
}

sph_fctable*
sph_fctable_load_fits(const char* filename) {
    sph_fctable*    self    = NULL;
    cpl_table*      tab     = NULL;
    int             next    = 0;
    int             ee      = 0;
    int             rr      = 0;
    int             bp      = 0;
    double          mjd     = 0.0;

    cpl_ensure(filename,CPL_ERROR_NULL_INPUT,NULL);

    next = cpl_fits_count_extensions(filename);
    for (ee = 0; ee < next + 1; ++ee) {
        tab = cpl_table_load( filename, ee, 0 );
        if ( tab == NULL ) {
            cpl_error_reset();
        }
        else {
            break;
        }
    }

    cpl_ensure(tab,CPL_ERROR_FILE_IO,NULL);
    self = sph_fctable_new();
    if ( strcasecmp(cpl_table_get_column_name(tab),
                "TIME-UT") != 0 ) {
        sph_fctable_delete(self);self = NULL;
        cpl_ensure(self,CPL_ERROR_BAD_FILE_FORMAT,NULL);
    }
    if ( strcasecmp(cpl_table_get_column_name(NULL),
            "CENTRE LEFT X") == 0 ) {
        self->mode = SPH_FCTAB_DOUBLE;
    }
    else if ( strcasecmp(cpl_table_get_column_name(NULL),
                "CENTRE Y") == 0 ) {
        self->mode = SPH_FCTAB_SINGLE;
    }
    else {
        sph_fctable_delete(self);self = NULL;
        cpl_ensure(self,CPL_ERROR_BAD_FILE_FORMAT,NULL);
    }

    for (rr = 0; rr < cpl_table_get_nrow(tab); ++rr) {
        if ( self->mode == SPH_FCTAB_DOUBLE ) {
            sph_time_mjd_from_string(&mjd,
                    cpl_table_get_string(tab,"TIME-UT",rr));

            sph_fctable_add_row_double(self,mjd,
                    cpl_table_get_double(tab,"CENTRE LEFT X",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE LEFT Y",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE RIGHT X",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE RIGHT Y",rr,&bp),
                    cpl_table_get_double(tab,"ANGLE",rr,&bp) );
        }
        else {
            sph_time_mjd_from_string(&mjd,
                    cpl_table_get_string(tab,"TIME-UT",rr));
            sph_fctable_add_row_single(self,mjd,
                    cpl_table_get_double(tab,"CENTRE X",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE Y",rr,&bp),
                    cpl_table_get_double(tab,"ANGLE",rr,&bp) );
        }
    }
    cpl_table_delete(tab); tab = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return self;
}

sph_fctable*
sph_fctable_load_fits_IRDIS(const char* filename) {
    sph_fctable*    self    = NULL;
    cpl_table*      tab     = NULL;
    int             next    = 0;
    int             ee      = 0;
    int             rr      = 0;
    int             bp      = 0;
    double          mjd     = 0.0;

    cpl_ensure(filename,CPL_ERROR_NULL_INPUT,NULL);

    next = cpl_fits_count_extensions(filename);
    for (ee = 0; ee < next + 1; ++ee) {
        tab = cpl_table_load( filename, ee, 0 );
        if ( tab == NULL ) {
            cpl_error_reset();
        }
        else {
            break;
        }
    }

    cpl_ensure(tab,CPL_ERROR_FILE_IO,NULL);
    self = sph_fctable_new();
    if ( strcasecmp(cpl_table_get_column_name(tab),
                "TIME-UT") != 0 ) {
        sph_fctable_delete(self);self = NULL;
        cpl_ensure(self,CPL_ERROR_BAD_FILE_FORMAT,NULL);
    }
    if ( strcasecmp(cpl_table_get_column_name(NULL),
            "CENTRE LEFT X") == 0 ) {
        self->mode = SPH_FCTAB_DOUBLE;
    }
    else if ( strcasecmp(cpl_table_get_column_name(NULL),
                "CENTRE Y") == 0 ) {
        self->mode = SPH_FCTAB_SINGLE;
    }
    else {
        sph_fctable_delete(self);self = NULL;
        cpl_ensure(self,CPL_ERROR_BAD_FILE_FORMAT,NULL);
    }

    for (rr = 0; rr < cpl_table_get_nrow(tab); ++rr) {
        if ( self->mode == SPH_FCTAB_DOUBLE ) {
            sph_time_mjd_from_string(&mjd,
                    cpl_table_get_string(tab,"TIME-UT",rr));

            sph_fctable_add_row_double(self,mjd,
                    cpl_table_get_double(tab,"CENTRE LEFT X",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE LEFT Y",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE RIGHT X",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE RIGHT Y",rr,&bp), 0.0);
        }
        else {
            sph_time_mjd_from_string(&mjd,
                    cpl_table_get_string(tab,"TIME-UT",rr));
            sph_fctable_add_row_single(self,mjd,
                    cpl_table_get_double(tab,"CENTRE X",rr,&bp),
                    cpl_table_get_double(tab,"CENTRE Y",rr,&bp), 0.0);
        }
    }
    cpl_table_delete(tab); tab = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return self;
}


cpl_vector*
sph_fctable_get_times(sph_fctable* self) {
    cpl_vector* vector = NULL;
    int         n = 0;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n>0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    vector = cpl_vector_duplicate(self->times);
    cpl_vector_set_size(vector,n);
    return vector;
}

cpl_vector*
sph_fctable_get_angles(sph_fctable* self) {
    cpl_vector* vector = NULL;
    int         n = 0;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n>0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    vector = cpl_vector_duplicate(self->angles);
    cpl_vector_set_size(vector,n);
    return vector;
}


double
sph_fctable_get_time(sph_fctable* self, int idx) {
    cpl_ensure(self, CPL_ERROR_NULL_INPUT, -1.0);
	return cpl_vector_get(self->times, idx);
}

double
sph_fctable_get_angle(sph_fctable* self, int idx) {
    cpl_ensure(self, CPL_ERROR_NULL_INPUT, -1.0);
	return cpl_vector_get(self->angles, idx);
}

double sph_fctable_get_centre_x_element(sph_fctable* self, int idx){
	return sph_fctable_get_centre_x_left_element(self, idx);
}

double
sph_fctable_get_centre_y_element(sph_fctable* self, int idx){
	return sph_fctable_get_centre_y_left_element(self, idx);
}

cpl_vector* sph_fctable_get_centre_x(sph_fctable* self){
	return sph_fctable_get_centre_x_left(self);
}

cpl_vector* sph_fctable_get_centre_y(sph_fctable* self){
	return sph_fctable_get_centre_y_left(self);
}

double
sph_fctable_get_centre_x_left_element(sph_fctable* self, int idx){
    cpl_ensure(self, CPL_ERROR_NULL_INPUT, -1.0);
    return cpl_vector_get(self->centresL_x, idx);
}

double
sph_fctable_get_centre_y_left_element(sph_fctable* self, int idx){
    cpl_ensure(self, CPL_ERROR_NULL_INPUT, -1.0);
    return cpl_vector_get(self->centresL_y, idx);
}

cpl_vector*
sph_fctable_get_centre_x_left(sph_fctable* self) {
    cpl_vector* vector = NULL;
    int         n = 0;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n>0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    vector = cpl_vector_duplicate(self->centresL_x);
    cpl_vector_set_size(vector,n);
    return vector;
}

cpl_vector*
sph_fctable_get_centre_y_left(sph_fctable* self) {
    cpl_vector* vector = NULL;
    int         n = 0;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n>0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    vector = cpl_vector_duplicate(self->centresL_y);
    cpl_vector_set_size(vector,n);
    return vector;
}

cpl_vector*
sph_fctable_get_centre_x_right(sph_fctable* self) {
    cpl_vector* vector = NULL;
    int         n = 0;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n > 0 ,CPL_ERROR_ILLEGAL_INPUT,NULL);
    if ( self->mode != SPH_FCTAB_DOUBLE ) return NULL;
    vector = cpl_vector_duplicate(self->centresR_x);
    cpl_vector_set_size(vector,n);
    return vector;
}

cpl_vector*
sph_fctable_get_centre_y_right(sph_fctable* self) {
    cpl_vector* vector = NULL;
    int         n = 0;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    n = sph_fctable_get_size(self);
    cpl_ensure(n>0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    if ( self->mode != SPH_FCTAB_DOUBLE ) return NULL;
    vector = cpl_vector_duplicate(self->centresR_y);
    cpl_vector_set_size(vector,n);
    return vector;
}

void sph_fctable_delete(sph_fctable* self) {
    if ( self ) {
        cpl_vector_delete(self->times);
        cpl_vector_delete(self->angles);
        cpl_vector_delete(self->angles_err);
        cpl_vector_delete(self->centresL_x);
        cpl_vector_delete(self->centresL_x_err);
        cpl_vector_delete(self->centresL_y);
        cpl_vector_delete(self->centresL_y_err);
        cpl_vector_delete(self->centresR_x);
        cpl_vector_delete(self->centresR_x_err);
        cpl_vector_delete(self->centresR_y);
        cpl_vector_delete(self->centresR_y_err);
        free(self);
    }
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Set the center information for a specific row
 * @param self 		pointer to self
 * @param row	    the row number
 * @param clx0	x coordinate of the center of the left image of the frames
 * @param cly0	y coordinate of the center of the left image of the frames
 * @param crx0	x coordinate of the center of the right image of the frames
 * @param cry0	y coordinate of the center of the right image of the frames
 *
 *
 * @return error code
 *
 * Set the centre information for a specific row in the fctable
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fctable_set_cent_double(
		sph_fctable* self,
		int row,
		double clx, double cly, double crx, double cry)
{

	cpl_ensure_code( self, CPL_ERROR_NULL_INPUT);
	cpl_ensure_code( row >=0, CPL_ERROR_ILLEGAL_INPUT);
	cpl_ensure_code(
			cpl_vector_set( self->centresL_x, row, clx) == CPL_ERROR_NONE,
			cpl_error_get_code() );
	cpl_ensure_code(
			cpl_vector_set( self->centresL_y, row, cly) == CPL_ERROR_NONE,
			cpl_error_get_code() );
	cpl_ensure_code(
			cpl_vector_set( self->centresR_x, row, crx) == CPL_ERROR_NONE,
			cpl_error_get_code() );
	cpl_ensure_code(
			cpl_vector_set( self->centresR_y, row, cry) == CPL_ERROR_NONE,
			cpl_error_get_code() );
	return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Create a new field centre table
 * @param raw_science_frames the list of raw science frames
 * @param field_center_frame a field centre table input frame (maybe NULL)
 * @param use_dither   flag to control if dithering info should be used
 * @param clx0	x coordinate of the center of the left image of the frames
 * @param cly0	y coordinate of the center of the left image of the frames
 * @param crx0	x coordinate of the center of the right image of the frames
 * @param cry0	y coordinate of the center of the right image of the frames
 *
 *
 * @return new field centre table or NULL on error.
 *
 * Create a field centre table. This may either be a straight reading in
 * of the field centre table from the field_center frame if one is provided
 * or a default table is constructed for the input list of raw frames.
 * In the latter case, if the use_dither flag is not set, the table returned
 * is simply filled with the pixel centers as determined from the instrument
 * model (the central pixel for each frame would usually then be at 512 in x and y).
 * If the use_dither flag is set to 1, the dithering is taken into
 * account when constructing the default table.
 *
 */
/*----------------------------------------------------------------------------*/
sph_fctable*
sph_fctable_create_fctable( cpl_frame* raw_science_frame, int mode )
{
    sph_error_code        rerr 	  = CPL_ERROR_NONE;
    sph_fctable*          fctab   = NULL;
    sph_skycalc_input	  skyinp[7];
    sph_skycalc_output	  skyout;
    cpl_propertylist*	  pl 	  = NULL;
    cpl_image*			  image	  = NULL;
    char  				  ra[20];
    char				  dec[20];
    int					  epoch	  = 2000;
    double                crx 	  = 0.0;
    double                cry	  = 0.0;
    double                clx 	  = 0.0;
    double                cly 	  = 0.0;
    double				  xdith	  = 0.0;
    double				  ydith	  = 0.0;
    int                   ff 	  = 0;
    cpl_vector*           timev   = NULL;
    int                   nplanes = 0;


    cpl_ensure( raw_science_frame, CPL_ERROR_NULL_INPUT, NULL );


    snprintf( ra, sizeof(ra), "NONE");
    snprintf( dec, sizeof(ra), "NONE");

    fctab = sph_fctable_new();
    if ( !fctab ) {
        SPH_ERR("Could not create centre positions table.");
        return NULL;
    }

    if ( mode == SPH_FCTAB_SINGLE || mode == SPH_FCTAB_DOUBLE ){
    	fctab->mode = mode;
    } else {
    	fctab->mode = SPH_FCTAB_SINGLE;
    }

    nplanes = sph_fits_get_nplanes( cpl_frame_get_filename(
            raw_science_frame), 0 );

    timev = sph_time_extract_times_from_frame(
    		raw_science_frame );

    //SPH_ERROR_ENSURE_GOTO_EXIT( timev != NULL, rerr );
    if ( timev == NULL ) {
    	timev = cpl_vector_new( nplanes );
        cpl_vector_fill( timev, 0.0);
    }

    if ( fctab->mode == SPH_FCTAB_SINGLE ) {
    	SPH_INFO_MSG("fctab->mode == SPH_FCTAB_SINGLE");
    	image = cpl_image_load(cpl_frame_get_filename( raw_science_frame ), CPL_TYPE_DOUBLE, 0, 0  );
    	if ( image != NULL ){
    		clx = (double) (cpl_image_get_size_x( image ))/ 2.0;
    		cly = (double) (cpl_image_get_size_y( image ));
    		cpl_image_delete( image );
    	} else {
    		clx = 0.0; cly = 0.0;
    	}
    } else if (fctab->mode == SPH_FCTAB_DOUBLE ) {
    	clx = 0.0; cly = 0.0;
    	crx = 0.0; cry = 0.0;
    } else {

    }

    //pl = cpl_propertylist_load( cpl_frame_get_filename(raw_science_frame), 0);
    pl = sph_keyword_manager_load_properties(cpl_frame_get_filename(raw_science_frame), 0);

    if ( pl ) {

      //dithering
      if ( cpl_propertylist_has( pl, SPH_COMMOM_KEYWORD_CHIP_INDEX) ) {

    	  if ( cpl_propertylist_get_int( pl, SPH_COMMOM_KEYWORD_CHIP_INDEX )
    			  == SPH_COMMON_KEYWORD_VALUE_CHIP_INDEX_CAM1 ) {
    		  //cam1
    		  if ( cpl_propertylist_has ( pl, SPH_COMMON_KEYWORD_CAM1_DITHERING_X) ){
    			  xdith = cpl_propertylist_get_double( pl, SPH_COMMON_KEYWORD_CAM1_DITHERING_X );
    		  } else {
       		   SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword (which sets up xdith for cam1) not found!\n",SPH_COMMON_KEYWORD_CAM1_DITHERING_X );
    		  }
    		  if ( cpl_propertylist_has ( pl, SPH_COMMON_KEYWORD_CAM1_DITHERING_Y) ){
    			  ydith = cpl_propertylist_get_double( pl, SPH_COMMON_KEYWORD_CAM1_DITHERING_Y );
    		  } else {
       		   SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword (which sets up ydith for cam1) not found!\n",SPH_COMMON_KEYWORD_CAM1_DITHERING_Y );
    		  }
    	  } else if ( cpl_propertylist_get_int( pl, SPH_COMMOM_KEYWORD_CHIP_INDEX )
        			  == SPH_COMMON_KEYWORD_VALUE_CHIP_INDEX_CAM2 ) {
    		  //cam2
    		  if ( cpl_propertylist_has ( pl, SPH_COMMON_KEYWORD_CAM2_DITHERING_X) ){
    			  xdith = cpl_propertylist_get_double( pl, SPH_COMMON_KEYWORD_CAM2_DITHERING_X );
    		  } else {
       		   SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword (which sets up xdith for cam2) not found!\n",SPH_COMMON_KEYWORD_CAM2_DITHERING_X );
    		  }
    		  if ( cpl_propertylist_has ( pl, SPH_COMMON_KEYWORD_CAM2_DITHERING_Y) ){
    			  ydith = cpl_propertylist_get_double( pl, SPH_COMMON_KEYWORD_CAM2_DITHERING_Y );
    		  } else {
       		   SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword (which sets up ydith for cam2) not found!\n",SPH_COMMON_KEYWORD_CAM2_DITHERING_Y );
    		  }

    	  } else {
      		   SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "Chip index has a worng value: %d (it must be 1 or 2)!\n",  cpl_propertylist_get_int( pl, SPH_COMMOM_KEYWORD_CHIP_INDEX) );
    	  }
      }


      if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_DROT2_MODE ) ) {

        if ( strcmp( cpl_propertylist_get_string( pl, SPH_COMMON_KEYWORD_DROT2_MODE ),
    		  SPH_COMMON_KEYWORD_VALUE_DROT2_MODE_PUPIL_STAB ) == 0) {
           SPH_INFO_MSG("Pupil stabilazed mode: extracting RA, DEC & Epoch to calculate PA")
    	   //get ra, dec, epoch
    	   if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_RA ) ) {
    		   snprintf( ra, sizeof(ra), cpl_propertylist_get_string( pl, SPH_COMMON_KEYWORD_RA ));
    	   } else {
    		   SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword not found!\n",SPH_COMMON_KEYWORD_RA );
    	   }
    		if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_DEC ) ) {
    			snprintf( dec, sizeof(dec), cpl_propertylist_get_string( pl, SPH_COMMON_KEYWORD_DEC ) );
    		} else {
    			SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword not found!\n",SPH_COMMON_KEYWORD_DEC );
    		}
    		if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_EPOCH ) ) {
    			epoch = (int) cpl_propertylist_get_double( pl, SPH_COMMON_KEYWORD_EPOCH );
    		} else {
    			SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword not found!",SPH_COMMON_KEYWORD_EPOCH );
    			SPH_INFO_MSG("Set default EPOCH = 2000");
    		}
        }
      } else {
    	  SPH_INFO_MSG("SPH_COMMON_KEYWORD_DROT2_MODE is not presented in the header,"
    			  " consider the mode as a pupil stabilized!");

    	 if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_RA ) ) {
   		   snprintf( ra, sizeof(ra), cpl_propertylist_get_string( pl, SPH_COMMON_KEYWORD_RA ));
    	 } else {
    		 SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword not found!\n",SPH_COMMON_KEYWORD_RA );
    	 }
    	 if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_DEC ) ) {
   			snprintf( dec, sizeof(dec), cpl_propertylist_get_string( pl, SPH_COMMON_KEYWORD_DEC ) );
    	 } else {
    		 SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword not found!\n",SPH_COMMON_KEYWORD_DEC );
    	 }
    	 if ( cpl_propertylist_has( pl, SPH_COMMON_KEYWORD_EPOCH ) ) {
    		 epoch = (int) cpl_propertylist_get_double( pl, SPH_COMMON_KEYWORD_EPOCH );
    	 } else {
    		 SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "%s keyword not found!",SPH_COMMON_KEYWORD_EPOCH );
    		 SPH_INFO_MSG("Set default EPOCH = 2000");
    	 }
      }
    }
    cpl_propertylist_delete(pl); pl = NULL;


    //setup sph_skycalc and create first raw
    ff = 0;
    if ( strcmp( ra, "NONE") != 0 && strcmp( dec, "NONE") != 0 ) {
    	sph_skycalc_setup_mjd( skyinp, ra, dec, epoch, cpl_vector_get(timev,ff) );
    	SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "RA=%c, DEC=%c, EPOC=%c, TIMEZ=%c, DATE=%c, TIME=%c, EQUAL=%c",
    			skyinp[RA].command, skyinp[DEC].command,
    			skyinp[EPOCH].command, skyinp[TIMEZ].command, skyinp[DATE].command, skyinp[TIME].command,
    			skyinp[EQUAL].command);
    	sph_skycalc( skyinp, &skyout );
    	SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "Time = %f, PA = %f", cpl_vector_get(timev,ff), skyout.pa );
        SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "RA=%c, DEC=%c, EPOC=%c, TIMEZ=%c, DATE=%c, TIME=%c, EQUAL=%c",
                        skyinp[RA].command, skyinp[DEC].command,
                        skyinp[EPOCH].command, skyinp[TIMEZ].command, skyinp[DATE].command, skyinp[TIME].command,
                        skyinp[EQUAL].command);
 
    } else {
    	skyout.pa = 0.0;
    }

    if ( fctab->mode == SPH_FCTAB_SINGLE ){
   		rerr = sph_fctable_add_row_single( fctab, cpl_vector_get( timev, ff ),
    												  clx - xdith, cly - ydith, skyout.pa );
    } else {
    	rerr = sph_fctable_add_row_double( fctab, cpl_vector_get( timev, ff ),
    												  clx, cly, crx, cry, skyout.pa );
   	}

    SPH_ERROR_ENSURE_GOTO_EXIT( rerr == CPL_ERROR_NONE, rerr );

    for (ff = 1; ff < nplanes; ++ff) {

        if ( strcmp( ra, "NONE") != 0 && strcmp( dec, "NONE") != 0 ){
        	sph_skycalc_set_mjd( skyinp, cpl_vector_get(timev,ff) );
        	sph_skycalc( skyinp, &skyout );
        	SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "iTime = %f, iPA = %f", cpl_vector_get(timev,ff), skyout.pa );
        } else {
        	skyout.pa = 0.0;
        }

        if ( fctab->mode == SPH_FCTAB_SINGLE ){
       		rerr = sph_fctable_add_row_single( fctab, cpl_vector_get( timev, ff ),
        												  clx - xdith, cly - ydith, skyout.pa );
        } else {
        	rerr = sph_fctable_add_row_double( fctab, cpl_vector_get( timev, ff ),
        												  clx, cly, crx, cry, skyout.pa );
       	}

        SPH_ERROR_ENSURE_GOTO_EXIT( rerr == CPL_ERROR_NONE, rerr );

    }

    cpl_vector_delete(timev); timev = NULL;
    return fctab;

EXIT:
    cpl_vector_delete(timev); timev = NULL;
    sph_fctable_delete(fctab); fctab = NULL;
    return NULL;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Find a new field centre table
 * @param science_frame the science frame to find corr. fctable for
 * @param field_center_frames list of field centre frames to search
 *
 * @return new field centre table or NULL on error.
 *
 * Searches the field center list of frames for a field center that corresponds
 * to the input science frame.
 *
 * Matching is done on the filename for txt files and on the PRO FIRST RAW FRAME
 * for FITS files.
 *
 * If no corresponding frame is found NULL is returned.
 */
/*----------------------------------------------------------------------------*/
sph_fctable*
sph_fctable_find_fctable(
        cpl_frame* raw_science_frame,
        cpl_frameset* star_centre_calib_frameset
        )
{
    sph_fctable*    field_centre_tab = NULL;
    int             ff = 0;
    cpl_frame*      current_frame = NULL;
    cpl_frame*      star_centre_calib_frame = NULL;
    cpl_propertylist*   pl = NULL;
    char           base[256];
    char           ext[256];
    char           base_raw[256];
    char*          fname = NULL;
    char*          fname2 = NULL;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    cpl_ensure(raw_science_frame,CPL_ERROR_NULL_INPUT,NULL);
    cpl_ensure(star_centre_calib_frameset,CPL_ERROR_NULL_INPUT,NULL);

    fname = sph_filemanager_remove_dir(
            cpl_frame_get_filename(raw_science_frame)
            );
    sph_filemanager_split(fname,base_raw,ext);
    for (ff = 0; ff < cpl_frameset_get_size(star_centre_calib_frameset); ++ff) {
        current_frame = cpl_frameset_get_frame(star_centre_calib_frameset,ff);
        fname2 = sph_filemanager_remove_dir(cpl_frame_get_filename(current_frame));
        sph_filemanager_split(fname2,base,ext);
        if ( strcasecmp(ext,"txt") != 0 ) {
            pl = sph_keyword_manager_load_properties(
                    cpl_frame_get_filename(current_frame),
                    0);
            if ( cpl_propertylist_has(pl,
                    SPH_COMMON_KEYWORD_PRO_FIRST_RAW_FRAME) ) {
                sph_filemanager_split(
                        cpl_propertylist_get_string(pl,
                        SPH_COMMON_KEYWORD_PRO_FIRST_RAW_FRAME) ,
                        base, ext);
                if ( strncasecmp(base,base_raw,
                        strlen(base_raw)) == 0 ) {
                    star_centre_calib_frame = current_frame;
                }
            }
            else {
                SPH_ERROR_RAISE_ERR( CPL_ERROR_BAD_FILE_FORMAT,
                        "The star center frame %s is missing "
                        "the %s keyword!",
                        cpl_frame_get_filename(current_frame),
                        SPH_COMMON_KEYWORD_PRO_FIRST_RAW_FRAME);
                break;
            }
        }
        if ( star_centre_calib_frame == NULL ) {
            if ( strncasecmp(base,base_raw,strlen(base_raw)) == 0 ) {
                star_centre_calib_frame = current_frame;
            }
        }
        cpl_free(fname2); fname2 = NULL;
    }
    cpl_free(fname); fname = NULL;
    if ( star_centre_calib_frame ) {
        field_centre_tab = sph_fctable_load_ascii(
                cpl_frame_get_filename( star_centre_calib_frame ));
        if ( field_centre_tab ) {
            return field_centre_tab;
        }
        else {
            cpl_error_reset();
            field_centre_tab = sph_fctable_load_fits(
                    cpl_frame_get_filename( star_centre_calib_frame ));
        }
    }

    return NULL;
}

/**@}*/
