/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#define _sph_interpolation_C
#include <cpl.h>
#include "sph_error.h"
#include "sph_interpolation.h"
#include "sph_kdtree.h"
#include <math.h>
#undef _sph_interpolation_C

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides functionality for apertures, extending the functionality
 * as it exists for cpl_apertures.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

sph_interpolation*
sph_interpolation_new(cpl_array* xpositions,
        cpl_array* ypositions, int n) {
    sph_interpolation* self = NULL;
    int             ii = 0;
    double          pos[2];

    cpl_ensure(xpositions,CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(ypositions,CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(n > 0,CPL_ERROR_NULL_INPUT, NULL);
    self = cpl_calloc(1,sizeof(sph_interpolation));
    self->xpositions = xpositions;
    self->ypositions = ypositions;
    self->n = n;
    self->indices = cpl_calloc(
            cpl_array_get_size(xpositions),
            sizeof(int));
    self->kdtree = sph_kd_create(2);
    for (ii = 0; ii < cpl_array_get_size(xpositions); ++ii) {
        pos[0] = cpl_array_get(xpositions,ii,NULL);
        pos[1] = cpl_array_get(ypositions,ii,NULL);
        self->indices[ii] = ii;
        sph_kd_insert(self->kdtree,pos,&self->indices[ii]);
    }
    return self;
}

void
sph_interpolation_delete( sph_interpolation* self) {
    if ( self ) {
        cpl_array_delete( self->xpositions );
        cpl_array_delete( self->ypositions );
        cpl_free(self->indices);
        if ( self->kdtree ) sph_kd_free(self->kdtree);
        cpl_imagelist_delete(self->index_imagelist);
        cpl_imagelist_delete(self->distances_imagelist);
        cpl_free(self);
    }
}

cpl_vector*
sph_interpolation_find_closest_points(
        sph_interpolation* self,
           double xpos, double ypos)
{
    cpl_vector*     result = NULL;
    int             ii      = 0;
    double          pos[2];
    sph_kdres*   kdresult = NULL;
    int*            intp = NULL;
    double          dist = 0.0;

    cpl_ensure(self,CPL_ERROR_NULL_INPUT, NULL);

    pos[0] = xpos; pos[1] = ypos;
    kdresult = sph_kd_nearest(self->kdtree,pos,0.0);
    result = cpl_vector_new(self->n);

    cpl_vector_fill(result,-1.0);
    for (ii = 0; ii < self->n; ++ii) {
        if ( kdresult ) {
            intp = (int*)kdresult->riter->item->data;
            dist = kdresult->riter->dist_sq;
            cpl_vector_set(result,ii,*intp);
            sph_kd_res_free(kdresult);
            kdresult = NULL;
            kdresult = sph_kd_nearest(self->kdtree,pos,dist + 0.001);
        }
        else {
            cpl_vector_delete(result);result = NULL;
            break;
        }
    }
    if ( kdresult ) sph_kd_res_free(kdresult);
    return result;
}
cpl_image*
sph_interpolation_interpolate(
        sph_interpolation* self,
        cpl_array*     values) {
    cpl_image*      result = NULL;
    int             nx      = 0;
    int             ny      = 0;
    int             xx      = 0;
    int             yy      = 0;
    int             pp      = 0;
    double          value   = 0.0;
    double          weight   = 0.0;
    double          z = 0.0;
    double          sq_dist = 0.0;
    int             bp = 0;
    int             index = 0;
    cpl_ensure(self, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(self->index_imagelist,
            CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(self->distances_imagelist,
            CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(values, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure( cpl_array_get_size( values ) ==
            cpl_array_get_size(self->xpositions),
            CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    cpl_ensure( cpl_imagelist_get_size(
            self->index_imagelist ) ==
            self->n,
            CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    cpl_ensure( cpl_imagelist_get_size(
            self->distances_imagelist ) ==
            self->n,
            CPL_ERROR_INCOMPATIBLE_INPUT, NULL);

    nx = cpl_image_get_size_x(
            cpl_imagelist_get(self->index_imagelist,0) );
    ny = cpl_image_get_size_y(
            cpl_imagelist_get(self->index_imagelist,0) );

    result = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            value = 0.0;
            weight = 0.0;
            for (pp = 0; pp < self->n; ++pp) {
                sq_dist = cpl_image_get(
                        cpl_imagelist_get(
                                self->distances_imagelist,pp),
                        xx + 1, yy + 1, &bp);
                index = cpl_image_get(
                        cpl_imagelist_get(
                                self->index_imagelist,pp),
                        xx + 1, yy + 1, &bp);
                z = cpl_array_get(values,index,NULL);
                if ( sq_dist > 0.0000001 ) {
                    value += z/sq_dist;
                    weight += 1.0/sq_dist;
                }
            }
            if ( weight ) value = value / weight;
            cpl_image_set( result, xx + 1, yy + 1, value);
        }
    }
    return result;
}
sph_error_code
sph_interpolation_prepare_index_assocs(
            sph_interpolation* self,
            int nx,
            int ny)
{
    double              xpos = 0.0;
    double              ypos = 0.0;
    int                 xx = 0;
    int                 yy = 0;
    int                 pp = 0;
    cpl_image*          im = NULL;
    cpl_mask*           m = NULL;
    cpl_vector*         indices = NULL;
    int                 index = 0;
    double              sq_dist = 0.0;
    double              dx = 0.0;
    double              dy = 0.0;
    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT);

    if ( self->index_imagelist ) {
        cpl_imagelist_delete(self->index_imagelist);
        self->index_imagelist = NULL;
    }
    if ( self->distances_imagelist ) {
        cpl_imagelist_delete(self->distances_imagelist);
        self->distances_imagelist = NULL;
    }
    self->index_imagelist = cpl_imagelist_new();
    self->distances_imagelist = cpl_imagelist_new();
    for (pp = 0; pp < self->n; ++pp) {
        im = cpl_image_new(nx,ny,CPL_TYPE_INT);
        m = cpl_image_get_bpm(im);
        cpl_mask_not(m);
        cpl_imagelist_set(
                self->index_imagelist,
                im,pp);
        im = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE);
        m = cpl_image_get_bpm(im);
        cpl_mask_not(m);
        cpl_imagelist_set(self->distances_imagelist,
                im,pp);
    }

    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            xpos = xx + 0.5; // pixel midpoint
            ypos = yy + 0.5;
            indices = sph_interpolation_find_closest_points(
                    self,xpos,ypos);
            if ( indices ) {
                if ( cpl_vector_get_size(indices) == self->n ) {
                    for (pp = 0; pp < self->n; ++pp) {
                        index = cpl_vector_get( indices, pp);
                        if ( index >=0 ) {
                            im = cpl_imagelist_get(
                                    self->index_imagelist,pp);
                            cpl_image_accept(im,xx+1,yy+1);
                            cpl_image_set(im,
                                    xx + 1, yy + 1,index);
                            im = cpl_imagelist_get(
                                    self->distances_imagelist,
                                    pp);
                            cpl_image_accept(im,xx+1,yy+1);
                            dx = cpl_array_get(
                                    self->xpositions,
                                    index,NULL) - xpos;
                            dy = cpl_array_get(
                                    self->ypositions,
                                    index,NULL) - ypos;
                            sq_dist = dx * dx + dy * dy;
                            cpl_image_set(im,
                                    xx + 1, yy + 1,sq_dist);

                        }
                    }
                    cpl_vector_delete(indices); indices = NULL;
                }
                else {
                    SPH_ERROR_RAISE_ERR(
                            SPH_ERROR_GENERAL,
                            "Could not find %d neigbours for %d,%d",self->n,xx,yy);
                }
            }
            else {
                SPH_ERROR_RAISE_ERR(
                        SPH_ERROR_GENERAL,
                        "Could not find neigbours for %d,%d",xx,yy);
            }
        }
    }
    return cpl_error_get_code();
}
/**@}*/
