/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include <strings.h>
#include "sph_error.h"
#define SPH_POINT_PATTERN_C_
#include "sph_point_pattern.h"
#include "sph_fitting.h"
#include <math.h>
#define SPH_POINT_PATTERN_COL1_NAME "POINT ID"
#define SPH_POINT_PATTERN_COL2_NAME "POINT X"
#define SPH_POINT_PATTERN_COL3_NAME "POINT Y"
#define SPH_POINT_PATTERN_COL4_NAME "POINT DX"
#define SPH_POINT_PATTERN_COL5_NAME "POINT DY"
#undef SPH_POINT_PATTERN_C_

static void write_bad_col_err( const char* colname, const char* fn, int ext ) {
    sph_error_raise( CPL_ERROR_FILE_IO,
            __FILE__,
            __func__,
            __LINE__,
            SPH_ERROR_ERROR,
            "Could not load table from %s, "
            "extension %d, "
            "table has bad format (missing %s column)",
            fn,ext, colname);
}

/* Read one line from standard input, */
/* copying it to line array (but no more than max chars). */
/* Does not place terminating \n in line array. */
/* Returns line length, or 0 for empty line, or EOF for end-of-file. */
static int __sph_point_pattern_getline(FILE* f,char line[], int max)
{
    int nch = 0;
    int c;
    max = max - 1;          /* leave room for '\0' */

    while((c = fgetc(f)) != EOF)
    {
        if(c == '\n')
            break;

        if(nch < max)
        {
            line[nch] = c;
            nch = nch + 1;
        }
    }

    if(c == EOF && nch == 0)
        return EOF;

    line[nch] = '\0';
    return nch;
}
/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/

sph_point_pattern* sph_point_pattern_new(void) {
    sph_point_pattern* result = cpl_calloc( sizeof(sph_point_pattern), 1);

    result->points_x = cpl_vector_new(1);
    result->points_y = cpl_vector_new(1);
    return result;
}
sph_error_code
sph_point_pattern_save_dfs(
        sph_point_pattern* self,
        const char* outfilename,
        cpl_frameset* allframes,
        cpl_frame* template_frame,
        cpl_parameterlist* params,
        const char* tag,
        const char* recipe,
        const char* pipename,
        cpl_propertylist* plist )
{
    sph_error_code        rerr = CPL_ERROR_NONE;
    int np = sph_point_pattern_get_size(self);
    int ii = 0;
    sph_point* p = NULL;
    cpl_table*    tab = cpl_table_new(np);
    cpl_propertylist*       alist = NULL;

    alist = cpl_propertylist_new();
    if ( plist ) {
        cpl_propertylist_append(alist,plist);
    }
    cpl_table_new_column(tab, "POINT ID", CPL_TYPE_INT);
    cpl_table_new_column(tab, "POINT X", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POINT Y", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POINT DX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POINT DY", CPL_TYPE_DOUBLE);
    for (ii = 0; ii < np; ++ii) {
        cpl_table_set(tab,"POINT ID",ii,ii + 1);
        p = sph_point_pattern_get_point(self,ii);
        if ( p ) {
            cpl_table_set(tab,"POINT X",ii,p->x);
            cpl_table_set(tab,"POINT Y",ii,p->y);
            cpl_table_set(tab,"POINT DX",ii,0.0);
            cpl_table_set(tab,"POINT DY",ii,0.0);
            sph_point_delete(p);
        }
    }
    cpl_propertylist_append_string(alist,CPL_DFS_PRO_CATG,tag);
    cpl_dfs_save_table(allframes,
            NULL,
            params,
            allframes,
            template_frame,
            tab,
            plist,
            recipe,
            alist,
            NULL,
            pipename,outfilename);
    cpl_table_delete(tab);
    cpl_propertylist_delete(alist);
    return rerr;
}
sph_error_code
sph_point_pattern_save( sph_point_pattern* self, const char* filename ) {
    sph_error_code        rerr = CPL_ERROR_NONE;
    int np = sph_point_pattern_get_size(self);
    int ii = 0;
    sph_point* p = NULL;
    cpl_table*    tab = cpl_table_new(np);

    cpl_table_new_column(tab, "POINT ID", CPL_TYPE_INT);
    cpl_table_new_column(tab, "POINT X", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POINT Y", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POINT DX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POINT DY", CPL_TYPE_DOUBLE);
    for (ii = 0; ii < np; ++ii) {
        cpl_table_set(tab,"POINT ID",ii,ii + 1);
        p = sph_point_pattern_get_point(self,ii);
        if ( p ) {
            cpl_table_set(tab,"POINT X",ii,p->x);
            cpl_table_set(tab,"POINT Y",ii,p->y);
            cpl_table_set(tab,"POINT DX",ii,0.0);
            cpl_table_set(tab,"POINT DY",ii,0.0);
            sph_point_delete(p);
        }
    }
    cpl_table_save(tab,NULL,NULL,filename,CPL_IO_DEFAULT);
    cpl_table_delete(tab);
    return rerr;
}

sph_error_code
sph_point_pattern_save_ascii(
        sph_point_pattern* self,
        const char* filename ) {
    sph_error_code        rerr = CPL_ERROR_NONE;
    int np = sph_point_pattern_get_size(self);
    int ii = 0;
    sph_point* p = NULL;
    FILE*       f = NULL;

    cpl_ensure_code(filename,CPL_ERROR_NULL_INPUT);
    f = fopen(filename,"w");
    cpl_ensure_code(f,CPL_ERROR_FILE_IO);
    for (ii = 0; ii < np; ++ii) {
        p = sph_point_pattern_get_point(self,ii);
        if ( p ) {
            fprintf(f,"     %10.4f %10.4f\n",p->x, p->y);
            sph_point_delete(p);
        }
    }
    fclose(f); f = NULL;
    return rerr;
}

sph_point_pattern*
sph_point_pattern_load_ascii(const char* filename) {
    FILE*       f = NULL;
    char       line[256];
    double      clx = 0.0;
    double      cly = 0.0;
    int         nmatches = 0;
    int         linesize = 0;
    sph_point_pattern*    self = NULL;

    cpl_ensure(filename,CPL_ERROR_NULL_INPUT,NULL);
    f = fopen(filename,"r");
    cpl_ensure(f,CPL_ERROR_FILE_NOT_FOUND,NULL);

    self = sph_point_pattern_new();

    while ( (linesize = __sph_point_pattern_getline(f,line,256)) != EOF ) {
        if ( line[0]== '#' || linesize == 0 ) {
            ;
        }
        else {
            nmatches = sscanf(line,"%lf %lf",
                    &clx,&cly);
            if ( nmatches == 2 ) {
                sph_point_pattern_add_point(self,clx,cly);
            }
            else {
                SPH_ERROR_ENSURE_GOTO_EXIT(0,CPL_ERROR_BAD_FILE_FORMAT);
            }
        }
    }
    fclose(f); f = NULL;

    return self;

EXIT:
    sph_point_pattern_delete(self); self = NULL;
    return NULL;
}


sph_point_pattern* sph_point_pattern_new_from_table( cpl_frame* framein, int ext ) {
    sph_point_pattern* result = NULL;
    cpl_table*    tab = NULL;
    int ii = 0;
    int bv = 0;
    double x;
    double y;
    if ( framein == NULL || ext < 0 ) {
        return NULL;
    }
    tab = cpl_table_load( cpl_frame_get_filename(framein),ext,0);
    if ( tab == NULL ) {
        return NULL;
    }
    result = cpl_calloc( sizeof(sph_point_pattern), 1);
    result->points_x = cpl_vector_new(1);
    result->points_y = cpl_vector_new(1);
    if ( strcasecmp( cpl_table_get_column_name(tab),SPH_POINT_PATTERN_COL1_NAME) != 0 ) {
        write_bad_col_err( SPH_POINT_PATTERN_COL1_NAME, cpl_frame_get_filename(framein), ext );
        return NULL;
    }
    if ( strcasecmp( cpl_table_get_column_name(NULL),SPH_POINT_PATTERN_COL2_NAME) != 0 ) {
        write_bad_col_err( SPH_POINT_PATTERN_COL2_NAME, cpl_frame_get_filename(framein), ext );
        return NULL;
    }
    if ( strcasecmp( cpl_table_get_column_name(NULL),SPH_POINT_PATTERN_COL3_NAME) != 0 ) {
        write_bad_col_err( SPH_POINT_PATTERN_COL3_NAME, cpl_frame_get_filename(framein), ext );
        return NULL;
    }
    if ( strcasecmp( cpl_table_get_column_name(NULL),SPH_POINT_PATTERN_COL4_NAME) != 0 ) {
        write_bad_col_err( SPH_POINT_PATTERN_COL4_NAME, cpl_frame_get_filename(framein), ext );
        return NULL;
    }
    if ( strcasecmp( cpl_table_get_column_name(NULL),SPH_POINT_PATTERN_COL5_NAME) != 0 ) {
        write_bad_col_err( SPH_POINT_PATTERN_COL5_NAME, cpl_frame_get_filename(framein), ext );
        return NULL;
    }
    if ( cpl_table_get_ncol(tab) != 5 ) {
        sph_error_raise( CPL_ERROR_FILE_IO,
                __FILE__,
                __func__,
                __LINE__,
                SPH_ERROR_ERROR,
                "Could not load table from %s, "
                "extension %d, "
                "table has bad format (expected 5 columns, got %d)",
                cpl_frame_get_filename(framein),ext,
                (int)cpl_table_get_ncol(tab));
        return NULL;
    }
    for (ii = 0; ii < cpl_table_get_nrow(tab); ++ii) {
        x = cpl_table_get_double(tab, SPH_POINT_PATTERN_COL2_NAME, ii, &bv);
        y = cpl_table_get_double(tab, SPH_POINT_PATTERN_COL3_NAME, ii, &bv);
        if ( bv == 0 ) {
            sph_point_pattern_add_point(result,x,y);
        }
    }
    cpl_table_delete(tab); tab = NULL;
    return result;
}

sph_point_pattern* sph_point_pattern_new_from_image(
        cpl_image* image,
        cpl_image* errim,
        double sigma,
        double point_width_estimate)
{
    cpl_apertures* aps = NULL;
    sph_point_pattern* result = NULL;
    double                    totpix = 0.0;
    sph_error_code rerr = CPL_ERROR_NONE;
    int ii = 0;
    double          median, med_dist;
    double          threshold;
    double          xpos = 0.0;
    double          ypos = 0.0;
    double          xpos0 = 0.0;
    double          ypos0 = 0.0;
    double          xw = 0.0;
    double          yw = 0.0;
    cpl_image*      internal_errim = NULL;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    result = sph_point_pattern_new();
    if ( result == NULL ) {
        sph_error_raise( SPH_ERROR_MEMORY_FAIL,
                __FILE__,
                __func__,
                __LINE__,
                SPH_ERROR_ERROR,
                "Could not allocate point pattern.");
        return NULL;
    }


    aps = cpl_apertures_extract_sigma(image, sigma);
    if (aps == NULL) {
        median = cpl_image_get_median_dev(image, &med_dist);
        threshold = median + sigma * med_dist;
        SPH_RAISE_CPL;
        sph_error_raise( CPL_ERROR_DATA_NOT_FOUND,
                __FILE__,
                __func__,
                __LINE__,
                SPH_ERROR_WARNING,
                "Could not find any points, "
                "using a sigma of %f and a calculated "
                "threshold of %f. "
                "Try lowering the sigma.", sigma, threshold);
        return NULL;
    }
    if (cpl_apertures_get_size(aps)<1) {
        sph_error_raise( CPL_ERROR_DATA_NOT_FOUND,
                __FILE__,
                __func__,
                __LINE__,
                SPH_ERROR_WARNING,
                "Could not find any points, using a sigma of %f. "
                "Try lowering the sigma.", sigma);
        return NULL;
    }

    for (ii = 0; ii < cpl_apertures_get_size(aps); ++ii) {
        totpix += cpl_apertures_get_npix( aps, ii +1 );
    }

    totpix /= cpl_apertures_get_size(aps);

    if ( errim == NULL ) {
        internal_errim = errim = cpl_image_abs_create(image);
        cpl_image_power(internal_errim,0.5);
    }

    for (ii = 0; ii < cpl_apertures_get_size(aps); ++ii) {
        if ( fabs(cpl_apertures_get_npix(aps,ii+1) - totpix ) < 40000 )
            xpos = xpos0 = cpl_apertures_get_centroid_x(aps,ii + 1) - 0.5;
            ypos = ypos0 = cpl_apertures_get_centroid_y(aps,ii + 1) - 0.5;
            xw = point_width_estimate;
            yw = point_width_estimate;
            xpos = xpos + 0.5;
            ypos = ypos + 0.5;
            if ( sph_fitting_fit_gauss2D(image,errim,&xpos,&ypos,&xw,&yw) ==
                    CPL_ERROR_NONE)
            {
                rerr = sph_point_pattern_add_point( result,xpos,ypos);
            }
            else {
                SPH_ERROR_RAISE_WARNING(
                        cpl_error_get_code(),
                        "Could not fit the point at %5.2f,%5.2f with a gauss, "
                        "so am using aperture centroids for coordinates",
                        xpos0,ypos0);
                cpl_error_reset();
                rerr = sph_point_pattern_add_point( result,xpos0,ypos0);
            }
    }
    median = cpl_image_get_median_dev(image, &med_dist);
    threshold = median + sigma * med_dist;
    SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
            "Found %d points from %d apertures "
            "using a sigma of %f and a calculated "
            "threshold of %f. ",
            sph_point_pattern_get_size(result),
            (int)cpl_apertures_get_size(aps),
            sigma, threshold);
    cpl_image_delete(internal_errim);
    cpl_apertures_delete(aps);
    return result;
}
//sph_point_pattern* sph_point_pattern_new_from_image(cpl_image* image, double sigma) {
//    cpl_apertures* aps = NULL;
//    sph_point_pattern* result = NULL;
//    double                    totpix = 0.0;
//    sph_error_code rerr = CPL_ERROR_NONE;
//    int ii = 0;
//    double          median, med_dist = 0.0;
//    double          threshold = 0.0 ;
//    cpl_mask    *   selection = NULL;
//    cpl_image   *   labels = NULL;
//    int            nlabels = 0;
//
//    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
//
//    median = cpl_image_get_median_dev(image, &med_dist);
//    threshold = median + sigma * med_dist;
//
//    selection=cpl_mask_threshold_image_create(image, threshold, DBL_MAX);
//    cpl_ensure(selection, CPL_ERROR_ILLEGAL_OUTPUT, NULL);
//    /* Labelise the different detected apertures */
//    if ((labels = cpl_image_labelise_mask_create(selection, &nlabels))==NULL) {
//        cpl_mask_delete(selection);
//        cpl_ensure(0, CPL_ERROR_ILLEGAL_OUTPUT, NULL);
//    }
//    cpl_mask_delete(selection);
//
//    /* Nothing detected */
//    if (nlabels == 0) {
//        cpl_image_delete(labels);
//        return NULL;
//    }
//
//    result = sph_point_pattern_new();
//    if ( result == NULL ) {
//        sph_error_raise( SPH_ERROR_MEMORY_FAIL,
//                __FILE__,
//                __func__,
//                __LINE__,
//                SPH_ERROR_ERROR,
//                "Could not allocate point pattern.");
//        cpl_image_delete(labels);
//        return NULL;
//    }
//    aps = cpl_apertures_new_from_image(image, labels);
//    cpl_image_delete(labels);
//    if (aps == NULL) {
//        SPH_RAISE_CPL;
//        sph_error_raise( CPL_ERROR_DATA_NOT_FOUND,
//                __FILE__,
//                __func__,
//                __LINE__,
//                SPH_ERROR_WARNING,
//                "Could not find any points, "
//                "using a sigma of %f and a calculated "
//                "threshold of %f", sigma, threshold);
//        return NULL;
//    }
//    if (cpl_apertures_get_size(aps)<1) {
//        sph_error_raise( CPL_ERROR_DATA_NOT_FOUND,
//                __FILE__,
//                __func__,
//                __LINE__,
//                SPH_ERROR_WARNING,
//                "Could not find any points, using a sigma of %f", sigma);
//        return NULL;
//    }
//
//    for (ii = 0; ii < cpl_apertures_get_size(aps); ++ii) {
//        totpix += cpl_apertures_get_npix( aps, ii +1 );
//    }
//
//    totpix /= cpl_apertures_get_size(aps);
//
//    for (ii = 0; ii < cpl_apertures_get_size(aps); ++ii) {
//        if ( fabs(cpl_apertures_get_npix(aps,ii+1) - totpix ) < 2 )
//            rerr = sph_point_pattern_add_point( result, cpl_apertures_get_centroid_x(aps,ii + 1),
//                                   cpl_apertures_get_centroid_y(aps,ii + 1));
//
//    }
//    return result;
//}
sph_error_code
sph_point_pattern_add_point( sph_point_pattern* self, double x, double y ) {
    int vs = 0;
    int rerr = CPL_ERROR_NONE;
    if (!self )
        return CPL_ERROR_NULL_INPUT;
    if ( !self->points_x || !self->points_y ) {
        return CPL_ERROR_DATA_NOT_FOUND;
    }
    vs = cpl_vector_get_size(self->points_x);
    if ( (rerr = cpl_vector_set( self->points_x, vs - 1, x)) != CPL_ERROR_NONE)
        return rerr;
    if ( (rerr = cpl_vector_set( self->points_y, vs - 1, y)) != CPL_ERROR_NONE)
        return rerr;
    if ( (rerr = cpl_vector_set_size( self->points_x, vs + 1)) != CPL_ERROR_NONE)
        return rerr;
    if ( (rerr = cpl_vector_set_size( self->points_y, vs + 1)) != CPL_ERROR_NONE)
        return rerr;
    return rerr;
}

sph_error_code
sph_point_pattern_rotate(
        sph_point_pattern* self,
        double cx, double cy,
        double angle )
{
    int vs = 0;
    int rerr = CPL_ERROR_NONE;
    int vv = 0;
    sph_point* apoint = NULL;
    double      sinang = sin(angle / 180.0 * CPL_MATH_PI);
    double      cosang = cos(angle / 180.0 * CPL_MATH_PI);

    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( self->points_x && self->points_y,
            CPL_ERROR_NULL_INPUT);

    vs = sph_point_pattern_get_size(self);

    apoint = sph_point_new(0.0,0.0);
    cpl_ensure_code( apoint, CPL_ERROR_ILLEGAL_OUTPUT );

    for (vv = 0; vv < vs; ++vv) {
        apoint->x = cpl_vector_get( self->points_x,vv);
        apoint->y = cpl_vector_get( self->points_y,vv);
        sph_point_rotate_around( apoint, cx,cy,
                cosang, sinang, 0.0, 0.0);
        cpl_vector_set( self->points_x, vv, apoint->x);
        cpl_vector_set( self->points_y, vv, apoint->y);
    }
    sph_point_delete(apoint);
    return rerr;
}

sph_error_code
sph_point_pattern_scale_around(
        sph_point_pattern* self,
        double cx, double cy,
        double scale )
{
    int vs = 0;
    int rerr = CPL_ERROR_NONE;
    int vv = 0;
    sph_point* apoint = NULL;

    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( self->points_x && self->points_y,
            CPL_ERROR_NULL_INPUT);

    vs = sph_point_pattern_get_size(self);

    apoint = sph_point_new(0.0,0.0);
    cpl_ensure_code( apoint, CPL_ERROR_ILLEGAL_OUTPUT );

    for (vv = 0; vv < vs; ++vv) {
        apoint->x = cpl_vector_get( self->points_x,vv);
        apoint->y = cpl_vector_get( self->points_y,vv);

        apoint->x = ( apoint->x - cx ) * scale + cx;
        apoint->y = ( apoint->y - cy ) * scale + cy;

        cpl_vector_set( self->points_x, vv, apoint->x);
        cpl_vector_set( self->points_y, vv, apoint->y);
    }
    sph_point_delete(apoint);
    return rerr;
}

sph_error_code
sph_point_pattern_clear( sph_point_pattern* self ) {
    int rerr = CPL_ERROR_NONE;
    if (!self )
        return CPL_ERROR_NULL_INPUT;
    if ( !self->points_x || !self->points_y ) {
        return CPL_ERROR_DATA_NOT_FOUND;
    }
    cpl_vector_delete( self->points_x );
    cpl_vector_delete( self->points_y );
    if ( (self->points_x = cpl_vector_new(1)) == NULL )
        return cpl_error_get_code();
    if ( (self->points_y = cpl_vector_new(1)) == NULL )
        return cpl_error_get_code();

    return rerr;
}

sph_error_code
sph_point_pattern_centralise( sph_point_pattern* self ) {
    double x0 = 0.0;
    double y0 = 0.0;
    double xx = 0.0;
    double yy = 0.0;
    int ii;
    if (!self )
        return CPL_ERROR_NULL_INPUT;
    if ( !self->points_x || !self->points_y ) {
        return CPL_ERROR_DATA_NOT_FOUND;
    }
    for (ii = 0; ii < cpl_vector_get_size(self->points_x)-1; ++ii) {
        x0 = x0 + cpl_vector_get(self->points_x,ii);
        y0 = y0 + cpl_vector_get(self->points_y,ii);
    }
    x0 = x0 / ( cpl_vector_get_size(self->points_x) - 1);
    y0 = y0 / ( cpl_vector_get_size(self->points_y) - 1);
    for (ii = 0; ii < cpl_vector_get_size(self->points_x)-1; ++ii) {
        xx = cpl_vector_get(self->points_x,ii);
        yy = cpl_vector_get(self->points_y,ii);
        cpl_vector_set( self->points_x, ii, xx - x0 );
        cpl_vector_set( self->points_y, ii, yy - y0 );
    }
    return CPL_ERROR_NONE;
}
int
sph_point_pattern_get_central_point(
        sph_point_pattern* self,
        double* x, double* y )
{
    double x0 = 0.0;
    double y0 = 0.0;
    double xx = 0.0;
    double yy = 0.0;
    int ii;
    double  l = 0.0;
    double mindist = DBL_MAX;
    int minind = 0;

    if (!self )
        return CPL_ERROR_NULL_INPUT;
    if ( !self->points_x || !self->points_y ) {
        return CPL_ERROR_DATA_NOT_FOUND;
    }
    for (ii = 0; ii < cpl_vector_get_size(self->points_x)-1; ++ii) {
        x0 = x0 + cpl_vector_get(self->points_x,ii);
        y0 = y0 + cpl_vector_get(self->points_y,ii);
    }
    x0 = x0 / ( cpl_vector_get_size(self->points_x) );
    y0 = y0 / ( cpl_vector_get_size(self->points_y) );
    for (ii = 0; ii < cpl_vector_get_size(self->points_x) - 1; ++ii) {
        xx = cpl_vector_get(self->points_x,ii);
        yy = cpl_vector_get(self->points_y,ii);

        l = ( xx - x0 ) * ( xx - x0 ) + ( yy - y0 ) * ( yy - y0 );
        if ( l < mindist ) {
            mindist = l;
            minind = ii;
        }
    }

    *x = cpl_vector_get(self->points_x,minind);
    *y = cpl_vector_get(self->points_y,minind);
    return minind;
}

sph_error_code
sph_point_pattern_add_offset( sph_point_pattern* self, double x0, double y0 ) {
    double xx = 0.0;
    double yy = 0.0;
    int ii;

    if (!self )
        return CPL_ERROR_NULL_INPUT;
    if ( !self->points_x || !self->points_y ) {
        return CPL_ERROR_DATA_NOT_FOUND;
    }

    for (ii = 0; ii < cpl_vector_get_size(self->points_x)-1; ++ii) {
        xx = cpl_vector_get(self->points_x,ii);
        yy = cpl_vector_get(self->points_y,ii);
        cpl_vector_set( self->points_x, ii, xx + x0 );
        cpl_vector_set( self->points_y, ii, yy + y0 );
    }
    return CPL_ERROR_NONE;
}

sph_error_code
sph_point_pattern_centralise_wrt_central_point( sph_point_pattern* self ) {
    double x0 = 0.0;
    double y0 = 0.0;
    int minind = 0;

    if (!self )
        return CPL_ERROR_NULL_INPUT;
    if ( !self->points_x || !self->points_y ) {
        return CPL_ERROR_DATA_NOT_FOUND;
    }

    minind = sph_point_pattern_get_central_point(self,&x0,&y0);
    return sph_point_pattern_add_offset(self,-1.0 * x0, -1.0 * y0);
}

sph_point*
sph_point_pattern_get_point( sph_point_pattern* self, int n ) {
    sph_point* result = NULL;
    if (!self )
        return NULL;
    if ( !self->points_x || !self->points_y ) {
        return NULL;
    }
    if ( n >= cpl_vector_get_size(self->points_x) ||
            n >= cpl_vector_get_size(self->points_y) ||
            n < 0 ) {
        return NULL;
    }
    result = sph_point_new(cpl_vector_get(self->points_x,n),cpl_vector_get(self->points_y,n) );
    return result;
}

sph_point*
sph_point_pattern_find_closest( sph_point_pattern* self, double x, double y ) {
    double d = 0.0;
    double xx = 0.0;
    double yy = 0.0;
    double mind = -1.0;
    int best_ind = -1;
    int ii;
    if (!self )
        return NULL;
    if ( !self->points_x || !self->points_y ) {
        return NULL;
    }

    for (ii = 0; ii < cpl_vector_get_size(self->points_x) - 1; ++ii) {
        xx = cpl_vector_get(self->points_x,ii);
        yy = cpl_vector_get(self->points_y,ii);
        d = (x - xx) * (x - xx) + (y - yy) * (y - yy);
        if (mind < -0.5 || d < mind) {
            mind = d;
            best_ind = ii;
        }
    }
    if ( best_ind < 0 ||
         best_ind > cpl_vector_get_size(self->points_x) - 1 ||
         best_ind > cpl_vector_get_size(self->points_y) - 1 ) {
        sph_error_raise( SPH_ERROR_GENERAL, __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR, "Could not find a point.");
        return NULL;
    }
    return sph_point_pattern_get_point(self, best_ind);
}
int
sph_point_pattern_get_size( sph_point_pattern* self ) {
    if (!self )
        return -1;
    if ( !self->points_x || !self->points_y ) {
        return -1;
    }
    return (cpl_vector_get_size( self->points_x ) - 1);
}
//cpl_image*
//sph_point_pattern_create_image( sph_point_pattern* self, int nx, int ny, double psize ) {
//    cpl_image* image = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE );
//    cpl_image* smallim = NULL;
//    cpl_image* extractim = NULL;
//    int size0 = (int)psize * 20;
//    int size = size0;
//    int ddx = 0;
//    int ddy = 0;
//    sph_error_code rerr = CPL_ERROR_NONE;
//    double xpos = 0.0;
//    double ypos = 0.0;
//    double xpos0 = 0.0;
//    double ypos0 = 0.0;
//    int ii = 0;
//    for ( ii = 0; ii < cpl_vector_get_size(self->points_x)-1; ++ii) {
//        xpos = cpl_vector_get(self->points_x,ii);
//        ypos = cpl_vector_get(self->points_y,ii);
//        size = size0;
//        xpos0 = xpos - size * 0.5;
//        ypos0 = ypos - size * 0.5;
//        if ( xpos0 + size < 0 || ypos0 + size < 0 ) {
//            continue;
//        }
//        if ( xpos0 >= nx || ypos0 >= ny ) {
//            continue;
//        }
//        if ((int)xpos0 + size >= nx) {
//            size = nx - (int)xpos0;
//        }
//        if ((int)ypos0 + size >= ny)
//            size = ny - (int)ypos0;
//        if ((int)xpos0 < 0)
//            xpos0 = 0;
//        if ((int)ypos0 < 0)
//            ypos0 = 0;
//        ddx = xpos - (int)(xpos0);
//        ddy = ypos - (int)(ypos0);
//        smallim = cpl_image_new(size,size,CPL_TYPE_DOUBLE);
//        rerr = cpl_image_fill_gaussian( smallim, ddx + 1.0,
//                ddy + 1.0,
//                1000.0,
//                psize, psize );
//        extractim = cpl_image_extract(image,(int)(xpos0) + 1, (int)(ypos0) + 1,
//                (int)(xpos0) + size, (int)(ypos0) + size);
//        rerr = cpl_image_add(extractim,smallim);
//        rerr = cpl_image_copy( image, extractim, (int)(xpos0) + 1, (int)(ypos0) + 1);
//        cpl_image_delete(smallim);smallim = NULL;
//        cpl_image_delete(extractim);extractim = NULL;
//    }
//    return image;
//}

cpl_image*
sph_point_pattern_create_image( sph_point_pattern* self, int nx, int ny, double psize )
{
    cpl_image* image = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE );
    int         ii      = 0;

    for (ii = 0; ii < sph_point_pattern_get_size(self); ++ii) {
        __sph_point_pattern_add_gauss_to_image(
                image,
                cpl_vector_get(self->points_x,ii),
                cpl_vector_get(self->points_y,ii),
                psize, 1000.0
                );
    }
    return image;
}

int
sph_point_pattern_find_closest_neighbour( sph_point_pattern* self, int n, double* pdist ) {
    double d = 0.0;
    double xx = 0.0;
    double yy = 0.0;
    double x = 0.0;
    double y = 0.0;
    double mind = -1.0;
    int best_ind = -1;
    int ii;
    if (!self )
        return -1;
    if ( !self->points_x || !self->points_y ) {
        return -1;
    }
    if ( n >= cpl_vector_get_size(self->points_x) ||
            n >= cpl_vector_get_size(self->points_y) ||
            n < 0 ) {
        return -1;
    }
    x = cpl_vector_get(self->points_x,n);
    y = cpl_vector_get(self->points_y,n);
    for (ii = 0; ii < cpl_vector_get_size(self->points_x) - 1; ++ii) {
        if ( ii != n ) {
            xx = cpl_vector_get(self->points_x,ii);
            yy = cpl_vector_get(self->points_y,ii);
            d = (x - xx) * (x - xx) + (y - yy) * (y - yy);
            if (mind < -0.5 || d < mind) {
                mind = d;
                best_ind = ii;
            }
        }
    }
    if ( best_ind < 0 ||
         best_ind > cpl_vector_get_size(self->points_x) - 1 ||
         best_ind > cpl_vector_get_size(self->points_y) - 1 ) {
        sph_error_raise( SPH_ERROR_GENERAL, __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR, "Could not find a point.");
        return -1;
    }
    if ( pdist ) {
        *pdist = sqrt(mind);
    }
    return best_ind;
}

void sph_point_pattern_delete( sph_point_pattern* self ) {
    if ( !self ) return;
    if ( self->points_x ) cpl_vector_delete( self->points_x );
    if ( self->points_y ) cpl_vector_delete( self->points_y );
    cpl_free ( self );
}

static
int
__sph_point_pattern_add_gauss_to_image( cpl_image* inimage,
        double dxpos,
        double dypos,
        double psize,
        double value )
{
    int size0 = (int)psize * 20;
    int size;
    int xpos = (int)dxpos;
    int ypos = (int)dypos;
    int ypos0;
    int xpos0;
    int nx = cpl_image_get_size_x(inimage);
    int ny = cpl_image_get_size_y(inimage);
    double ddx;
    double ddy;
    sph_error_code        rerr = CPL_ERROR_NONE;
    cpl_image*    smallim = NULL;
    cpl_image*  extractim = NULL;

    cpl_ensure_code(inimage,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    if ( size0 < 2 ) {
        SPH_ERR("Too small gaussian to add to image.");
        return CPL_ERROR_ILLEGAL_INPUT;
    }
    size = size0;
    xpos0 = xpos - size * 0.5;
    ypos0 = ypos - size * 0.5;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    if ((int)xpos0 < 0)
        xpos0 = 0;
    if ((int)ypos0 < 0)
        ypos0 = 0;

    if ((int)xpos0 + size >= nx) {
        size = nx - (int)xpos0;
    }
    if ((int)ypos0 + size >= ny)
        size = ny - (int)ypos0;
    if ( size  < 1 ) return CPL_ERROR_NONE;
    ddx = dxpos - (int)(xpos0);
    ddy = dypos - (int)(ypos0);
    smallim = cpl_image_new(size,size,CPL_TYPE_DOUBLE);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    //SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,"ddx=%f,ddy=%f",ddx,ddy);
    rerr = cpl_image_fill_gaussian( smallim, ddx,
                             ddy,
                             value,
                             psize, psize );
    SPH_RAISE_CPL_RESET
    extractim = cpl_image_extract(inimage,(int)(xpos0) + 1, (int)(ypos0) + 1,
                                  (int)(xpos0) + size, (int)(ypos0) + size);
    if ( extractim == NULL ) {
        SPH_RAISE_CPL;
        SPH_ERROR_RAISE_ERR(SPH_ERROR_GENERAL,"Could not extract image size %d from %d,%d.",size,(int)xpos0,(int)ypos0);
        return SPH_ERROR_GENERAL;
    }
    rerr = cpl_image_add(extractim,smallim);
    SPH_RAISE_CPL_RESET
    rerr = cpl_image_copy( inimage, extractim, (int)(xpos0) + 1, (int)(ypos0) + 1);
    cpl_image_delete(smallim);smallim = NULL;
    cpl_image_delete(extractim);extractim = NULL;
    SPH_RAISE_CPL_RESET;
    return rerr;
}


/**@}*/
