/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <Basic.h>
#include <CUnit.h>
#include "sph_error.h"
#include "sph_test.h"
#include "sph_test_ngc_ir_simulator.h"
#include "sph_test_pupilimage_creator.h"
#include "sph_test_image_tools.h"
#include "cutest_ifs_science.h"
#include "sph_ifs_test_fixtures_calib_products.h"
#include "sph_ldt.h"
#include "sph_framecombination.h"
#include "sph_utils.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A CUnit Test Suite -- representing a collection of testcases
 * @par Synopsis:
 * @code
 * @endcode
 * @par Desciption:
 *
 * This module provides a collection of tests for one specific, distinct
 * module or set-up. The testing code is implemented using the CUnit
 * framework.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("A testsuite", cutest_ifs_science_init_testsuite,
                          cutest_ifs_science_clean_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "Testing coilapse LDTs",
                             cutest_ifs_science_collapse_ldts ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "Testing combine LDTs",
                             cutest_ifs_science_combine_ldts ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}


/*----------------------------------------------------------------------------*/
/*-                            INTERNAL HELPER FUNCTIONS                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to initiailise the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_ifs_science_init_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de",CPL_MSG_DEBUG);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to clean the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_ifs_science_clean_testsuite() {
    sph_error_dump( SPH_ERROR_INFO );
    sph_end_test();
    return 0;
}
/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_science_collapse_ldts(void) {
    cpl_frameset*           inframes = NULL;
    cpl_frameset*           allframes = NULL;
    cpl_frameset*           collapsed = NULL;
    sph_master_frame*       result = NULL;
    double                  median = 0.0;
    inframes = cutest_ifs_science_collapse_ldts_fixture(1);
    allframes = cpl_frameset_duplicate(inframes);
    collapsed = sph_ifs_science_collapse_ldts( inframes,
            allframes,NULL,"SOMERECIPE", "SOMETAG");
    CU_ASSERT_PTR_NOT_NULL_FATAL( collapsed );
    CU_ASSERT_EQUAL( cpl_frameset_get_size(collapsed), 1 );
    result = sph_master_frame_load(
            cpl_frame_get_filename(
                    cpl_frameset_get_first(collapsed) ), 0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( result );
    median = sph_master_frame_get_median(result,NULL);
    sph_master_frame_divide_double(result,median);
    sph_master_frame_mask_tolerance(result,0.000001,1000000.0);
    //CU_ASSERT_DOUBLE_EQUAL( sph_master_frame_get_mean(result,NULL), 1.0, 0.01);

    // The accuracy for the check below is so bad, because
    // the lenslet model slits are wider than the fixture model slits
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_max(result->image), 1.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_median(result->image), 1.0, 0.01);
    sph_master_frame_save( result, "result.fits", NULL);
    sph_master_frame_delete(result);
    cpl_frameset_delete(collapsed);
    cpl_frameset_delete(inframes);
    cpl_frameset_delete(allframes);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_science_combine_ldts(void) {
    cpl_frameset*           inframes = NULL;
    cpl_frameset*           allframes = NULL;
    sph_ldt*           collapsed = NULL;
    sph_master_frame*       result = NULL;
    double                  median = 0.0;
    inframes = cutest_ifs_science_collapse_ldts_fixture(1);
    allframes = cpl_frameset_duplicate(inframes);
    collapsed = sph_ifs_science_combine_ldts( inframes,
            SPH_COLL_ALG_MEDIAN, NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL( collapsed );
    result = sph_ldt_collapse_interpolate(collapsed,0.0,1.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL( result );
    median = sph_master_frame_get_median(result,NULL);
    sph_master_frame_divide_double(result,median);
    sph_master_frame_mask_tolerance(result,0.000001,1000000.0);
    //CU_ASSERT_DOUBLE_EQUAL( sph_master_frame_get_mean(result,NULL), 1.0, 0.01);

    // The accuracy for the check below is so bad, because
    // the lenslet model slits are wider than the fixture model slits
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_max(result->image), 1.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_median(result->image), 1.0, 0.01);
    sph_master_frame_save( result, "result_combined.fits", NULL);
    sph_master_frame_delete(result);
    sph_ldt_delete(collapsed);
    cpl_frameset_delete(inframes);
    cpl_frameset_delete(allframes);
    return;
}



static
cpl_frameset* cutest_ifs_science_collapse_ldts_fixture(int n) {
    sph_ldt*            ldt = NULL;
    sph_test_ngc_ir_simulator* ngc = NULL;
    sph_test_ifs_instrument_model* ifsmodel = NULL;
    cpl_frameset*           inframes = NULL;
    cpl_frameset*           rawframes = NULL;
    cpl_frame*              aframe = NULL;
    int                     ii      = 0;

    ngc = sph_test_ngc_ir_simulator_new(1024,1024);
    CU_ASSERT_PTR_NOT_NULL_FATAL( ngc );
    ifsmodel = sph_test_ifs_instrument_model_new(ngc->det_size_x,
           SPH_IFS_LENSLET_MODEL_LENSLET_SIZE_MICRONS,
            SPH_IFS_LENSLET_MODEL_LENSLETS_SIDE);
    CU_ASSERT_PTR_NOT_NULL_FATAL( ifsmodel );
    inframes = cpl_frameset_new();
    for (ii = 0; ii < n; ++ii) {
        ldt = sph_ifs_test_fixtures_calib_products_simple_ldt(
                ngc, ifsmodel );

        CU_ASSERT_PTR_NOT_NULL_FATAL(ldt);
        aframe = sph_filemanager_create_temp_frame(
                "LDT_collapse",
                "TESTTAG",
                CPL_FRAME_GROUP_CALIB);

        sph_ldt_save_cube( ldt, cpl_frame_get_filename(aframe),
                NULL, NULL, NULL, NULL, NULL, NULL, NULL );

        cpl_frameset_insert( inframes, aframe );

        sph_ldt_delete(ldt);
    }


    sph_test_ngc_ir_simulator_delete( ngc );
    sph_test_ifs_instrument_model_delete( ifsmodel );
    return inframes;
}


/**@}*/
