/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <Basic.h>
#include <CUnit.h>
#include "sph_error.h"
#include "sph_test.h"
#include "sph_ifs_tags.h"
#include "sph_test_ngc_ir_simulator.h"
#include "sph_test_pupilimage_creator.h"
#include "sph_test_image_tools.h"
#include "sph_ifs_spec_deconv_run.c"
#include "sph_ifs_spec_deconv.c"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A CUnit Test Suite -- representing a collection of testcases
 * @par Synopsis:
 * @code
 * @endcode
 * @par Desciption:
 *
 * This module provides a collection of tests for one specific, distinct
 * module or set-up. The testing code is implemented using the CUnit
 * framework.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/*-                            INTERNAL HELPER FUNCTIONS                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to initiailise the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_init_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de",CPL_MSG_DEBUG);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to clean the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_clean_testsuite() {
    sph_error_dump( SPH_ERROR_ERROR );
    return sph_end_test();
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Fixtures
 *
/*----------------------------------------------------------------------------*/
static
sph_ldt*
cutest_ifs_spec_deconv_fixture_empty_ldt() {
    sph_ldt*      ldt = NULL;
    sph_ifs_lenslet_model*              model = NULL;
    model = sph_ifs_lenslet_model_new();
    ldt = sph_ldt_new(model);
    return ldt;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
cpl_imagelist* cutest_ifs_spec_deconv_fixture_imcube(int nx, int ny,
        double xcent,
        double ycent,
        cpl_vector** lams,
        int no_planet) {
    cpl_imagelist*      imlist = NULL;
    int                 nplanes = 0;
    cpl_imagelist*      result = NULL;
    int                 pp      = 0;
    double              xpos    = 0.0;
    cpl_image*          im      = NULL;
    double              lambda  = 0.0;
    double              dlambda = 0.04;
    double              lambda_min = 1.2;
    double              xpos0       = 20.0;
    double              p0          = 5.0;
    double              p1          = 3.0;
    sph_ifs_lenslet_model*      model = NULL;
    cpl_vector*         reflam = NULL;

    model = sph_ifs_lenslet_model_new();
    reflam = sph_ifs_lenslet_model_get_lambda_vector(model);
    nplanes = cpl_vector_get_size(reflam);
    imlist = cpl_imagelist_new();
    *lams = cpl_vector_new(nplanes);
    for (pp = 0; pp < nplanes; ++pp) {
        im = sph_test_image_tools_create_flat_image_double(nx,ny,0.0);
        lambda = cpl_vector_get(reflam,pp);
        //oldcode :_ pp * dlambda + lambda_min;
        cpl_vector_set(*lams,pp,lambda);
        xpos = xpos0 * lambda + xcent;
        sph_test_image_tools_add_gauss(im,xcent,ycent,p0 * lambda,10000.0);
        if ( no_planet == 0 ) {
            sph_test_image_tools_add_gauss(im,xpos,ycent,p1 * lambda,1000.0);
        }
        cpl_imagelist_set(imlist,im,pp);
        SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    }
    cpl_imagelist_save(imlist,"input_imlist.fits", CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    sph_ifs_lenslet_model_delete(model);
    return imlist;
}


/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
cpl_frame*
cutest_ifs_spec_deconv_fixture_standard(int nx, int ny,
        double xcent,
        double ycent,
        cpl_vector** lams,
        int noplanet) {
    cpl_imagelist*      imlist = NULL;
    int                 nplanes = 0;
    cpl_frame*          outframe = NULL;
    cpl_imagelist*      result = NULL;
    int                 pp      = 0;
    double              xpos    = 0.0;
    cpl_image*          im      = NULL;
    double              lambda  = 0.0;
    double              dlambda = 0.06;
    double              lambda_min = 1.2;
    double              xpos0       = 20.0;
    double              p0          = 3.0;
    double              p1          = 1.0;
    int                 ii          = 0.0;

    imlist = cutest_ifs_spec_deconv_fixture_imcube(
            nx,ny,xcent,ycent,lams,noplanet);
    CU_ASSERT_EQUAL_FATAL(cpl_imagelist_get_size(imlist),
            cpl_vector_get_size(*lams));
    sph_ifs_test_fixtures_science_add_random_speckles(
            imlist,
            xcent,ycent,
            200,*lams,0.9,NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(imlist);
    nplanes = cpl_imagelist_get_size(imlist);
    for (ii = 0; ii < nplanes; ++ii) {
        im = cpl_imagelist_get(imlist,ii);
        if ( noplanet == 0 ) {
            sph_test_image_tools_add_gauss(im, xcent + 20.0,
                    ycent + 50.0, 2.0, 20.0);
        }
        sph_test_image_tools_add_noise(im,0.05,NULL);
    }
    outframe = sph_filemanager_create_temp_frame("cutest_ifs_spec_deconv",
            SPH_IFS_TAG_SPEC_DECONV_RAW,
            CPL_FRAME_GROUP_RAW);
    CU_ASSERT_EQUAL(
            cpl_imagelist_save(
                    imlist,
                    cpl_frame_get_filename(outframe),
                    CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE),
            CPL_ERROR_NONE);
    cpl_imagelist_delete(imlist); imlist = NULL;
    return outframe;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_spec_deconv_resize_image(void) {
    int                     nx                = 256;
    int                     ny                = 256;
    cpl_image*              imA               = NULL;

    imA = sph_test_image_tools_create_flat_image_double(nx,ny,10.0);
    CU_ASSERT_EQUAL_FATAL(
            sph_ifs_spec_deconv_resize_image(&imA,190,190),
            CPL_ERROR_NONE);
    CU_ASSERT_PTR_NOT_NULL_FATAL(imA);
    CU_ASSERT_EQUAL(cpl_image_get_size_x(imA),190);
    CU_ASSERT_EQUAL(cpl_image_get_size_y(imA),190);
    cpl_image_save(imA,"imA.fits",CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    cpl_image_delete(imA);
    CU_ASSERT_EQUAL(cpl_error_get_code(),CPL_ERROR_NONE);
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_spec_deconv_fit_imcube(void) {
    cpl_imagelist*          imlist       = NULL;
    cpl_imagelist*          bpixlist       = NULL;
    cpl_imagelist*          rmslist       = NULL;
    cpl_imagelist*          result       = NULL;
    cpl_vector*             lambda           = NULL;
    int                     nx                = 256;
    int                     ny                = 256;
    int                     pp              = 0;
    sph_error_code          rerr = CPL_ERROR_NONE;
    cpl_image*              im = NULL;
    double                  val = 0.0;
    double                  xval = 0.0;
    imlist = cpl_imagelist_new();
    bpixlist = cpl_imagelist_new();
    rmslist = cpl_imagelist_new();
    lambda = cpl_vector_new(30);
    for (pp = 0; pp < 30; ++pp) {
        val = 1021.0 + (pp + 0.5) * 100.0 + 0.75 * (pp + 0.5) * (pp + 0.5);
        cpl_vector_set(lambda, pp , 1.0 / (pp + 0.5));
        im = sph_test_image_tools_create_flat_image_double(nx,ny,val);
        cpl_imagelist_set(imlist,im,pp);
        im = sph_test_image_tools_create_flat_image_double(nx,ny,0.0);
        cpl_imagelist_set(bpixlist,im,pp);
        im = sph_test_image_tools_create_flat_image_double(nx,ny,1.0);
        cpl_imagelist_set(rmslist,im,pp);
    }
    result = sph_ifs_spec_deconv_fit_imcube(imlist,bpixlist,rmslist,lambda,2);
    CU_ASSERT_PTR_NOT_NULL_FATAL(result);
    cpl_imagelist_subtract(result,imlist);
    cpl_imagelist_save(result,"resultlist.fits",CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    for (pp = 0; pp < 30; ++pp) {
        CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_absflux(cpl_imagelist_get(result,pp)),0.0,0.00001);
    }

    cpl_imagelist_delete(imlist); imlist = NULL;
    cpl_imagelist_delete(bpixlist); bpixlist = NULL;
    cpl_imagelist_delete(rmslist); rmslist = NULL;
    cpl_imagelist_delete(result); result = NULL;
    cpl_vector_delete(lambda); lambda = NULL;

}
/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_spec_deconv_scale_imcube_inverse(void) {
    cpl_imagelist*          inputlist       = NULL;
    cpl_imagelist*          comp_list       = NULL;
    cpl_vector*             lambda           = NULL;
    int                     nx                = 256;
    int                     ny                = 256;
    int                     pp              = 0;
    sph_error_code          rerr = CPL_ERROR_NONE;
    cpl_image*              refim = NULL;
    cpl_image*              diffim = NULL;
    double                  maxval  = 0.0;
    sph_transform*         transform = NULL;
    inputlist = cutest_ifs_spec_deconv_fixture_imcube(
            nx,ny,nx/2.0+0.5,ny/2.0+0.5,&lambda,0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(inputlist);
    comp_list = cpl_imagelist_duplicate(inputlist);
    CU_ASSERT_PTR_NOT_NULL_FATAL(comp_list);
    for (pp = 0; pp < cpl_imagelist_get_size(inputlist); ++pp) {
        CU_ASSERT_EQUAL(cpl_image_get_type(cpl_imagelist_get(inputlist,pp)), CPL_TYPE_DOUBLE);
        CU_ASSERT_EQUAL(cpl_image_get_size_x(cpl_imagelist_get(inputlist,pp)), nx);
        CU_ASSERT_EQUAL(cpl_image_get_size_y(cpl_imagelist_get(inputlist,pp)), ny);
    }
    transform = sph_transform_new_default();

    rerr = sph_ifs_spec_deconv_scale_imcube(inputlist,NULL,NULL,NULL,
    		lambda,cpl_vector_get_max(lambda),
    		nx/2.0+0.5,ny/2.0+0.5,0,transform, -1.0 );
    rerr = sph_ifs_spec_deconv_scale_imcube(inputlist,
    		NULL,NULL,NULL,
    		lambda,cpl_vector_get_max(lambda),
    		nx/2.0+0.5,ny/2.0+0.5,1,transform, -1.0 );
    CU_ASSERT_EQUAL_FATAL(rerr,CPL_ERROR_NONE);
    cpl_imagelist_subtract(inputlist,comp_list);
    for (pp = 0; pp < cpl_imagelist_get_size(inputlist); ++pp) {
        CU_ASSERT_EQUAL(cpl_image_get_type(cpl_imagelist_get(inputlist,pp)), CPL_TYPE_DOUBLE);
        CU_ASSERT_EQUAL(cpl_image_get_size_x(cpl_imagelist_get(inputlist,pp)), nx);
        CU_ASSERT_EQUAL(cpl_image_get_size_y(cpl_imagelist_get(inputlist,pp)), ny);
        diffim = cpl_image_duplicate(cpl_imagelist_get(inputlist,pp));
        CU_ASSERT_PTR_NOT_NULL_FATAL(diffim);
        cpl_image_abs(diffim);
        CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_max_window(diffim,107,107,166,152),0.0,0.1);
        CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_max_window(diffim,107,107,166,152),0.0,0.02);
        cpl_image_delete(diffim); diffim = NULL;
    }
    cpl_imagelist_save(inputlist,"output_imlist_inv.fits", CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    cpl_imagelist_delete(inputlist); inputlist = NULL;
    cpl_imagelist_delete(comp_list); comp_list = NULL;
    cpl_vector_delete(lambda); lambda = NULL;
    sph_transform_delete(transform); transform = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_spec_deconv_scale_imcube(void) {
    cpl_imagelist*          inputlist       = NULL;
    cpl_vector*             lambda           = NULL;
    int                     nx                = 256;
    int                     ny                = 256;
    int                     pp              = 0;
    sph_error_code          rerr = CPL_ERROR_NONE;
    cpl_image*              refim = NULL;
    cpl_image*              diffim = NULL;
    double                  maxval  = 0.0;
    sph_transform*         transform = NULL;
    inputlist = cutest_ifs_spec_deconv_fixture_imcube(
            nx,ny,(double)nx/2.0+0.5,(double)ny/2.0+0.5,&lambda,0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(inputlist);
    for (pp = 0; pp < cpl_imagelist_get_size(inputlist); ++pp) {
        CU_ASSERT_EQUAL(cpl_image_get_type(cpl_imagelist_get(inputlist,pp)), CPL_TYPE_DOUBLE);
        CU_ASSERT_EQUAL(cpl_image_get_size_x(cpl_imagelist_get(inputlist,pp)), nx);
        CU_ASSERT_EQUAL(cpl_image_get_size_y(cpl_imagelist_get(inputlist,pp)), ny);
    }
    transform = sph_transform_new_default();
    rerr = sph_ifs_spec_deconv_scale_imcube(inputlist,
    		NULL,NULL,NULL,
    		lambda,cpl_vector_get_max(lambda),
    		nx/2.0+0.5,ny/2.0+0.5,0,transform, -1.0 );
    CU_ASSERT_EQUAL_FATAL(rerr,CPL_ERROR_NONE);
    refim = cpl_imagelist_get(inputlist,0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(refim);
    maxval = cpl_image_get_max(refim);
    for (pp = 0; pp < cpl_imagelist_get_size(inputlist); ++pp) {
        CU_ASSERT_EQUAL(cpl_image_get_type(cpl_imagelist_get(inputlist,pp)), CPL_TYPE_DOUBLE);
        CU_ASSERT_EQUAL(cpl_image_get_size_x(cpl_imagelist_get(inputlist,pp)), nx);
        CU_ASSERT_EQUAL(cpl_image_get_size_y(cpl_imagelist_get(inputlist,pp)), ny);
        diffim = cpl_image_subtract_create(cpl_imagelist_get(inputlist,pp),refim);
        CU_ASSERT_PTR_NOT_NULL_FATAL(diffim);
        cpl_image_divide(diffim,refim);
        cpl_image_abs(diffim);
        CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_max_window(diffim,107,107,166,152),0.0,0.8);
        CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_max_window(diffim,107,107,166,152),0.0,0.8);
        cpl_image_delete(diffim); diffim = NULL;
    }
    cpl_imagelist_save(inputlist,"output_imlist.fits", CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    cpl_imagelist_delete(inputlist); inputlist = NULL;
    cpl_vector_delete(lambda); lambda = NULL;
    sph_transform_delete(transform); transform = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_ifs_spec_deconv_run(void) {
    cpl_frame*              inframe         = NULL;
    cpl_frame*              inframeset      = NULL;
    cpl_imagelist*          inputlist       = NULL;
    cpl_imagelist*          outlist       = NULL;
    cpl_vector*             lambda           = NULL;
    int                     nx                = 291;
    int                     ny                = 291;
    int                     pp              = 0;
    sph_ifs_spec_deconv*    deconv          = NULL;
    cpl_parameterlist*      plist           = NULL;


    CU_ASSERT_EQUAL_FATAL( cpl_error_get_code(), CPL_ERROR_NONE);
    plist = sph_ifs_spec_deconv_create_paramlist();
    CU_ASSERT_PTR_NOT_NULL_FATAL(plist);

    inframeset = cpl_frameset_new();
    inframe = cutest_ifs_spec_deconv_fixture_standard(
            nx,ny,nx/2.0+10.5,ny/2.0+3.5,&lambda,0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(inframe);
    cpl_frameset_insert(inframeset,inframe);

    deconv = sph_ifs_spec_deconv_new(inframeset,plist);
    CU_ASSERT_PTR_NOT_NULL_FATAL(deconv);

    deconv->cx = nx/2.0 + 10.5;
    deconv->cy = ny/2.0 + 3.5;
    deconv->user_cent = 1;
    CU_ASSERT_EQUAL(sph_ifs_spec_deconv_run(deconv), CPL_ERROR_NONE);

    sph_ifs_spec_deconv_delete(deconv); deconv = NULL;
    cpl_vector_delete(lambda); lambda = NULL;
    cpl_frameset_delete(inframeset); inframeset = NULL;
    cpl_parameterlist_delete(plist); plist = NULL;
    CU_ASSERT_EQUAL(cpl_error_get_code(),CPL_ERROR_NONE);
    return;
}



/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("A testsuite", cutest_init_testsuite,
                          cutest_clean_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "Resizing image",
                             cutest_ifs_spec_deconv_resize_image) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "Scaling imcube",
                             cutest_ifs_spec_deconv_scale_imcube) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "Run",
                             cutest_ifs_spec_deconv_run) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
//    if ( NULL == CU_add_test(pSuite, "Fitting imcube",
//                             cutest_ifs_spec_deconv_fit_imcube) )
//    {
//       CU_cleanup_registry();
//       return CU_get_error();
//    }
//    if ( NULL == CU_add_test(pSuite, "Scaling imcube inverse",
//    		cutest_ifs_spec_deconv_scale_imcube_inverse) )
//    {
//    	CU_cleanup_registry();
//    	return CU_get_error();
//    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
