/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_ird_master_dark.h"
#include "sph_ird_tags.h"
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"
#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include "cutest_ird_master_dark.h"
#include "Basic.h"
#include "CUnit.h"
#include "sph_test.h"
#include "cutest_ird_master_dark.h"
#include "sph_utils.h"
#include "sph_test_image_tools.h"
#include "sph_test_ngc_ir_simulator.h"
/*----------------------------------------------------------------------------*/
/**
 * @defgroup  cutest_ird_master_dark  Unit test of ird_master_dark recipe and \
 *                                       associated functions.
 *
 * This tests verifies the very basic requirements of the sph_ird_master_dark
 * recipe. All functions of the recipe object are tested in only one,
 * standard, test scenario. In this scenario, the parameters for the
 * recipe are the standard default parameters. The input raw frames
 * are created using the sph_test_create_ifs_raw_dark_frameset function.
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/
int cutest_init_master_dark_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de", CPL_MSG_DEBUG);
    return CPL_ERROR_NONE;
}

int cutest_clean_master_dark_testsuite(void) {
    sph_error_dump( SPH_ERROR_ERROR );
    return sph_end_test();
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Common setup for the unit tests
 *
 * @param    frames        Address of pointer to the input frames to set up
 * @param   params        Address of pointer to the input parameterlist to set up
 *
 * This sets up the standard recipe parameter list and the input frames.
 * The frames an params pointers should not be initiallised yet.
 *
 * On return the pointers will point to the corresponding cpl structures that
 * are now filled with the setup data for the test.
 *
 */
/*----------------------------------------------------------------------------*/
void cutest_ird_master_dark_common_setup( cpl_frameset** frames,
                                          cpl_parameterlist** params,
                                          int nframes)
{
    sph_error_code              rerr                = CPL_ERROR_NONE;
    cpl_image*                    adark                = NULL;
    sph_test_ngc_ir_simulator*        ngc                    = NULL;
    cpl_frame*                    rawframe            = NULL;

    cpl_error_reset();
    CU_ASSERT_PTR_NOT_NULL_FATAL( frames );
    CU_ASSERT_PTR_NOT_NULL_FATAL( params );

    ngc = sph_test_ngc_ir_simulator_new(256,256);
    CU_ASSERT_PTR_NOT_NULL( ngc );

    adark = sph_test_image_tools_create_flat_image_int( 256, 256, 50 +  5 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( adark );
    rawframe = sph_test_frame_image( adark,
            CPL_BPP_32_SIGNED, SPH_IRD_TAG_DARK_RAW,
            CPL_FRAME_GROUP_RAW );
    CU_ASSERT_PTR_NOT_NULL_FATAL( rawframe );
    *frames = sph_test_ngc_ir_simulator_create_raw_frameset(
            ngc, rawframe, nframes,1,
            "raw_dark.fits", NULL, SPH_IRD_TAG_DARK_RAW );
    CU_ASSERT_PTR_NOT_NULL_FATAL( *frames );
    cpl_image_delete( adark );adark = NULL;
    cpl_frame_delete( rawframe ); rawframe = NULL;
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    *params = sph_ird_master_dark_create_paramlist();

    rerr = cpl_error_get_code();
    if ( rerr ) {
        SPH_RAISE_CPL
    }
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( *frames );
    CU_ASSERT_PTR_NOT_NULL_FATAL( *params );
    sph_test_ngc_ir_simulator_delete(ngc);
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Common teardown for the unit tests
 *
 * @param    frames        pointer to the input frames to clean up
 * @param   params        pointer to the input parameterlist to clean up
 *
 * This cleans up the frameset and parameterlist
 */
/*----------------------------------------------------------------------------*/
static void cutest_ird_master_dark_common_teardown( cpl_frameset* frames, cpl_parameterlist* params) {

    CU_ASSERT_PTR_NOT_NULL_FATAL( frames );
    CU_ASSERT_PTR_NOT_NULL_FATAL( params );

    cpl_parameterlist_delete( params );
    cpl_frameset_delete( frames );
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_ird_master_dark check frames function.
 */
/*----------------------------------------------------------------------------*/
void cutest_ird_master_dark_check_frames(void) {
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;
    sph_error_code              rerr                = CPL_ERROR_NONE;

    /* Setup */
    cpl_error_reset();
    cutest_ird_master_dark_common_setup( &inframes, &inparams, 3 );

    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );
    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark );
    // delete to avoid duplicates and mem leak
    cpl_frameset_delete(ird_master_dark->rawframes);
    cpl_frameset_delete(ird_master_dark->current_raw_frameset);
    /* Running code... */

    rerr = sph_ird_master_dark_check_frames( ird_master_dark );

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark->inframes );

    CU_ASSERT_EQUAL( cpl_frameset_get_size( ird_master_dark->inframes), 3 );

    /* Terdown */
    sph_ird_master_dark_delete( ird_master_dark );
    cutest_ird_master_dark_common_teardown( inframes, inparams );
    return;

}
void cutest_ird_master_dark_check_params(void) {
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;
    sph_error_code              rerr                = CPL_ERROR_NONE;

    /* Setup */
    cutest_ird_master_dark_common_setup( &inframes, &inparams, 3 );
    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark );

    /* Running ... */

    rerr = sph_ird_master_dark_check_params( ird_master_dark );

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark->master_dark_outfilename );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark->static_badpix_filename );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark->framecomb_parameterlist );

    CU_ASSERT_STRING_EQUAL( ird_master_dark->master_dark_outfilename, "master_dark.fits" );

    CU_ASSERT_STRING_EQUAL( ird_master_dark->static_badpix_filename, "static_badpixels.fits" );

    CU_ASSERT_EQUAL( cpl_parameterlist_get_size(
                                    ird_master_dark->framecomb_parameterlist),
                                    2 );
    /* Terdown */
    cutest_ird_master_dark_common_teardown( inframes, inparams );
    sph_ird_master_dark_delete( ird_master_dark );
    return;

}
void cutest_ird_master_dark_new(void)
{
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;

    cutest_ird_master_dark_common_setup( &inframes, &inparams, 3 );

    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );

    CU_ASSERT_PTR_NOT_NULL( ird_master_dark );
    /* Terdown */
    sph_ird_master_dark_delete( ird_master_dark );
    cutest_ird_master_dark_common_teardown( inframes, inparams );
    return;
}

void cutest_ird_master_dark_delete(void) {
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;
    sph_error_code              rerr                = CPL_ERROR_NONE;

    /* Setup */
    cutest_ird_master_dark_common_setup( &inframes, &inparams, 3 );

    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark );
    /* Run code... */
    rerr = sph_ird_master_dark_delete( ird_master_dark );
    /* Verify */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE );
    /* Terdown */
    cutest_ird_master_dark_common_teardown( inframes, inparams );

    return;
}
void cutest_ird_master_dark_run(void) {
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;
    cpl_image*                    bpixs                = NULL;
    sph_error_code              rerr                = CPL_ERROR_NONE;

    /* Setup */
    cutest_ird_master_dark_common_setup( &inframes, &inparams, 3 );

    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark );

    /* Running */
    rerr |= sph_ird_master_dark_run( ird_master_dark );

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );

    bpixs = cpl_image_load( ird_master_dark->master_dark_outfilename, CPL_TYPE_INT, 0, 1 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( bpixs );
    CU_ASSERT_TRUE( cpl_image_get_absflux( bpixs ) == 0 );

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE );
    cpl_image_delete( bpixs ); bpixs = NULL;
     /* Terdown */
    sph_ird_master_dark_delete( ird_master_dark );
    cutest_ird_master_dark_common_teardown( inframes, inparams );
    return;
}

void cutest_ird_master_dark_run_test_illegal_reject(void) {
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;
    cpl_parameter*              p                   = NULL;
    sph_error_code              rerr                = CPL_ERROR_NONE;

    /* Setup */
    cutest_ird_master_dark_common_setup( &inframes, &inparams, 3 );

    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark );

    ird_master_dark->clean_mean_reject_high = 2;
    ird_master_dark->clean_mean_reject_low = 1;
    p = cpl_parameterlist_find( ird_master_dark->framecomb_parameterlist, "clean_mean.reject_low");
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );
    cpl_parameter_set_int( p, ird_master_dark->clean_mean_reject_low );
    p = cpl_parameterlist_find( ird_master_dark->framecomb_parameterlist, "clean_mean.reject_high");
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );
    cpl_parameter_set_int( p, ird_master_dark->clean_mean_reject_high );

    /* Running */
    rerr |= sph_ird_master_dark_run( ird_master_dark );

    /* Verification */
    CU_ASSERT_NOT_EQUAL( rerr, CPL_ERROR_NONE );
    sph_ird_master_dark_delete( ird_master_dark );

    /* Terdown */
    cutest_ird_master_dark_common_teardown( inframes, inparams );
    sph_error_reset();
    return;
}


void cutest_ird_master_dark_run_more_than50(void)
{
    sph_ird_master_dark*        ird_master_dark     = NULL;
    cpl_parameterlist*          inparams            = NULL;
    cpl_frameset*               inframes            = NULL;
    sph_error_code              rerr                = CPL_ERROR_NONE;

    /* Setup */
    inframes = cpl_frameset_new();
    rerr = sph_test_create_ifs_raw_dark_frameset_big( inframes, 50 );
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    inparams = sph_ird_master_dark_create_paramlist();

    ird_master_dark = sph_ird_master_dark_new( inframes, inparams );

    CU_ASSERT_PTR_NOT_NULL_FATAL( ird_master_dark );

    /* Running */
    rerr |= sph_ird_master_dark_run( ird_master_dark );

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE );
    /* Terdown */
    cutest_ird_master_dark_common_teardown( inframes, inparams );
    return;
}


/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int result       = 0;

    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("master_dark_test", cutest_init_master_dark_testsuite,
                          cutest_clean_master_dark_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */

    if ( NULL == CU_add_test(pSuite, "sph_ird_master_dark_new",
                             cutest_ird_master_dark_new ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    if ( NULL == CU_add_test(pSuite, "sph_ird_master_dark_check_frames",
                             cutest_ird_master_dark_check_frames ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    if ( NULL == CU_add_test(pSuite, "sph_ird_master_dark_check_params",
                             cutest_ird_master_dark_check_params ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    //    if ( NULL == CU_add_test(pSuite, "sph_ird_master_dark_run_more_than50",
//                             cutest_ird_master_dark_run_more_than50 ) )
//    {
//       CU_cleanup_registry();
//       return CU_get_error();
//    }

    if ( NULL == CU_add_test(pSuite, "sph_ird_master_dark_run",
                             cutest_ird_master_dark_run ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    if ( NULL == CU_add_test(pSuite, "cutest_ird_master_dark_run_test_illegal_reject",
                             cutest_ird_master_dark_run_test_illegal_reject ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }


    if ( NULL == CU_add_test(pSuite, "sph_ird_master_dark_delete",
                             cutest_ird_master_dark_delete ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
