/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_dataset_stats.h"
#include "sph_dataset.h"
#include "sph_common_keywords.h"
#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup techcal_master_test  Unit test of techcal_master recipe and 
 *                                  associated functions.               
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

sph_dataset*
cutest_util_dataset_new_load() {
    sph_dataset*       dat       = NULL;
    cpl_vector*         lams       = NULL;
    cpl_vector*         vals       = NULL;
    int                 ii         = 0;
    
    dat = sph_dataset_new( 100, 1.0, 100.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat );
    
    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams );
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals );

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        cpl_vector_set( vals, ii, (double)ii);
    }
    
    CU_ASSERT_EQUAL( cpl_vector_get( lams, 0 ), 1.0);
    CU_ASSERT_EQUAL( cpl_vector_get( lams, 99 ), 100.0);
    return dat;
}
sph_dataset*
cutest_util_dataset_new_gauss() {
    sph_dataset*        dat         = NULL;
    cpl_vector*         lams        = NULL;
    cpl_vector*         vals        = NULL;
    double              val         = 0.0;        
    int                 ii          = 0;
    
    dat = sph_dataset_new( 100, 1.0, 100.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat );
    
    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams );
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals );

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        val = cpl_vector_get(lams, ii ) - 50.0;
        val = exp ( -val*val / 25.0 );
        cpl_vector_set( vals, ii, val );
    }
    
    return dat;
}
sph_dataset*
cutest_util_dataset_new_double_gauss() {
    sph_dataset*        dat         = NULL;
    cpl_vector*         lams        = NULL;
    cpl_vector*         vals        = NULL;
    double              val         = 0.0;        
    double              valB        = 0.0;        
    int                 ii          = 0;
    
    dat = sph_dataset_new( 100, 1.0, 100.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat );
    
    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams );
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals );

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        val = cpl_vector_get(lams, ii ) - 25.0;
        val = exp ( -val*val / 15.0 );
        valB = cpl_vector_get(lams, ii ) - 75.0;
        valB = exp ( -valB*valB / 15.0 );
        val = val + valB;
        cpl_vector_set( vals, ii, val );
    }
    
    return dat;
}
 
int cutest_init_dataset_stats_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia.de",CPL_MSG_DEBUG);
    return 0;
}

int cutest_clean_dataset_stats_testsuite() {
    sph_end_test();
    return 0;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_dataset_stats_new function. 
 */
/*----------------------------------------------------------------------------*/
void cutest_dataset_stats_new() {
    sph_dataset*              dataset           = NULL;
    sph_dataset_stats*        dat               = NULL;
    
    /* Setup and run ...*/
    dataset = cutest_util_dataset_new_load();
    CU_ASSERT_PTR_NOT_NULL_FATAL( dataset );
    
    dat = sph_dataset_stats_new( dataset );
    
    /*Verify */
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat );
    sph_dataset_stats_delete(dat); dat = NULL;
    sph_dataset_delete(dataset); dataset = NULL;
    return;   
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_dataset_stats_find_peaks function. 
 */
/*----------------------------------------------------------------------------*/
void cutest_dataset_stats_find_peaks() {
    sph_dataset*            dataset     = NULL;
    sph_dataset_stats*      dat         = NULL;
    sph_error_code          rerr        = CPL_ERROR_NONE;
    
    dataset = cutest_util_dataset_new_gauss();
    CU_ASSERT_PTR_NOT_NULL_FATAL( dataset );
    dat = sph_dataset_stats_new( dataset );
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat ); 

    rerr = sph_dataset_stats_find_peaks( dat, 0.9, 3 );
        
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL_FATAL( dat->npeaks, 1);
    CU_ASSERT_DOUBLE_EQUAL( dat->peakloc[0], 50.0, 0.5);

    rerr |= sph_dataset_stats_delete( dat );

    return;   
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_dataset_stats_find_peaks function. 
 */
/*----------------------------------------------------------------------------*/
void cutest_dataset_stats_find_peaks_exact_number() {
    sph_dataset*            dataset     = NULL;
    sph_dataset_stats*      dat         = NULL;
    sph_error_code          rerr        = CPL_ERROR_NONE;
    
    dataset = cutest_util_dataset_new_double_gauss();
    CU_ASSERT_PTR_NOT_NULL_FATAL( dataset );
    dat = sph_dataset_stats_new( dataset );
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat ); 

    rerr = sph_dataset_stats_find_peaks_exact_number( dat, 2, 0.01 );
        
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL_FATAL( dat->npeaks, 2);
    CU_ASSERT_DOUBLE_EQUAL( dat->peakloc[0], 25.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( dat->peakloc[1], 75.0, 0.5);

    rerr |= sph_dataset_stats_delete( dat );
    sph_dataset_delete(dataset);dataset = NULL;
    return;   
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_dataset_stats_delete function. 
 */
/*----------------------------------------------------------------------------*/
void cutest_dataset_stats_delete() {
    sph_dataset*            dataset           = NULL;
    sph_dataset_stats*        dat        = NULL;
    sph_error_code            rerr        = CPL_ERROR_NONE;
    
    dataset = cutest_util_dataset_new_load();
    CU_ASSERT_PTR_NOT_NULL_FATAL( dataset );
    dat = sph_dataset_stats_new( dataset );
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat ); 

    rerr = sph_dataset_stats_delete( dat );
    rerr = sph_dataset_delete( dataset );
    
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    return;   
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("dataset_stats_test", cutest_init_dataset_stats_testsuite, 
                          cutest_clean_dataset_stats_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "sph_dataset_stats_new", 
                             cutest_dataset_stats_new ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_dataset_stats_find_peaks",
                             cutest_dataset_stats_find_peaks ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_dataset_stats_find_peaks_exact_number",
                             cutest_dataset_stats_find_peaks_exact_number ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_dataset_stats_delete",
                             cutest_dataset_stats_delete ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
