/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_ldt.h"
#include "sph_dataset.h"
#include "sph_common_keywords.h"
#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"
#include "sph_utils.h"
#include "sph_test_image_tools.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup techcal_master_test  Unit test of techcal_master recipe and
 *                                  associated functions.
 *
 */
/*----------------------------------------------------------------------------*/
static
sph_ldt* __ldt = NULL;
/**@{*/


static
sph_ldt*
create_ldt_standard() {
    sph_ifs_lenslet_model*                model        = NULL;
    cpl_image*                            inimage        = NULL;
    sph_pixel_description_table*        pdt         = NULL;
    sph_ldt*        ldt         = NULL;
    int                                    nx = 256;
    int                                    ny = 256;
    int                                    ll    = 0;
    sph_lenslet_descriptor*                ldescr = NULL;
    cpl_vector*                            vec = NULL;

    model = sph_test_create_small_lenslet_model();

    inimage = sph_test_image_tools_create_flat_image_double(nx,ny,0.0);
    pdt = sph_pixel_description_table_new_from_model(model,0.0,0.0);
    ldt = sph_ldt_new_from_pdt_image(pdt,model,inimage);
    sph_pixel_description_table_delete(pdt);
    cpl_image_delete(inimage);
    vec = cpl_vector_duplicate(ldt->reflambdas);
    for (ll = 0; ll < ldt->nlens; ++ll) {
        ldescr = ldt->arr[ll];
        if ( ldescr ) {
            cpl_vector_fill(vec,2.0);
            sph_lenslet_descriptor_fill_values(ldescr,vec);
            cpl_vector_fill(vec,0.0);
            sph_lenslet_descriptor_fill_bpix(ldescr,vec);
            cpl_vector_fill(vec,3.0);
            sph_lenslet_descriptor_fill_rms(ldescr,vec);
            cpl_vector_fill(vec,4.0);
            sph_lenslet_descriptor_fill_ncomb(ldescr,vec);
        }
    }

    cpl_vector_delete(vec); vec = NULL;
    return ldt;
}

static
int cutest_init_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de", CPL_MSG_DEBUG);
    __ldt = create_ldt_standard();
    if ( !__ldt ) return 1;
    if ( cpl_error_get_code() ) return 1;
    return 0;
}
static
int cutest_clean_testsuite() {
    if (__ldt ) sph_ldt_delete(__ldt);
    __ldt = NULL;
    return sph_end_test();
}

static
void cutest_standard_happy_case(void) {
    sph_master_frame*                mframe = NULL;
    int                                    nx = 256;
    int                                    ny = 256;
    cpl_propertylist*                    plist = NULL;

    CU_ASSERT_PTR_NOT_NULL_FATAL(__ldt);
    mframe = sph_ldt_create_mframe(__ldt,nx,ny,1,0.0,NULL);
    CU_ASSERT_EQUAL_FATAL(
            cpl_image_save( mframe->image, "image.fits", CPL_TYPE_DOUBLE,
            plist = sph_ifs_lenslet_model_get_as_propertylist( __ldt->model ),
            CPL_IO_CREATE ), CPL_ERROR_NONE);
    //CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_flux(mframe->image),__ldt->nlens * 2.0, 1.0);
    sph_error_raise( SPH_ERROR_GENERAL,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO,
            "flux:%f vs. %f\n",
            cpl_image_get_flux(mframe->image),__ldt->nlens * 2.0 );
    sph_master_frame_delete(mframe);
    cpl_propertylist_delete(plist); plist = NULL;
}

static
void cutest_standard_happy_case_interpolate(void) {
    sph_master_frame*                mframe = NULL;
    int                                    nx = 256;
    int                                    ny = 256;
    cpl_propertylist*                    plist = NULL;
    sph_interpolation*                  interp = NULL;

    CU_ASSERT_PTR_NOT_NULL_FATAL(__ldt);
    mframe = sph_ldt_create_mframe_interpolate(__ldt,nx,ny,1,0.0,&interp);
    CU_ASSERT_PTR_NOT_NULL_FATAL( mframe );
    CU_ASSERT_PTR_NOT_NULL_FATAL( interp );
    CU_ASSERT_EQUAL_FATAL(
            cpl_image_save( mframe->image, "image2.fits", CPL_TYPE_DOUBLE,
            plist = sph_ifs_lenslet_model_get_as_propertylist( __ldt->model ),
            CPL_IO_CREATE ), CPL_ERROR_NONE);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_image_get_min(mframe->image),
            __ldt->arr[0]->values[0], 0.0001);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_image_get_max(mframe->image),
            __ldt->arr[0]->values[0], 0.0001);
    sph_error_raise( SPH_ERROR_GENERAL,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO,
            "flux:%f vs. %f\n",
            cpl_image_get_flux(mframe->image),__ldt->nlens * 2.0 );
    sph_master_frame_delete(mframe);
    cpl_propertylist_delete(plist); plist = NULL;
    sph_interpolation_delete(interp);
}


static
void cutest_standard_happy_case_interpolate_lensdist(void) {
    sph_master_frame*                mframe = NULL;
    int                                    nx = 256;
    int                                    ny = 256;
    cpl_propertylist*                    plist = NULL;
    sph_interpolation*                  interp = NULL;
    cpl_size                            pows[2] = {1,1};
    sph_polynomial_distortion_model*    lensdistmodel = NULL;

    lensdistmodel =
            sph_polynomial_distortion_model_new(NULL,NULL);

    cpl_polynomial_set_coeff(lensdistmodel->polyx,pows,2.0);

    CU_ASSERT_PTR_NOT_NULL_FATAL(__ldt);
    sph_ifs_lenslet_model_set_lenslet_distortion(__ldt->model, lensdistmodel );
    mframe = sph_ldt_create_mframe_interpolate(__ldt,nx,ny,1,0.0,&interp);
    CU_ASSERT_PTR_NOT_NULL_FATAL( mframe );
    CU_ASSERT_PTR_NOT_NULL_FATAL( interp );
    CU_ASSERT_EQUAL_FATAL(
            cpl_image_save( mframe->image, "image3.fits", CPL_TYPE_DOUBLE,
            plist = sph_ifs_lenslet_model_get_as_propertylist( __ldt->model ),
            CPL_IO_CREATE ), CPL_ERROR_NONE);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_image_get_min(mframe->image),
            __ldt->arr[0]->values[0], 0.0001);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_image_get_max(mframe->image),
            __ldt->arr[0]->values[0], 0.0001);
    sph_error_raise( SPH_ERROR_GENERAL,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO,
            "flux:%f vs. %f\n",
            cpl_image_get_flux(mframe->image),__ldt->nlens * 2.0 );
    sph_master_frame_delete(mframe);
    cpl_propertylist_delete(plist); plist = NULL;
    sph_interpolation_delete(interp);
}


#ifdef SPHERE_TEST_FULL
static
void cutest_big_happy_case(void) {
    sph_master_frame*                mframe = NULL;
    int                                    nx = 402;
    int                                    ny = 402;
    int                                    xx = 0;
    int                                    yy = 0;
    int                                    bad = 0;
    int                                    bp = 0;
    sph_ldt*        ldt = create_ldt_big();
    cpl_propertylist*                    plist = NULL;
    CU_ASSERT_PTR_NOT_NULL_FATAL(ldt);
    mframe = sph_ldt_create_mframe(ldt,nx,ny,1,0.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(mframe);
    CU_ASSERT_EQUAL_FATAL(
            cpl_image_save( mframe->image, "bigimage.fits", CPL_TYPE_DOUBLE,
            plist = sph_ifs_lenslet_model_get_as_propertylist( ldt->model ),
            CPL_IO_CREATE ), CPL_ERROR_NONE);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_flux(mframe->image),ldt->nlens * 2.0, 1.0 );
    sph_master_frame_delete(mframe);
    cpl_propertylist_delete(plist); plist = NULL;
    sph_ldt_delete(ldt);
}
#endif
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("ldt_create_image", cutest_init_testsuite,
                          cutest_clean_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "standard happy case",
                             cutest_standard_happy_case ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "standard happy case interpolate",
                             cutest_standard_happy_case_interpolate ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "interpolate with lensdistortion",
                             cutest_standard_happy_case_interpolate_lensdist ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    /* add the tests to the suite */
#ifdef SPHERE_TEST_FULL
    if ( NULL == CU_add_test(pSuite, "big happy case",
                             cutest_big_happy_case ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
#endif

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
