/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>
#include <math.h>
#include <Basic.h>
#include <CUnit.h>

#include "sph_master_frame.h"
#include "sph_common_keywords.h"
#include "sph_error.h"
#include "sph_test.h"
#include "sph_test_image_tools.h"
#include "sph_utils.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cutest_sph_master_frame Unit test for the sph_master_frame module            
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

sph_master_frame*        __mframe = NULL;

static
sph_master_frame*
create_mframe() {
    int            nx = 256;
    int            ny = 256;
    if ( __mframe ) return __mframe;
    __mframe = sph_master_frame_new(nx,ny);
    cpl_image_delete(__mframe->badpixelmap);__mframe->badpixelmap = NULL;
    __mframe->badpixelmap = sph_test_image_tools_create_flat_image_int(nx,ny,0);
    cpl_image_delete(__mframe->rmsmap);__mframe->rmsmap = NULL;
    __mframe->rmsmap = sph_test_image_tools_create_flat_image_double(nx,ny,10.0);
    cpl_image_delete(__mframe->ncombmap); __mframe->ncombmap = NULL;
    __mframe->ncombmap = sph_test_image_tools_create_flat_image_double(nx,ny,10.0);
    cpl_image_delete(__mframe->image); __mframe->image = NULL;
    __mframe->image = sph_test_image_tools_create_flat_image_double(nx,ny,10.0);
    return __mframe;
}

static
int cutest_init_testsuite() {
    unlink("sphere.cfg");
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de", CPL_MSG_DEBUG);
    __mframe = create_mframe();
    if ( __mframe == NULL ) return 1;
    return 0;
}

static
int cutest_clean_testsuite() {
    /*--------------------------------------------------------------------
     * -    Close and dump errors.
     * -------------------------------------------------------------------*/
    sph_master_frame_delete(__mframe );
    sph_error_dump( SPH_ERROR_ERROR );

    return sph_end_test();
}
static
void
cutest_null_params(void) {
    sph_error_code            rerr = CPL_ERROR_NONE;

    CU_ASSERT_PTR_NOT_NULL_FATAL( __mframe );
    rerr = sph_master_frame_save_dfs(__mframe,
            "master_frame_dfs.fits",
            NULL,
            NULL,
            NULL,
            "TEST_TAG",
            NULL,
            NULL,
            NULL );
    CU_ASSERT_NOT_EQUAL(rerr,CPL_ERROR_NONE);
    SPH_RAISE_CPL_RESET;
}
static
void
cutest_happy_case(void) {
    sph_error_code            rerr = CPL_ERROR_NONE;
    cpl_frameset*            allframes = NULL;
    cpl_parameterlist*        params = NULL;
    cpl_image*                imA = NULL;

    allframes = cpl_frameset_new();
    imA = sph_test_image_tools_create_flat_image_double(256,256,10.0);
    cpl_frameset_insert(allframes,sph_test_frame_image(imA,CPL_BPP_32_SIGNED,"ATAG", CPL_FRAME_GROUP_RAW ));

    params = cpl_parameterlist_new();
    cpl_parameterlist_append(params, cpl_parameter_new_value("ANAME",CPL_TYPE_INT,"Descr","Conext",1.0) );

    CU_ASSERT_PTR_NOT_NULL_FATAL( __mframe );
    CU_ASSERT_PTR_NOT_NULL_FATAL( allframes );
    CU_ASSERT_PTR_NOT_NULL_FATAL( params );

    rerr = sph_master_frame_save_dfs( __mframe,
            "master_frame_dfs.fits",
            allframes,
            cpl_frameset_get_first(allframes),
            params,
            "TEST_TAG",
            "NORECIPE",
            "NOPIPE",
            NULL );
    CU_ASSERT_EQUAL(rerr,CPL_ERROR_NONE);
    cpl_frameset_delete(allframes);
    cpl_parameterlist_delete(params);
    cpl_image_delete(imA);
}

static
void cutest_happy_case_extnames(void) {
    sph_error_code            rerr = CPL_ERROR_NONE;
    cpl_frameset*            allframes = NULL;
    cpl_parameterlist*        params = NULL;
    cpl_image*                imA = NULL;

    allframes = cpl_frameset_new();
    imA = sph_test_image_tools_create_flat_image_double(256,256,10.0);
    cpl_frameset_insert(allframes,sph_test_frame_image(imA,CPL_BPP_32_SIGNED,"ATAG", CPL_FRAME_GROUP_RAW ));

    params = cpl_parameterlist_new();
    cpl_parameterlist_append(params, cpl_parameter_new_value("ANAME",CPL_TYPE_INT,"Descr","Conext",1.0) );

    CU_ASSERT_PTR_NOT_NULL_FATAL( __mframe );
    CU_ASSERT_PTR_NOT_NULL_FATAL( allframes );
    CU_ASSERT_PTR_NOT_NULL_FATAL( params );

    rerr = sph_master_frame_save_dfs( __mframe,
            "master_frame_dfs.fits",
            allframes,
            cpl_frameset_get_first(allframes),
            params,
            "TEST_TAG",
            "NORECIPE",
            "NOPIPE",
            NULL );
    CU_ASSERT_EQUAL(rerr,CPL_ERROR_NONE);
    //check extnames of the double image extensions saved in master frame
    CU_ASSERT_EQUAL( 0, cpl_fits_find_extension("master_frame_dfs.fits",
  		  SPH_COMMON_KEYWORD_VALUE_MASTER_FRAME_IMAGE_EXTNAME) );
    CU_ASSERT_EQUAL( 1, cpl_fits_find_extension("master_frame_dfs.fits",
  		  SPH_COMMON_KEYWORD_VALUE_MASTER_FRAME_BADPIXMAP_EXTNAME) );
    CU_ASSERT_EQUAL( 2, cpl_fits_find_extension("master_frame_dfs.fits",
  		  SPH_COMMON_KEYWORD_VALUE_MASTER_FRAME_NCOMBMAP_EXTNAME) );
    CU_ASSERT_EQUAL( 3, cpl_fits_find_extension("master_frame_dfs.fits",
  		  SPH_COMMON_KEYWORD_VALUE_MASTER_FRAME_RMSMAP_EXTNAME) );

    cpl_frameset_delete(allframes);
    cpl_parameterlist_delete(params);
    cpl_image_delete(imA);
}

/*----------------------------------------------------------------------------*/
/**
 @brief   Main to run the tests.
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result  = 0;
    CU_pSuite   pSuite  = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("cutest_sph_master_frame_save_dfs", cutest_init_testsuite,
                          cutest_clean_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "testing NULL params",
                             cutest_null_params ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "testing happy case",
                             cutest_happy_case ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "testing happy case extnames check",
                             cutest_happy_case_extnames ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
