/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <Basic.h>
#include <CUnit.h>
#include "sph_error.h"
#include "sph_test.h"
#include "sph_test_ngc_ir_simulator.h"
#include "sph_test_pupilimage_creator.h"
#include "sph_fitting.h"
#include "sph_pixel_polyfit_table.h"
#include "sph_utils.h"
#include "sph_test_image_tools.h"
#include <gsl_rng.h>
#include <gsl_randist.h>
#include <gsl_vector.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A CUnit Test Suite -- representing a collection of testcases
 * @par Synopsis:
 * @code
 * @endcode
 * @par Desciption:
 *
 * This module provides a collection of tests for one specific, distinct
 * module or set-up. The testing code is implemented using the CUnit
 * framework.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/*-                            INTERNAL HELPER FUNCTIONS                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to initiailise the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_init_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de",CPL_MSG_DEBUG);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to clean the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_clean_testsuite() {
    sph_error_dump( SPH_ERROR_ERROR );
    sph_end_test();
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_test_pixel_polyfit_table_new() {
    sph_pixel_polyfit_table*        ptab = NULL;

    ptab = sph_pixel_polyfit_table_new(256,128,2);
    CU_ASSERT_PTR_NOT_NULL_FATAL(ptab);
    CU_ASSERT_EQUAL(ptab->nx,256);
    CU_ASSERT_EQUAL(ptab->ny,128);
    CU_ASSERT_EQUAL(ptab->order,2);
    CU_ASSERT_PTR_NOT_NULL(ptab->table);
    sph_pixel_polyfit_table_delete(ptab); ptab = NULL;
    return;
}



/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_test_pixel_polyfit_table_new_from_imlist() {
    sph_pixel_polyfit_table*        ptab = NULL;
    cpl_imagelist*                  imlist = NULL;

    imlist = cpl_imagelist_new();
    cpl_imagelist_set(imlist,sph_test_image_tools_create_flat_image_double(256,128,1.0),0);
    cpl_imagelist_set(imlist,sph_test_image_tools_create_flat_image_double(256,128,2.0),1);
    cpl_imagelist_set(imlist,sph_test_image_tools_create_flat_image_double(256,128,3.0),2);

    ptab = sph_pixel_polyfit_table_new_from_imlist(imlist,NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(ptab);
    CU_ASSERT_EQUAL(ptab->nx,256);
    CU_ASSERT_EQUAL(ptab->ny,128);
    CU_ASSERT_EQUAL(ptab->order,2);
    CU_ASSERT_PTR_NOT_NULL(ptab->table);
    sph_pixel_polyfit_table_save_cube(ptab,"ptab.fits");
    sph_pixel_polyfit_table_delete(ptab); ptab = NULL;
    cpl_imagelist_delete(imlist);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_test_pixel_polyfit_table_set_invalid() {
    sph_pixel_polyfit_table*        ptab = NULL;
    cpl_imagelist*                  imlist = NULL;
    cpl_mask*                       bpixmask = NULL;
    sph_master_frame*               mframe = NULL;
    int                             bpix = 0;
    imlist = cpl_imagelist_new();
    cpl_imagelist_set(imlist,sph_test_image_tools_create_flat_image_double(256,128,1.0),0);
    cpl_imagelist_set(imlist,sph_test_image_tools_create_flat_image_double(256,128,2.0),1);
    cpl_imagelist_set(imlist,sph_test_image_tools_create_flat_image_double(256,128,3.0),2);

    ptab = sph_pixel_polyfit_table_new_from_imlist(imlist,NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(ptab);
    CU_ASSERT_EQUAL(ptab->nx,256);
    CU_ASSERT_EQUAL(ptab->ny,128);
    CU_ASSERT_EQUAL(ptab->order,2);
    CU_ASSERT_PTR_NOT_NULL(ptab->table);

    sph_pixel_polyfit_table_set_invalid(ptab,12,12);

    bpixmask = sph_pixel_polyfit_table_create_badpixel_mask(ptab,-100000.0,10000000.0,10000.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(bpixmask);
    CU_ASSERT_EQUAL(cpl_mask_get(bpixmask,13,13),1);
    CU_ASSERT_EQUAL(cpl_mask_get(bpixmask,12,12),0);

    cpl_mask_delete(bpixmask);

    mframe = sph_pixel_polyfit_table_get_coeff_as_image(ptab,1);
    CU_ASSERT_PTR_NOT_NULL_FATAL(mframe);
    CU_ASSERT_EQUAL(cpl_image_get(mframe->badpixelmap,13,13,&bpix),1);
    CU_ASSERT_EQUAL(cpl_image_get(mframe->badpixelmap,12,12,&bpix),0);

    sph_master_frame_delete(mframe);

    sph_pixel_polyfit_table_delete(ptab); ptab = NULL;
    cpl_imagelist_delete(imlist);
    return;
}


/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("Testing sph_pixel_polyfit_table", cutest_init_testsuite,
                          cutest_clean_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "new",
                             cutest_test_pixel_polyfit_table_new) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "new from imlist",
                             cutest_test_pixel_polyfit_table_new_from_imlist) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "set invalid",
                             cutest_test_pixel_polyfit_table_set_invalid) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
