/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef _OPENMP
#include <omp.h>
#endif
/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_polygon.h"
#include "sph_common_keywords.h"
#include <math.h>
#include "sph_error.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"
#include "sph_time.h"
#include "sph_test_image_tools.h"
#include "sph_utils.h"
/*----------------------------------------------------------------------------*/
/**
 * @defgroup techcal_master_test  Unit test of techcal_master recipe and
 *                                  associated functions.
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/
int cutest_init_polygon_testsuite() {
    return sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de", CPL_MSG_DEBUG);
}

int cutest_clean_polygon_testsuite() {
    sph_error_dump( SPH_ERROR_ERROR );
    return sph_end_test();
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_new() {
    sph_polygon*        result    = NULL;

    result = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( result );

    sph_polygon_delete(result);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_duplicate() {
    sph_polygon*        poly_old    = NULL;
    sph_polygon*        poly    = NULL;
    sph_error_code        rerr    = 0;

    poly_old = sph_polygon_new();
    sph_polygon_add_point(poly_old,1.0,2.0);
    sph_polygon_add_point(poly_old,-1.0,-2.0);

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly_old );
    poly = sph_polygon_duplicate(poly_old);
    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );

    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( poly->npoints, 2 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 2.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->x, -1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->y, -2.0, 1.0e-15);
    sph_polygon_delete(poly);
    sph_polygon_delete(poly_old);
    poly = sph_polygon_duplicate( NULL );
    CU_ASSERT_PTR_NULL( poly );
    cpl_error_reset();
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_add_point() {
    sph_polygon*        poly    = NULL;
    sph_error_code        rerr    = 0;
    poly = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );

    rerr = sph_polygon_add_point( poly, 1.0, 2.0 );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( poly->npoints, 1 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 2.0, 1.0e-15);
    sph_polygon_delete(poly);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_get_edge() {
    sph_polygon*        poly    = NULL;
    sph_error_code        rerr    = 0;
    sph_line*            edgeA    = NULL;
    sph_line*            edgeB    = NULL;
    sph_line*            edgeC    = NULL;
    sph_line*            edgeD    = NULL;

    poly = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    rerr = sph_polygon_add_point( poly, 0.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 0.0, 1.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    edgeA = sph_polygon_get_edge( poly, 1 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( edgeA );
    edgeB = sph_polygon_get_edge( poly, 2 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( edgeB );
    edgeC = sph_polygon_get_edge( poly, 3 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( edgeC );
    edgeD = sph_polygon_get_edge( poly, 4 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( edgeD );
    CU_ASSERT_DOUBLE_EQUAL( edgeA->p1->x, 0.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeA->p1->y, 0.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeA->p2->x, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeA->p2->y, 0.0, 1.0e-15);

    CU_ASSERT_DOUBLE_EQUAL( edgeB->p1->x, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeB->p1->y, 0.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeB->p2->x, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeB->p2->y, 1.0, 1.0e-15);

    CU_ASSERT_DOUBLE_EQUAL( edgeC->p1->x, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeC->p1->y, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeC->p2->x, 0.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeC->p2->y, 1.0, 1.0e-15);

    CU_ASSERT_DOUBLE_EQUAL( edgeD->p1->x, 0.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeD->p1->y, 1.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeD->p2->x, 0.0, 1.0e-15);
    CU_ASSERT_DOUBLE_EQUAL( edgeD->p2->y, 0.0, 1.0e-15);
    sph_line_delete(edgeA);
    sph_line_delete(edgeB);
    sph_line_delete(edgeC);
    sph_line_delete(edgeD);

    edgeD = sph_polygon_get_edge( poly, 5 );
    CU_ASSERT_PTR_NULL_FATAL( edgeD );
    edgeD = sph_polygon_get_edge( poly, 0 );
    CU_ASSERT_PTR_NULL_FATAL( edgeD );

    sph_polygon_delete(poly);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_test_inside() {
    sph_polygon*        poly    = NULL;
    sph_error_code        rerr    = 0;
    poly = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    rerr = sph_polygon_add_point( poly, 0.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 0.0, 1.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, 0.0, 0.0 ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, 1.0, 0.0 ), 1 );
    //CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, 0.0, 1.0 ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, 1.0, 1.0 ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, -1.0, 0.0 ), 0 );
    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, 1.0, -1.0 ), 0 );
    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, -1.0, 1.0 ), 0 );
    CU_ASSERT_EQUAL( sph_polygon_test_inside( poly, 0.5, 0.5 ), 1 );


    sph_polygon_delete(poly);
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_test_inside_im_hex_case() {
    sph_polygon*        poly    = NULL;
    sph_polygon*        pixelpoly    = NULL;
    sph_ifs_lenslet_model*    model    = NULL;

    int                    nx        = 582;
    int                    ny         = 582;
    int                    xx        = 0;
    int                    yy        = 0;
    int                    minx     = 0;
    int                    miny     = 0;
    int                    maxx     = 0;
    int                    maxy     = 0;
    cpl_image*            image    = NULL;
    double                area    = 0.0;
    double                overlap = 0.0;
    int                    nn    = 0;
    double                    cosan = cos(CPL_MATH_PI*0.0/180.0);
    double                    sinan = sin(CPL_MATH_PI*0.0/180.0);
    double                    pixsize = 0.0;
    sph_polygon_workspace*  ws = NULL;

    ws = sph_polygon_workspace_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( ws );

    model = sph_ifs_lenslet_model_new();
    poly = sph_ifs_lenslet_model_get_poly(model,0,0);
    sph_polygon_rotate_around(poly,0.0,0.0,cosan,sinan,0.0,0.0);
    pixsize = 4.0 * model->lensize_microns * model->lenslets_per_side / nx;

    for (nn = 0; nn < poly->npoints; ++nn) {
        poly->points[nn]->x/=pixsize;
        poly->points[nn]->y/=pixsize;
        poly->points[nn]->x+=nx/2.0;
        poly->points[nn]->y+=ny/2.0;
    }
    //sph_polygon_enlarge(poly,2.0);
    image = sph_test_image_tools_create_flat_image_double(nx,ny,0.0);
    area = sph_polygon_area(poly);
    maxx = sph_polygon_get_right(poly);
    if ( maxx >= nx ) maxx = nx-1;
    maxy = sph_polygon_get_top(poly);
    if ( maxy >= ny ) maxy = ny-1;
    minx = sph_polygon_get_left(poly);
    if ( minx < 0 ) minx = 0;
    miny = sph_polygon_get_bottom(poly);
    if ( miny < 0 ) miny = 0;
    for (xx = minx; xx <= maxx; ++xx) {
        for (yy = miny; yy <= maxy; ++yy) {
                pixelpoly = sph_polygon_new();
                sph_polygon_add_point(pixelpoly,xx,yy);
                sph_polygon_add_point(pixelpoly,xx+1.0,yy);
                sph_polygon_add_point(pixelpoly,xx+1.0,yy+1.0);
                sph_polygon_add_point(pixelpoly,xx,yy+1.0);
                sph_polygon_clip_to_target(pixelpoly,poly,ws);
                overlap = sph_polygon_area(pixelpoly);
                if ( overlap > 0.0 ) {
                    if ( sph_polygon_test_inside(poly,xx+1.0,yy) < 1 &&
                            sph_polygon_test_inside(poly,xx+1.0,yy+1.0) < 1 &&
                            sph_polygon_test_inside(poly,xx,yy+1.0) < 1 &&
                            sph_polygon_test_inside(poly,xx,yy) < 1) {
                        if ( sph_polygon_test_inside(poly,xx,yy) < 1 )
                            sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_ERROR,"Point was: %f, %f",xx+0.0,yy+0.0);
                        if ( sph_polygon_test_inside(poly,xx+1.0,yy) < 1 )
                            sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_ERROR,"Point was: %f, %f",xx+1.0,yy+0.0);
                        if ( sph_polygon_test_inside(poly,xx+1.0,yy+1.0) < 1 )
                            sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_ERROR,"Point was: %f, %f",xx+1.0,yy+1.0);
                        if ( sph_polygon_test_inside(poly,xx,yy+1.0) < 1 )
                            sph_error_raise( SPH_ERROR_GENERAL, __FILE__,__func__,__LINE__,SPH_ERROR_ERROR,"Point was: %f, %f",xx+0.0,yy+1.0);
                        CU_FAIL("Failed to have inside point with positive overlap!");
                    }
                    cpl_image_set(image,xx+1,yy+1,overlap);
                }
                sph_polygon_delete(pixelpoly); pixelpoly = NULL;
        }
    }
    cpl_image_save(image,"testim.fits", CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    sph_polygon_workspace_delete(ws);
    sph_polygon_delete(poly);
    sph_ifs_lenslet_model_delete(model);
    cpl_image_delete(image);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_test_same_side() {
    sph_polygon*            poly    = NULL;
    sph_error_code          rerr    = 0;
    sph_line*               workline = NULL;
    sph_point*              point1 = NULL;
    sph_point*              point2 = NULL;
    sph_point*              point3 = NULL;
    sph_point*              point4 = NULL;
    sph_point*              point5 = NULL;
    sph_point*              point6 = NULL;
    sph_point*              point7 = NULL;
    poly = sph_polygon_new();
    workline = sph_line_new(0.0,0.0,0.0,0.0);

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    CU_ASSERT_PTR_NOT_NULL_FATAL( workline );
    rerr = sph_polygon_add_point( poly, 0.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 0.0, 1.0);

    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    point1 = sph_point_new(0.0,0.0);
    point2 = sph_point_new(1.0,0.0);
    point3 = sph_point_new(1.0,1.0);
    point4 = sph_point_new(0.1,0.1);
    point5 = sph_point_new(1.1,1.1);
    point6 = sph_point_new(-0.1,-0.1);
    point7 = sph_point_new(0.0,0.5);

    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 1, point1, workline ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 1, point2, workline ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 1, point3, workline ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 2, point3, workline ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 3, point4, workline ), 1 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 3, point5, workline ), 0 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 4, point6, workline ), 0 );
    CU_ASSERT_EQUAL( sph_polygon_test_same_side( poly, 4, point7, workline ), 1 );

    sph_polygon_delete(poly); poly = NULL;
    sph_line_delete(workline); workline = NULL;
    sph_point_delete(point1); point1 = NULL;
    sph_point_delete(point2); point2 = NULL;
    sph_point_delete(point3); point3 = NULL;
    sph_point_delete(point4); point4 = NULL;
    sph_point_delete(point5); point5 = NULL;
    sph_point_delete(point6); point6 = NULL;
    sph_point_delete(point7); point7 = NULL;

    CU_ASSERT_EQUAL(cpl_error_get_code(),CPL_ERROR_NONE);
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_clipping function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_clip_to_target_caseA() {
    sph_polygon*        poly    = NULL;
    sph_polygon*        target  = NULL;
    sph_polygon_workspace*  ws = NULL;
    sph_error_code      rerr    = 0;

    poly = sph_polygon_new();
    ws = sph_polygon_workspace_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( ws );
    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    target = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_polygon_add_point( target, 0.0, 0.0);
    rerr |= sph_polygon_add_point( target, 2.0, 0.0);
    rerr |= sph_polygon_add_point( target, 2.0, 2.0);
    rerr |= sph_polygon_add_point( target, 0.0, 2.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    rerr = sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 3.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    rerr = sph_polygon_clip_to_target(  poly, target, ws );

    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL_FATAL( poly->npoints, 4 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->x, 2.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->y, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->x, 2.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->y, 2.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[3]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[3]->y, 2.0, 1.0e-15 );
    sph_polygon_workspace_delete(ws);
    sph_polygon_delete(poly);
    sph_polygon_delete(target);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_clipping function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_clip_to_target_caseB() {
    sph_polygon*        poly    = NULL;
    sph_polygon*        target    = NULL;
    sph_error_code        rerr    = 0;
    sph_polygon_workspace*  ws = NULL;

    poly = sph_polygon_new();
    ws = sph_polygon_workspace_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( ws );

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    target = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_polygon_add_point( poly, 0.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 1.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    rerr = sph_polygon_add_point( target, -20.0, -20.0);
    rerr |= sph_polygon_add_point( target, 2.0, 0.0);
    rerr |= sph_polygon_add_point( target, 10.0, 10.0);
    rerr |= sph_polygon_add_point( target, 0.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    rerr = sph_polygon_clip_to_target(  poly, target, ws );

    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL_FATAL( poly->npoints, 3 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 0.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->y, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->x, 0.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->y, 0.0, 1.0e-15 );
    sph_polygon_workspace_delete(ws);
    sph_polygon_delete(poly);
    sph_polygon_delete(target);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_clipping function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_clip_to_target_caseC() {
    sph_polygon*        poly    = NULL;
    sph_polygon*        target  = NULL;
    sph_error_code      rerr    = 0;
    sph_polygon_workspace*  ws = NULL;

    poly = sph_polygon_new();
    ws = sph_polygon_workspace_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( ws );

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    target = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_polygon_add_point( target, 0.0, 0.0);
    rerr |= sph_polygon_add_point( target, 1.0, 0.0);
    rerr |= sph_polygon_add_point( target, 1.0, 1.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    rerr = sph_polygon_add_point( poly, -20.0, -20.0);
    rerr |= sph_polygon_add_point( poly, 2.0, 0.0);
    rerr |= sph_polygon_add_point( poly, 10.0, 10.0);
    rerr |= sph_polygon_add_point( poly, 0.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    rerr = sph_polygon_clip_to_target(  poly, target, ws );

    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_TRUE( poly->npoints < 4 );
    CU_ASSERT_TRUE( poly->npoints > 2 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 0.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 0.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->y, 0.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->y, 1.0, 1.0e-15 );
    sph_polygon_workspace_delete(ws);
    sph_polygon_delete(poly);
    sph_polygon_delete(target);
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_clipping function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_calculate_overlap() {
    sph_polygon*        poly    = NULL;
    sph_polygon*        target  = NULL;
    sph_error_code      rerr    = 0;
    double              result  = 0.0;
    sph_polygon_workspace*  ws = NULL;

    poly = sph_polygon_new();
    ws = sph_polygon_workspace_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( ws );

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    target = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_polygon_add_point( target, 0.0, 0.0);
    rerr |= sph_polygon_add_point( target, 2.0, 0.0);
    rerr |= sph_polygon_add_point( target, 2.0, 2.0);
    rerr |= sph_polygon_add_point( target, 0.0, 2.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    rerr = sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 3.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    result = sph_polygon_calculate_overlap( poly, target, ws );

    CU_ASSERT_DOUBLE_EQUAL( result, 1.0, 1.0e-10 );

    sph_polygon_delete(poly);
    sph_polygon_delete(target);
    sph_polygon_workspace_delete(ws);
    return;
}
void cutest_polygon_calculate_overlap2() {
    sph_polygon*        poly    = NULL;
    sph_polygon*        target  = NULL;
    double                offset_x = 0.0;
    double                offset_y = 0.0;
    sph_polygon_workspace*  ws = NULL;

    ws = sph_polygon_workspace_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( ws );

    poly = sph_polygon_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    sph_polygon_add_point(poly,offset_x + 0.0, offset_y + 0.0);
    sph_polygon_add_point(poly,offset_x + 1.0, offset_y + 0.0);
    sph_polygon_add_point(poly,offset_x + 1.0, offset_y + 1.0);
    sph_polygon_add_point(poly,offset_x + 0.0, offset_y + 1.0);

    target = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( target );
    sph_polygon_add_point(target,0.0,0.0);
    sph_polygon_add_point(target,1.01,0.0);
    sph_polygon_add_point(target,1.01,1.01);
    sph_polygon_add_point(target,0.0,1.01);
    CU_ASSERT_DOUBLE_EQUAL( sph_polygon_calculate_overlap(poly,target,ws), 1.0, 0.1);
    CU_ASSERT_DOUBLE_EQUAL( sph_polygon_calculate_overlap(target,poly,ws), 1.0, 0.1);
    CU_ASSERT_DOUBLE_EQUAL( sph_polygon_calculate_overlap(poly,poly,ws), 1.0, 0.1);
    CU_ASSERT_DOUBLE_EQUAL( sph_polygon_calculate_overlap(target,target,ws), 1.0, 0.1);
    sph_polygon_delete(poly); poly = NULL;
    sph_polygon_delete(target); target = NULL;
    sph_polygon_workspace_delete(ws);
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_clipping function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_calculate_overlap_parallel() {
    sph_polygon*        target  = NULL;
    sph_error_code      rerr    = 0;
    double              result  = 0.0;
#ifdef _OPENMP
    sph_polygon_workspace*  ws[omp_get_max_threads()];
#else
    sph_polygon_workspace*  ws[4];
#endif
    int                 ii = 0;
    int                 npolys = 1000000;
    sph_polygon*        polys[npolys];
    int                 tn = 0;

    for (ii = 0; ii < npolys; ++ii) {
        polys[ii] = sph_polygon_new();
        rerr = sph_polygon_add_point( polys[ii], 1.0, 1.0);
        rerr |= sph_polygon_add_point( polys[ii], 3.0, 1.0);
        rerr |= sph_polygon_add_point( polys[ii], 3.0, 3.0);
        rerr |= sph_polygon_add_point( polys[ii], 1.0, 3.0);
    }

    target = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_polygon_add_point( target, 0.0, 0.0);
    rerr |= sph_polygon_add_point( target, 2.0, 0.0);
    rerr |= sph_polygon_add_point( target, 2.0, 2.0);
    rerr |= sph_polygon_add_point( target, 0.0, 2.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Start timer %d: ",
            sph_time_start_timer(1) );

#ifdef _OPENMP
    for (ii = 0; ii < omp_get_max_threads(); ++ii) {
#else
    for (ii = 0; ii < 4; ++ii) {
#endif
    	ws[ii] = sph_polygon_workspace_new();
    }
#pragma omp parallel for shared(ws) private(ii,tn)
    for (ii = 0; ii < npolys; ++ii) {
#ifdef _OPENMP
        tn = omp_get_thread_num();
#endif
        result = sph_polygon_calculate_overlap( polys[ii], target, ws[tn] );
//        if ( fabs (result - 1.0 ) > 1.0e-10 ) {
            //CU_FAIL_FATAL("Incorrect");
//        }
    }
    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Stop time %f: ",
            sph_time_stop_timer(1) );
    for (ii = 0; ii < npolys; ++ii) {
        sph_polygon_delete(polys[ii]);
    }
#ifdef _OPENMP
    for (ii = 0; ii < omp_get_max_threads(); ++ii) {
#else
    for (ii = 0; ii < 4; ++ii) {
#endif
        sph_polygon_workspace_delete(ws[ii]);
    }
    sph_polygon_delete(target); target = NULL;
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_rotate function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_rotate_around() {
    sph_polygon*        poly    = NULL;
    sph_error_code      rerr    = 0;

    poly = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    rerr = sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 3.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    rerr = sph_polygon_rotate_around(  poly, 1.0, 1.0,
                                       cos( 90.0/180.0*SPH_PI ), sin( 90.0/180.0*SPH_PI ),
                                       0.0, 0.0 );

    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL_FATAL( poly->npoints, 4 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->x, 1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->y, 3.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->x, -1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->y, 3.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[3]->x, -1.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[3]->y, 1.0, 1.0e-15 );
    sph_polygon_delete(poly); poly = NULL;
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_rotate function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_area() {
    sph_polygon*        poly    = NULL;
    double              result  = 0.0;
    sph_error_code      rerr    = CPL_ERROR_NONE;

    poly = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    rerr = sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 3.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    result = sph_polygon_area(  poly );


    CU_ASSERT_DOUBLE_EQUAL( result, 4.0, 1.0e-15 );
    sph_polygon_delete(poly); poly = NULL;
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_rotate function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_rotate_around_performance() {
    sph_polygon*        poly    = NULL;
    sph_error_code      rerr    = 0;
    double              cosang  = cos( 90.0/180.0*SPH_PI );
    double              sinang  = sin( 90.0/180.0*SPH_PI );
    int                 maxi    = 4 * 4200000 + 1;
    int                 ii      = 0;

    poly = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
    rerr = sph_polygon_add_point( poly, 1.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 1.0);
    rerr |= sph_polygon_add_point( poly, 3.0, 3.0);
    rerr |= sph_polygon_add_point( poly, 1.0, 3.0);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    for ( ii = 0; ii < maxi; ++ii ) {
        rerr = sph_polygon_rotate_around(  poly, 1.0, 1.0,
                                           cosang, sinang,
                                           0.0, 0.0 );
    }
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );

    CU_ASSERT_EQUAL_FATAL( poly->npoints, 4 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->x, 1.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[0]->y, 1.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->x, 1.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[1]->y, 3.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->x, -1.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[2]->y, 3.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[3]->x, -1.0, 1.0e-8 );
    CU_ASSERT_DOUBLE_EQUAL( poly->points[3]->y, 1.0, 1.0e-8 );
    sph_polygon_delete(poly); poly = NULL;
    return;
}


/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_polygon_delete function.
 */
/*----------------------------------------------------------------------------*/
void cutest_polygon_delete() {
    sph_polygon*        result    = NULL;

    result = sph_polygon_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL( result );

    sph_polygon_delete( result );
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {

    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("polygon_test", cutest_init_polygon_testsuite,
                          cutest_clean_polygon_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "sph_polygon_new",
                             cutest_polygon_new ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    if ( NULL == CU_add_test(pSuite, "sph_polygon_add_point",
                              cutest_polygon_add_point ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_get_edge",
                              cutest_polygon_get_edge ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_test_inside",
                              cutest_polygon_test_inside ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_test_same_side",
                              cutest_polygon_test_same_side ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_clip_to_target_caseA",
                              cutest_polygon_clip_to_target_caseA ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_clip_to_target_caseB",
                              cutest_polygon_clip_to_target_caseB ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_clip_to_target_caseC",
                              cutest_polygon_clip_to_target_caseC ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_calculate_overlap",
                              cutest_polygon_calculate_overlap ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_rotate_around",
                              cutest_polygon_rotate_around ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_rotate_around_performance",
                              cutest_polygon_rotate_around_performance ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_area",
                              cutest_polygon_area ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_inside_im_hex",
                              cutest_polygon_test_inside_im_hex_case ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_duplicate",
                              cutest_polygon_duplicate ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_calculate_overlap2",
                              cutest_polygon_calculate_overlap2 ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_calculate_overlap parralel",
                              cutest_polygon_calculate_overlap_parallel ) )
     {
        CU_cleanup_registry();
        return CU_get_error();
     }
    if ( NULL == CU_add_test(pSuite, "sph_polygon_delete",
                             cutest_polygon_delete ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
